/***********************************
*     PROJECT :ComboPic            
*        FILE :ComboSLE.PRG.prg
*  Programmer :James Loughner
*  Created    :11/24/03
*          All rights reserved
*  Combobox SLEPic Class   
*  requires SLEPic Class    
*  Version 1.0
*
*  ComboBoxSLE is a replacement for XbpComboBox. 
*  their behavours are very close but not identical
*  
*  The main difference you'll notice is that be default
*  ComboBoxSLE uses Clipper navigation keys ie up/down/return.
*  XbpComboBox uses UP/DOWN when the SLE is in focus to
*  scroll through the list. For ComboBoxSLE the default is
*  Shift-UP/DOWN scroll UP/DOWN moves focus to the NEXT/PREVIOUS
*  XBP. This can be overriden by the ScrollUpKey/ScrollDownKey
*  iVars. (see below) 
* 
*  Progarmming Differences
*     Reference to the SLE part is now SLEPic instead of XbpSLE!!!
*     To set a picture oXbp:SLEPic:Picture := "XX-99-!!"  etc.
*     (see SLEPic class file for more info)
*
*     note: Set/GetData now referes to SLE part. You may use
*           oXbp:SET/GETDATA() or oXbp:SLEPic:SET/GETDATA()
*           to reference list part use
*           oXbp:XbpListBox:SET/GETDATA()
*
*     New iVars
*         IsProtected  Is .T. if Protected
*         IsDisabled   Is .T. if disabled
*         BGColor      Display BG color
*         FGColor      Display FG color
*         SelectedBG   Selected BG color
*         SelectedFG   Selected FG color
*         DisabledBG   Disabled BG color
*         DisabledFG   Disabled FG color
*         ProtectBG    Protected BD color
*         ProtectFG    Protected FG color
*         SleHeight    You may set the height of the SLE part before 
*                      Create() or before a setsize() or SetPos()
*         ListActivateKey  Key to activate list   default F4
*         ScrollDownKey    key to scroll down when SLE is in focus 
*                          default Shift-down
*         ScrollUpKey      key to scroll up when SLE is in focus 
*                          default Shift-up
*         SLEAlignment .T. Set postion relative to SLE part
*                      .F. Set position relative to the List part
*
*    New Methods
*
*        METHOD Protect(lVal) Turn on/off protected mode    
*
*    Note SetColorFG/BG method sets both XXColor and SelectedXX to the 
*         same colors  
*    Note The easiest way to use is to set two define statements at the 
*         begining of any file that uses XbpComboBox and link the 
*         combosle and SLEPic files
*         #define XbpComboBox ComboBoxSLE      
*         #define XbpSLE SLEPic
*     
*************************************/
 
#include "Gra.ch"
#include "Xbp.ch"
#include "Common.ch"
#include "Appevent.ch"
#include "Font.ch"
#include "Appbrow.ch" 
#include "Dmlb.ch" 
#include "dll.ch"
#include "inkey.ch"
#include "Nls.ch"
// unRemark next line to run stand alone
//#define TESTCOMBOSLE

#ifdef TESTCOMBOSLE
PROCEDURE Main()
   LOCAL nEvent, mp1, mp2, oXbp 
   LOCAL oDlg,oSLE,oSLE1,oSLE2,oSLE3,oSLE4,oSLE5,oSLE6,oSLE7
    
   oDlg := XbpDialog():new( AppDesktop(),,{30,50}, {400,450} ) 
   oDlg:title := "Test Combo" 
   oDlg:create() 
   
   oSLE7 := ComboBoxSLE():New(oDlg:drawingarea,,{10,285}, {100,100})
   oSLE7:SleHeight := 20  // note this defaults to 20 
   oSLE7:TabStop := .T. 
//   oSle7:Bufferlength := 10  // test short buffer
   oSle7:FGColor := GRA_CLR_WHITE
   oSle7:BGColor := GRA_CLR_DARKGREEN
//   oSLE7:Type := XBPCOMBO_DROPDOWNLIST    // set  type
//   oSLE7:Type := XBPCOMBO_SIMPLE    // set  type
   oSle7:SLEAlignment := .T.
   oSle7:Create()        
   // note these colors must be set after the create if wanted!!!
   oSle7:SelectedFG := GRA_CLR_WHITE
   oSle7:SelectedBG := GRA_CLR_DARKRED
   oSle7:ProtectFG := GRA_CLR_RED
   oSle7:ProtectBG := GRA_CLR_WHITE 
   oSle7:AddItem("12:00"+CHR(9)+"test")
   oSle7:AddItem("12:30"+CHR(9)+"test")
   oSle7:AddItem("01:00"+CHR(9)+"test")
   oSle7:AddItem("01:30"+CHR(9)+"test")  
   oSle7:AddItem("02:00"+CHR(9)+"test")
   oSle7:AddItem("02:30"+CHR(9)+"test")
   oSle7:AddItem("03:00"+CHR(9)+"test")
   oSle7:AddItem("03:30"+CHR(9)+"test")  
   oSle7:AddItem("04:00"+CHR(9)+"test")
   oSle7:AddItem("04:30"+CHR(9)+"test")
   oSle7:AddItem("05:00"+CHR(9)+"test")
   oSle7:AddItem("05:30"+CHR(9)+"test")  
   oSle7:XbpListBox:SetTabStops( { 8 })
   oSle7:SLEPic:Picture := "99:99XXXXXXXXX"  // set pic after create
   oSle7:SLEPic:WindAtK := .F.  // set cursor to beginning
   oSle7:SLEPic:Setdata("03:00")

   // set scroll keys to normal windows
//   oSle7:ScrollDownKey := xbeK_DOWN    
//   oSle7:ScrollUpKey := xbeK_UP  
//   oSle7:Disable()
//   oSle7:Protect(.T.)

   oSLE := XbpPushButton():New(oDlg:drawingarea,,{10,250},{120,30})
   oSLE:Caption := "Test"
   oSLE:Activate := {|| msgbox("test")}
   oSLE:Create()
      
// regular Xbp
   oSLE6 := XbpComboBox():New(oDlg:drawingarea,,{10,150}, {100,100})
   oSLE6:TabStop := .T. 
//   oSLE6:Type := XBPCOMBO_DROPDOWNLIST 
//   oSLE6:Type := XBPCOMBO_SIMPLE    // set  type
   oSle6:Create()        
   oSle6:XbpSLE:Setdata("03:30")
   oSle6:AddItem("12:00"+CHR(9)+"test")
   oSle6:AddItem("12:30"+CHR(9)+"test")
   oSle6:AddItem("01:00"+CHR(9)+"test")
   oSle6:AddItem("01:30"+CHR(9)+"test")  
   oSle6:AddItem("02:00"+CHR(9)+"test")
   oSle6:AddItem("02:30"+CHR(9)+"test")
   oSle6:AddItem("03:00"+CHR(9)+"test")
   oSle6:AddItem("03:30"+CHR(9)+"test")  
   oSle6:AddItem("04:00"+CHR(9)+"test")
   oSle6:AddItem("04:30"+CHR(9)+"test")
   oSle6:AddItem("05:00"+CHR(9)+"test")
   oSle6:AddItem("05:30"+CHR(9)+"test")  
   oSle7:XbpListBox:SetTabStops( { 8 })

   nEvent := 0 
   DO WHILE nEvent <> xbeP_Close 
      nEvent := AppEvent( @mp1, @mp2, @oXbp ) 
      oXbp:handleEvent( nEvent, mp1, mp2 ) 
   ENDDO 
RETURN



PROCEDURE APPSYS()
RETURN

#endif

CLASS ComboBoxSLE FROM SLEPic
EXPORTED:
   METHOD Init, Create,destroy 
   METHOD CalcSizes 
   METHOD DrawTick
   METHOD Paint 
   METHOD ListboxFocus
   METHOD ItemMarked
   METHOD ItemSelected  
   METHOD KillListInputFocus 
   METHOD SetListInputFocus 
   METHOD Keyboard
   VAR CurSize
   VAR CurPos
   VAR aSLESize 
   VAR IsVis
   VAR aButtonSize
   VAR aListSize
   VAR changed
   VAR IsProtected,ProtectBG,ProtectFG
   VAR IsDisabled,DisabledBG,DisabledFG 
   VAR XbpListBox
   VAR BGColor,FGColor   
   VAR SelectedBG,SelectedFG
   VAR LastTabStop
   VAR oPush 
   VAR IsListfocus
   VAR SleHeight
   VAR Type 
   VAR ListActivateKey 
   VAR ScrollDownKey,ScrollUpKey 
   VAR SLEAlignment  
//   VAR XbpSLE
   ASSIGN METHOD SetDisabledBG VAR DisabledBG
   ASSIGN METHOD SetDisabledFG VAR DisabledFG
   ASSIGN METHOD SetProtectBG VAR ProtectBG
   ASSIGN METHOD SetProtectFG VAR ProtectFG
   ASSIGN METHOD SetSelectedBG VAR SelectedBG
   ASSIGN METHOD SetSelectedFG VAR SelectedFG
   METHOD AddItem 
   METHOD clear  
   METHOD delItem  
//   METHOD getData  
   METHOD getItem   
   METHOD getItemsHeight   
   METHOD getTabstops   
   METHOD getTopItem   
   METHOD getVisibleItems   
   METHOD InsItem   
   METHOD numItems   
   METHOD setColumnWidth   
//   METHOD setData   
   METHOD setItem   
   METHOD setItemsHeight   
   METHOD setTabstops   
   METHOD setTopItem   
   METHOD listboxSize  
   METHOD sleSize  
   METHOD SetSize
   METHOD SetPos 
   METHOD SetPosAndSize

   METHOD Disable
   METHOD Enable
   METHOD Protect
   METHOD SetColorFG
   METHOD SetColorBG
//   METHOD undo 
ENDCLASS

METHOD ComboBoxSLE:Init(oParent,oOwner,aPos,aSize,aPresParam,lVisible)

   IF lVisible != NIL
      ::IsVis := lVisable
   ENDIF   
   IF aSize != NIL
      ::CurSize := aSize
   ENDIF   
   IF aPos != NIL
      ::CurPos := aPos
   ENDIF   
   ::SLEPic:Init(oParent,oOwner,,,aPresParam,.T.)
   ::XbpListBox := XbpListBox():New(oParent,oOwner,,,aPresParam,.F.)
   ::oPush := XbpPushButton():New(::SLEPic,oOwner,,,aPresParam,.T.)
   ::IsProtected := .F.
   ::ProtectBG := GRA_CLR_BLACK
   ::ProtectFG := GRA_CLR_WHITE
   ::IsDisabled := .F.
   ::DisabledBG := GRA_CLR_PALEGRAY
   ::DisabledFG := GRA_CLR_DARKGRAY
   ::LastTabStop := ::TabStop
   ::SelectedBG := XBPSYSCLR_ENTRYFIELD
   ::SelectedFG := GRA_CLR_BLACK
   ::BGColor := XBPSYSCLR_ENTRYFIELD
   ::FGColor := GRA_CLR_BLACK 
   ::IsListfocus := .F. 
   ::IsDisabled := .F. 
   ::SleHeight := 20 
   ::Type := XBPCOMBO_DROPDOWN  
   ::ListActivateKey := xbeK_F4
   ::ScrollDownKey := xbeK_SH_DOWN    
   ::ScrollUpKey := xbeK_SH_UP
   ::SLEAlignment := .F.  
//   ::XbpSLE := ::SLEPic
RETURN Self

METHOD ComboBoxSLE:Create(oParent,oOwner,aPos,aSize,aPresParam,lVisible)
   IF lVisible != NIL
      ::IsVis := lVisable
   ENDIF   
   IF aSize != NIL
      ::CurSize := aSize
   ENDIF   
   IF aPos != NIL
      ::CurPos := aPos
   ENDIF   
//   ::oBack:create(oParent,oOwner,aPos,aSize)
//   ::oBack:SetColorBG(XBPSYSCLR_TRANSPARENT)
   ::SLEPic:ClipChildren:=.T.
   ::SLEPic:Create(oParent,oOwner,,,aPresParam,.T.)   
   ::XbpListBox:Create(oParent,oOwner,,,aPresParam,.F.)
   ::oPush:Activate := {|| IF(::XbpListBox:IsVisible(),::ListboxFocus(::Type=XBPCOMBO_SIMPLE),::ListboxFocus(.T.))  }
   ::oPush:pointerFocus := .F.
   ::oPush:Paint := {||::DrawTick()}
   ::oPush:Create(::SLEPic,,,,,.T.)
   ::XbpListBox:Keyboard := {|nKey,y,o|IF(nKey=::ListActivateKey.AND.!::IsDisabled.AND.::Type!=XBPCOMBO_SIMPLE,Eval(::oPush:Activate),NIL)} 
   ::CalcSizes(::CurSize)
   ::SLEPic:SetInputFocus := {||::DrawTick()}
   ::DrawTick()
   ::XbpListBox:ItemMarked := {||::ItemMarked()}
   ::XbpListBox:ItemSelected := {||::ItemSelected()}
   ::XbpListBox:KillInputFocus := {||::KillListInputFocus()}
   ::XbpListBox:SetInputFocus := {||::XbpListBox:SetColorBG(::SelectedBG),::XbpListBox:SetColorFG(::SelectedFG)}
   ::XbpListBox:SetColorBG(::BGColor)
   ::XbpListBox:SetColorFG(::FGColor)
   ::SLEPic:SetColorBG(::BGColor)
   ::SLEPic:SetColorFG(::FGColor)  
   IF ::IsVis = NIL
      ::IsVis = .T.
   ENDIF
   IF ::IsVis
      ::SLEPic:Show()   
      IF ::Type!=XBPCOMBO_SIMPLE 
         ::XbpListBox:Hide()
         ::oPush:Show()
      ELSE
         ::IsListfocus := .T. 
         ::XbpListBox:Show()
         ::XbpListBox:Configure()
         ::oPush:Hide()
      ENDIF   
   ENDIF   
   IF ::Type = XBPCOMBO_DROPDOWNLIST 
      ::SLEPic:DisabledBG := ::SelectedBG
      ::SLEPic:DisabledFG := ::SelectedFG
      ::SLEPic:Disable()
   ENDIF
RETURN Self

METHOD ComboBoxSLE:CalcSizes(aSize)
   IF ::SLEAlignment
      // Calc SLEPic size 
      ::SLEPic:SetSize({aSize[1],::SleHeight})
      ::aSLESize := {aSize[1],::SleHeight}
      ::SLEPic:SetPos(::CurPos)
      // Calc Push button
      ::oPush:SetSize({::aSLESize[2]/2,::aSLESize[2]})
      ::aButtonSize := {::aSLESize[2]/2,::aSLESize[2]}
      ::oPush:SetPos({::aSLESize[1]-::aButtonSize[1]-4,0})
      // Calc List
      ::XbpListBox:SetSize({aSize[1],aSize[2]-::aSLESize[2]})
      ::aListSize := {aSize[1],aSize[2]-::aSLESize[2]}
      ::XbpListBox:SetPos({::CurPos[1],::CurPos[2]-::aListSize[2]})
   ELSE
      // Calc SLEPic size 
      ::SLEPic:SetSize({aSize[1],::SleHeight})
      ::aSLESize := {aSize[1],::SleHeight}
      ::SLEPic:SetPos({::CurPos[1],::CurPos[2]+aSize[2]-::aSLESize[2]})
      // Calc Push button
      ::oPush:SetSize({::aSLESize[2]/2,::aSLESize[2]})
      ::aButtonSize := {::aSLESize[2]/2,::aSLESize[2]}
      ::oPush:SetPos({::aSLESize[1]-::aButtonSize[1]-4,0})
      // Calc List
      ::XbpListBox:SetSize({aSize[1],aSize[2]-::aSLESize[2]})
      ::aListSize := {aSize[1],aSize[2]-::aSLESize[2]}
      ::XbpListBox:SetPos(::CurPos)
   ENDIF   
RETURN Self  

METHOD ComboBoxSLE:SetSize(aSize)
   ::CurSize := aSize
   ::CalcSizes(aSize)
RETURN .T. 

METHOD ComboBoxSLE:SetPos(aPos)
   ::CurPos := aPos 
   ::CalcSizes(::CurSize)
RETURN .T. 

METHOD ComboBoxSLE:SetPosAndSize(aPos,aSize)
   ::SetSize(aSize)
   ::SetPos(aPos)
RETURN .T. 

METHOD ComboBoxSLE:SetDisabledBG(nHue)
LOCAL nRtn := ::DisabledBG
   IF nHue != NIL
      ::DisabledBG := nHue
      ::SLEPic:DisabledBG := nHue
   ENDIF    
RETURN nRtn

METHOD ComboBoxSLE:SetDisabledFG(nHue)  
LOCAL nRtn := ::DisabledFG 
   IF nHue != NIL
      ::DisabledFG := nHue
      ::SLEPic:DisabledFG := nHue
   ENDIF    
RETURN nRtn

METHOD ComboBoxSLE:SetProtectBG(nHue)  
LOCAL nRtn := ::ProtectBG 
   IF nHue != NIL
      ::ProtectBG := nHue
      ::SLEPic:ProtectBG := nHue
   ENDIF    
RETURN nRtn

METHOD ComboBoxSLE:SetProtectFG(nHue)  
LOCAL nRtn := ::ProtectFG 
   IF nHue != NIL
      ::ProtectFG := nHue
      ::SLEPic:ProtectFG := nHue
   ENDIF    
RETURN nRtn

METHOD ComboBoxSLE:SetSelectedBG(nHue)  
LOCAL nRtn := ::SelectedBG 
   IF nHue != NIL
      ::SelectedBG := nHue
      ::SLEPic:SelectedBG := nHue
   ENDIF    
RETURN nRtn

METHOD ComboBoxSLE:SetSelectedFG(nHue)  
LOCAL nRtn := ::SelectedFG 
   IF nHue != NIL
      ::SelectedFG := nHue
      ::SLEPic:SelectedFG := nHue
   ENDIF    
RETURN nRtn

METHOD ComboBoxSLE:Disable()
  ::XbpListBox:Disable()
  ::SLEPic:Disable()
  ::oPush:Disable()
  ::IsDisabled := .T.
RETURN Self

METHOD ComboBoxSLE:Enable
  ::XbpListBox:Enable()
  ::SLEPic:Enable()
  ::oPush:Enable()
  ::IsDisabled := .F.
RETURN Self

METHOD ComboBoxSLE:Protect(lState)
   IF EMPTY(lState)
      lState := !::IsProtected
   ENDIF   
   ::IsProtected := lState
   IF lstate
      ::LastTabStop := ::TabStop
      ::TabStop := .F.
      ::SLEPic:Protect(lstate)
      ::oPush:Disable()
   ELSE   
      ::TabStop := ::LastTabStop
      ::SLEPic:Protect(lstate)
      ::oPush:Enable()
   ENDIF
RETURN Self

METHOD ComboBoxSLE:SetColorFG(nHue)
LOCAL nRtn := ::FGColor 
   IF nHue != NIL
      ::FGColor := nHue
      ::SelectedFG := nHue
      ::XbpListBox:SetColorFG(nHue)
      ::SLEPic:SetColorFG(nHue)
   ENDIF   
RETURN nRtn

METHOD ComboBoxSLE:SetColorBG(nHue)
LOCAL nRtn := ::BGColor 
   IF nHue != NIL
      ::BGColor := nHue
      ::SelectedFG := nHue
      ::XbpListBox:SetColorBG(nHue)
      ::SLEPic:SetColorBG(nHue)
   ENDIF   
RETURN nRtn 


METHOD ComboBoxSLE:DrawTick() 
LOCAL oPS:=::oPush:LockPS() ,Attr :=Array( GRA_AL_COUNT )
   Attr[GRA_AL_COLOR] := GRA_CLR_BLACK
   GraSetAttrLine(oPS,Attr)
   GRALINE(oPS,{::oPush:CurrentSize()[1]/2,::oPush:CurrentSize()[2]/2-2},{::oPush:CurrentSize()[1]/2+2,::oPush:CurrentSize()[2]/2+1-2})
   GRALINE(oPS,,{::oPush:CurrentSize()[1]/2-2,::oPush:CurrentSize()[2]/2+1-2})
   GRALINE(oPS,,{::oPush:CurrentSize()[1]/2,::oPush:CurrentSize()[2]/2-2})
   ::oPush:UnLockPS()
RETURN Self

METHOD ComboBoxSLE:Paint() 
   ::DrawTick()
RETURN Self

METHOD ComboBoxSLE:ListBoxFocus(lVal)   
LOCAL lRtn := ::IsListfocus ,I,lFound:=.F.
   IF lVal != NIL
      IF !lVal 
         ::XbpListBox:Hide()
         ::IsListfocus := .F.
         SetAppFocus(::SLEPic)
      ELSE
         IF EMPTY(::XbpListBox:Getdata()) .AND.::XbpListBox:NumItems() > 0 .AND. EMPTY(::SLEPic:EditBuffer())
            ::XbpListBox:SetData({1})
            ::SLEPic:Setdata(::XbpListBox:GetItem(1))
         ENDIF
         IF !EMPTY(::SLEPic:EditBuffer()).AND.::XbpListBox:NumItems() > 0
            FOR I = 1 TO ::XbpListBox:NumItems()
               IF ::XbpListBox:GetItem(I)== ALLTRIM(::SLEPic:EditBuffer())
                  ::XbpListBox:SetData({I})
                  lFound := .T.
                  EXIT
               ENDIF
            NEXT I
            IF !lFound
               ::XbpListBox:SetData({1})
               ::SLEPic:Setdata(::XbpListBox:GetItem(1))
            ENDIF
         ELSEIF ::XbpListBox:NumItems() > 0
            ::XbpListBox:SetData({1})
            ::SLEPic:Setdata(::XbpListBox:GetItem(1))
         ENDIF
         ::XbpListBox:Show()
         ::XbpListBox:SetColorBG(::SelectedBG)
         ::XbpListBox:SetColorFG(::SelectedFG)
         ::XbpListBox:toFront()
         ::IsListfocus := .T.
         SetAppFocus(::XbpListBox)
      ENDIF   
      ::DrawTick()
   ENDIF   
RETURN lVal  

METHOD ComboBoxSLE:ItemMarked() 
LOCAL nSel   
      nSel := ::XbpListBox:Getdata()[1]
      ::SLEPic:Setdata(::XbpListBox:GetItem(nSel))
RETURN Self

METHOD ComboBoxSLE:ItemSelected()
   ::ListBoxFocus(::Type=XBPCOMBO_SIMPLE)
   IF ::Type = XBPCOMBO_SIMPLE
      SetAppFocus(::SLEPic)
   ENDIF 
RETURN Self

METHOD ComboBoxSLE:KillListInputFocus()
   IF ::Type != XBPCOMBO_SIMPLE
      ::XbpListBox:Hide()
   ENDIF
   ::XbpListBox:SetColorBG(::BGColor)         
   ::XbpListBox:SetColorFG(::FGColor)
   ::IsListfocus := .F.
RETURN Self

METHOD ComboBoxSLE:SetListInputFocus()
   ::XbpListBox:SetColorFG(::SelectedFG)
   ::XbpListBox:SetColorBG(::SelectedBG)
   ::IsListfocus := .T.
RETURN Self

METHOD ComboBoxSLE:AddItem(cItem)
RETURN ::XbpListBox:addItem(cItem)   

METHOD ComboBoxSLE:clear()
RETURN ::XbpListBox:Clear()   
  
METHOD ComboBoxSLE:delItem(nItem)  
RETURN ::XbpListBox:delItem(nItem)   

//METHOD ComboBoxSLE:getData()  
//RETURN ::XbpListBox:Getdata()
  
METHOD ComboBoxSLE:getItem(nItem)  
RETURN ::XbpListBox:getItem(nItem)   

METHOD ComboBoxSLE:getItemsHeight()  
RETURN ::XbpListBox:getItemsHeight()
   
METHOD ComboBoxSLE:getTabstops()   
RETURN ::XbpListBox:getTabstops()

METHOD ComboBoxSLE:getTopItem()
RETURN ::XbpListBox:getTopItem()
   
METHOD ComboBoxSLE:getVisibleItems()
RETURN ::XbpListBox:getVisibleItems()
   
METHOD ComboBoxSLE:InsItem(nItem,cItem)  
RETURN ::XbpListBox:InsItem(nItem,cItem)
   
METHOD ComboBoxSLE:numItems()
RETURN ::XbpListBox:numItems()

METHOD ComboBoxSLE:setColumnWidth(nWidth)
RETURN ::XbpListBox:setColumnWidth(nWidth)
   
//METHOD ComboBoxSLE:setData(anItems,lMarked)   
//RETURN ::XbpListBox:setData(anItems,lMarked)   

METHOD ComboBoxSLE:setItem(nItem,cItem)  
RETURN ::XbpListBox:setItem(nItem,cItem)   

METHOD ComboBoxSLE:setItemsHeight(nPix)
RETURN ::XbpListBox:setItemsHeight(nPix)
   
METHOD ComboBoxSLE:setTabstops(anTabstops)   
RETURN ::XbpListBox:setTabstops(anTabstops)
   
METHOD ComboBoxSLE:setTopItem(nItem)
RETURN ::XbpListBox:setTopItem(nItem)  

METHOD ComboBoxSLE:listboxSize()  
RETURN ::XbpListBox:CurrentSize()  

METHOD ComboBoxSLE:sleSize()  
RETURN ::SLEPic:CurrentSize()  


METHOD ComboBoxSLE:Keyboard(nKey)
LOCAL lFound := .F.
   IF !::IsDisabled
      DO CASE
         CASE nKey = ::ListActivateKey  
             Eval(::oPush:Activate)
         CASE nKey = ::ScrollUpKey  
             IF EMPTY(::XbpListBox:Getdata())
                  FOR I = 1 TO ::XbpListBox:NumItems()
                     IF ::XbpListBox:GetItem(I)== ALLTRIM(::SLEPic:EditBuffer())
                        ::XbpListBox:SetData({I})
                        lFound := .T.
                        EXIT
                     ENDIF
                  NEXT I
                  IF !lFound
                     ::XbpListBox:SetData({1})
                     ::SLEPic:Setdata(::XbpListBox:GetItem(1))
                  ENDIF
             ENDIF
             IF ::XbpListBox:Getdata()[1] = 1
                ::SLEPic:SetData(::XbpListBox:GetItem(1))
             ELSE 
                ::SLEPic:SetData(::XbpListBox:GetItem(::XbpListBox:Getdata()[1]-1))  
                ::XbpListBox:Setdata(::XbpListBox:Getdata()[1]-1)  
             ENDIF
         CASE nKey = ::ScrollDownKey  
             IF EMPTY(::XbpListBox:Getdata())
                  FOR I = 1 TO ::XbpListBox:NumItems()
                     IF ::XbpListBox:GetItem(I)== ALLTRIM(::SLEPic:EditBuffer())
                        ::XbpListBox:SetData({I})
                        lFound := .T.
                        EXIT
                     ENDIF
                  NEXT I
                  IF !lFound
                     ::XbpListBox:SetData({1})
                     ::SLEPic:Setdata(::XbpListBox:GetItem(1))
                  ENDIF
             ENDIF
             IF ::XbpListBox:Getdata()[1] = ::XbpListBox:NumItems()
                ::SLEPic:SetData(::XbpListBox:GetItem(::XbpListBox:NumItems()))
             ELSE 
                ::SLEPic:SetData(::XbpListBox:GetItem(::XbpListBox:Getdata()[1]+1))
                ::XbpListBox:Setdata(::XbpListBox:Getdata()[1]+1)  
             ENDIF
         OTHERWISE
            ::SLEPic:Keyboard(nKey)      
      ENDCASE   
   ENDIF
RETURN Self

METHOD ComboBoxSLE:Destroy()
   ::XbpListBox:Destroy()
   ::oPush:Destroy()
   ::SLEPic:destroy()
RETURN Self