///////////////////////////////////////////////////////////////////////////////
//
//  LEDDISP.PRG
//
//  Contents  :
//    An old-fashioned LED display for numbers
//    See TEST.PRG or XCALC.PRG for usage scenarios
//
//  Remarks:
//    This software is published under the GNU General Public License (GPL)
//  
//      http://www.opensource.org/licenses/gpl-license.php
//  
//    NO WARRANTY
//  
//    BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
//    FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT
//    WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER
//    PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, 
//    EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE 
//    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
//    PURPOSE. THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE 
//    PROGRAM IS WITH YOU. SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME 
//    THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
//  
//    IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
//    WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR 
//    REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
//    DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
//    DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
//    (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
//    INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
//    OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
//    HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF 
//    SUCH DAMAGES.
//
//  Author and copyright
//    (c) Hannes Ziegler 2005
//
//        www.knowleXbase.com
//
///////////////////////////////////////////////////////////////////////////////


#include "Common.ch"
#include "Xbp.ch"
#include "Leddisp.ch"


CLASS LedDisplay FROM XbpStatic, DataRef
   PROTECTED:
   CLASS VAR aDigIDs             // Bitmaps IDs for digits
   CLASS VAR aDigChars           // Character representing a bitmap
   CLASS VAR aDigBmp             // Xbpbitmap objects for digits
   CLASS VAR aExpIDs             // Same as above for the
   CLASS VAR aExpChars           // smaller bitmaps displaying exponents
   CLASS VAR aExpBmp
   CLASS VAR oDecBmp             // The decimal point bitmap

   VAR aDigLeds                  // Xbpbitmap objects displaying the
   VAR aExpLeds                  // current value of :nNumber

   VAR nNumber                   // The currently displayed value
   VAR nExponent                 // The exponent for the display
   VAR nCount                    // Number of digits for :nNumber
   VAR cBuffer                   // :nNumber formatted as string
   VAR nDecPos                   // Position of the decimal point

   METHOD formatBuffer           // formats :nNumber as string
   METHOD updateBitmaps          // translates the string into bitmap representation

   EXPORTED:
   CLASS METHOD initClass

   VAR presSpace     READONLY
                                 ** Creation flags
   VAR digits                    // Number of digits to display
   VAR minus                     // Display minus sign .T.|.F.
   VAR withDecPoint              // Display decimal point .T.|.F.
   VAR withExponent              // Display exponent .T.|.F.
   VAR padWithZeros              // Fill number with zeros to the left .T.|.F.

   METHOD INIT, create, paint    // Overloaded from XbpStatic

   METHOD setData                // Overloaded from DataRef
   METHOD getData
   METHOD editBuffer

   METHOD error                  // Displays "Error"
ENDCLASS



/*
 * Prepare the bitmaps for all digits.
 * The XbpBitmaps objects are used by all instances of the class
 */
CLASS METHOD LedDisplay:initClass()
   LOCAL i, imax

   ::aDigIDs  := { BMP_LED_BLANK, ;
                   BMP_LED_NEG  , ;
                   BMP_LED_0    , ;
                   BMP_LED_1    , ;
                   BMP_LED_2    , ;
                   BMP_LED_3    , ;
                   BMP_LED_4    , ;
                   BMP_LED_5    , ;
                   BMP_LED_6    , ;
                   BMP_LED_7    , ;
                   BMP_LED_8    , ;
                   BMP_LED_9    , ;
                   BMP_LED_A    , ;
                   BMP_LED_B    , ;
                   BMP_LED_C    , ;
                   BMP_LED_D    , ;
                   BMP_LED_E    , ;
                   BMP_LED_F    , ;
                   BMP_LED_O    , ;
                   BMP_LED_R      }

   ::aDigChars := { " " , ;
                    "-" , ;
                    "0" , ;
                    "1" , ;
                    "2" , ;
                    "3" , ;
                    "4" , ;
                    "5" , ;
                    "6" , ;
                    "7" , ;
                    "8" , ;
                    "9" , ;
                    "A" , ;
                    "B" , ;
                    "C" , ;
                    "D" , ;
                    "E" , ;
                    "F" , ;
                    "O" , ;
                    "R"   }

   imax      := Len( ::aDigIDs )
   ::aDigBmp := Array( imax )

   FOR i:=1 TO imax
      ::aDigBmp[i] := XbpBitmap():new():create( )
      ::aDigBmp[i]:load( , ::aDigIDs[i] )
   NEXT

   ::aExpIDs  := { BMP_LED_SBLANK, ;
                   BMP_LED_SNEG  , ;
                   BMP_LED_S0    , ;
                   BMP_LED_S1    , ;
                   BMP_LED_S2    , ;
                   BMP_LED_S3    , ;
                   BMP_LED_S4    , ;
                   BMP_LED_S5    , ;
                   BMP_LED_S6    , ;
                   BMP_LED_S7    , ;
                   BMP_LED_S8    , ;
                   BMP_LED_S9      }

   ::aExpChars := { " " , ;
                    "-" , ;
                    "0" , ;
                    "1" , ;
                    "2" , ;
                    "3" , ;
                    "4" , ;
                    "5" , ;
                    "6" , ;
                    "7" , ;
                    "8" , ;
                    "9"   }

   imax      := Len( ::aExpIDs )
   ::aExpBmp := Array( imax )

   FOR i:=1 TO imax
      ::aExpBmp[i] := XbpBitmap():new():create()
      ::aExpBmp[i]:load( , ::aExpIDs[i] )
   NEXT

   ::oDecBmp := XbpBitmap():new():create()
   ::oDecBmp:load( , BMP_LED_DOT )

RETURN SELF


METHOD LedDisplay:INIT( oParent, oOwner, aPos, aSize, nDigits )

   IF Valtype( aSize ) == "A"
      DEFAULT nDigits TO Int( aSize[1] / 24 )
   ELSE
      DEFAULT nDigits TO 10
   ENDIF

   ::xbpStatic:INIT( oParent, oParent, aPos, aSize )
   ::presSpace    := XbpPresSpace():new()

   ::digits       := nDigits
   ::nDecPos      := 0
   ::nCount       := 0
   ::nNumber      := 0
   ::nExponent    := 0
   ::cBuffer      := "0"

   ::withExponent := .F.
   ::minus        := .F.
   ::withDecPoint := .T.
   ::padWithZeros := .F.
RETURN SELF



METHOD LedDisplay:create( oParent, oOwner, aPos, aSize, nDigits )
   LOCAL i, imax

   IF Valtype( nDigits ) == "N"
      ::digits := nDigits
   ENDIF

   ::xbpStatic:type    := XBPSTATIC_TYPE_RAISEDBOX
   ::xbpStatic:caption := ""
   ::xbpStatic:create( oParent, oParent, aPos, aSize )
   ::presSpace:create( ::winDevice() )

   IF ::withExponent
      ::digits   -= 2                    // Exponent bitmaps use 2 digits
      ::aExpLeds := Array( 3 )           // minus sign plus 2 digits
      AFill( ::aExpLeds, ::aExpBmp[1] )  // Fill with blank bitmap
   ENDIF

   ::aDigLeds := Array( ::digits )

   IF ::padWithZeros
      AFill( ::aDigLeds, ::aDigBmp[3] )  // Fill with Zero bitmap
   ELSE
      AFill( ::aDigLeds, ::aDigBmp[1] )  // Fill with blank bitmap
   ENDIF

RETURN SELF


/*
 * Overloaded from DataRef
 */
METHOD LedDisplay:editBuffer
RETURN ::nNumber


/*
 * Set current value
 */
METHOD LedDisplay:setData( nNumber, cBuffer )

   DEFAULT nNumber TO ::dataRef:setData()

   ::nNumber := ::dataRef:setData( nNumber )

   IF cBuffer <> NIL
      ::cBuffer := cBuffer
   ENDIF

   ::updateBitmaps()
RETURN ::nNumber



/*
 * Let DataRef evaluate the :datalink code block
 */
METHOD LedDisplay:getData
   ::dataRef:getData()
RETURN ::nNumber



/*
 * Display the LED bitmaps
 * Note: Bitmaps are not scaled but displayed in original size
 */
METHOD LedDisplay:paint
   LOCAL aStart := { Int((::currentSize()[1]-::digits*::aDigBmp[1]:xSize)/2), 1 }
   LOCAL aPos

   aStart[1] += (::digits-1)*::aDigBmp[1]:xSize
   aPos      := AClone( aStart )

   IF ::withExponent
      aPos[1] += 2 * ::aExpLeds[1]:xSize
      AEval( ::aExpLeds, {|o| o:draw( ::presSpace, aPos ), aPos[1] -= o:xSize } )
      aPos[1] -= (::aDigBmp[1]:xSize - ::aExpBmp[1]:xSize )
   ENDIF

   AEval( ::aDigLeds, {|o| o:draw( ::presSpace, aPos ), aPos[1] -= o:xSize } )

   aPos := aStart
   IF ::nDecPos > 0
      aPos[1] -= (::nCount-::nDecPos+1) * ::aDigBmp[1]:xSize
   ENDIF

   IF ::withDecPoint
      aPos[1] -= ::oDecBmp:xSize
      ::oDecBmp:draw( ::presSpace, aPos )
   ENDIF
RETURN


/*
 * Get the bitmap objects that reflect the current buffer
 */
METHOD LedDisplay:updateBitmaps
   LOCAL cChr, nPos, i, imax

   ::formatBuffer()

   IF ::padWithZeros
      AFill( ::aDigLeds, ::aDigBmp[3] ) // fill with Zero betmap
   ELSE
      AFill( ::aDigLeds, ::aDigBmp[1] ) // fill with Blank bitmap
   ENDIF

   IF ::withExponent
      AFill( ::aExpLeds, ::aExpBmp[1] )

      IF ::nExponent <> 0
         IF ::nExponent < 0
            cChr := "-" + PadL( Abs(::nExponent), 2, "0" )
         ELSE
            cChr := PadL( ::nExponent, 3, "0" )
         ENDIF

         ::aExpLeds[3] := getBitmap( ::aExpBmp, ::aExpChars, cChr[1] )
         ::aExpLeds[2] := getBitmap( ::aExpBmp, ::aExpChars, cChr[2] )
         ::aExpLeds[1] := getBitmap( ::aExpBmp, ::aExpChars, cChr[3] )
      ENDIF
   ENDIF

   ::nCount  := imax := Len( StrTran( ::cBuffer, ".", "" ) )
   ::nDecPos := At( ".", ::cBuffer )

   IF ::digits < ::nCount
      imax     := ;
      ::nCount := ::digits
   ENDIF

   IF ::nDecPos > 0
      imax ++
   ENDIF

   nPos := 1
   FOR i:=1 TO imax
      cChr := ::cBuffer[i]

      IF cChr $  ",."
         LOOP
      ELSE
        ::aDigLeds[::nCount-nPos+1] := getBitmap( ::aDigBmp, ::aDigChars, cChr )
        nPos ++
      ENDIF
   NEXT

   ::paint()
RETURN SELF


/*
 * Format a floating point number with exponent
 */
METHOD LedDisplay:formatBuffer
   LOCAL nPos, nLog10, nNumber, cBuffer, cPre, cPost, bBlock, i

   IF ::nNumber <> 0
      IF ::withExponent .OR. ::withDecPoint
         nLog10      := Log( Abs(::nNumber) ) / Log(10)
         ::nExponent := Int( nLog10 )

         IF ::nExponent < -99 .OR. ::nExponent > 300
            // Exponent overflow ** RETURN **
            RETURN ::error()
         ENDIF

         nNumber   := Abs(::nNumber) / 10 ^ ::nExponent
         cPre      := LTrim( Str( Int( nNumber ) ) )
         cPost     := LTrim( Str( nNumber-Int( nNumber ), 2*::digits, ::digits+2 ) )
         RemoveTrailingZeros( @cPost )
         cPost     := SubStr( cPost, 3 )

         cBuffer   := cPre + "." + cPost
         IF ::nNumber < 0
            cBuffer := "-" + cBuffer
            ::minus := .T.
         ENDIF

         // adjust exponent for numbers that don't fit into the display
         bBlock := {|| Abs(::nExponent) > 0 .AND. Len( cBuffer ) < ::digits }
         FOR i:=1 TO 2
            DO WHILE ::eval( bBlock )
               nPos := At( ".", cBuffer )
               IF nPos == 0
                  cBuffer += "0"
               ELSEIF Abs(::nNumber) < 1
                  cBuffer := Stuff( cBuffer, nPos+1, 0 , "0" )
               ELSEIF nPos == Len( cBuffer )
                  cBuffer := Stuff( cBuffer, nPos, 0 , "0" )
               ELSE
                  cBuffer := Stuff( StrTran(cBuffer,".",""), nPos+1, 0, "." )
               ENDIF

               IF ::nExponent > 0
                  ::nExponent --
               ELSE
                  ::nExponent ++
               ENDIF
            ENDDO
            bBlock := {|| Abs(::nExponent)> 0 .AND. Abs(::nExponent) < ::digits-1 }
         NEXT

         IF Right(cBuffer,1) == "."
            cBuffer := SubStr( cBuffer, 1, Len(cBuffer)-1 )
         ENDIF
      ELSE
         // Only integers are displayed
         cBuffer := LTrim(Str(::nNumber))
      ENDIF

      ::cBuffer := cBuffer

   ELSE
      ::nExponent := 0
      IF Val( ::cBuffer ) <> 0 .OR. ::cBuffer == ""
         IF ::minus
            ::cBuffer := "-0"
         ELSE
            ::cBuffer := "0"
         ENDIF
      ENDIF
   ENDIF
RETURN SELF


METHOD LedDisplay:error
   LOCAL cString, i

   AFill( ::aDigLeds, ::aDigBmp[1] )

   IF ::withExponent
      AFill( ::aExpLeds, ::aExpBmp[1] )
   ENDIF

   ::nNumber   := 0
   ::nDecPos   := 0
   ::nExponent := 0
   ::nCount    := Min(::Digits, 5+Int((::Digits-5)/2) )
   ::cBuffer   := "0"
   ::minus     := .F.
   cString     := PadR( "ERROR", ::nCount )

   FOR i:=1 TO ::nCount
      ::aDigLeds[::nCount-i+1] := getBitmap( ::aDigBmp, ::aDigChars, cString[i] )
   NEXT
   ::paint()
RETURN


STATIC FUNCTION GetBitmap( aBitmaps, aChars, cChar )
   LOCAL i := AScan( aChars, cChar )
RETURN IIF( i==0, NIL, aBitmaps[i] )


STATIC FUNCTION RemoveTrailingZeros( cNumber )
   LOCAL i := Len( cNumber ) + 1

   DO WHILE --i > 0 .AND. cNumber[i] == "0"
   ENDDO
   IF cNumber[i] == "."
      i--
   ENDIF
   cNumber := SubStr( cNumber, 1, i)
RETURN cNumber