#pragma Library( "XppUI2.LIB" )

#include "Appevent.ch"
#include "Common.ch"
#include "Xbp.ch"
#include "Font.ch"

PROCEDURE EditDBF( name,cTitle )
   LOCAL nEvent, mp1, mp2, oXbp, oBrowse, aStruct, i, imax, bError, cDbe
   LOCAL aPP, oColumn,oDlg

   DEFAULT name TO GetFilename()
   DEFAULT cTitle TO "Edit File"

   IF Empty( name )
      MsgBox ( "No Filename specified" + CHR(10) + CHR(10)+ ;
               "Usage: Editbrow [<DbfName>]" )
      RETURN
   ENDIF

   bError := ErrorBlock( {|oErr| BREAK( oErr ) } )
   cDBE   := "DBFDBE"
   FOR i:=1 TO 2
      BEGIN SEQUENCE
         DbUseArea( .T., cDBE, name )
      RECOVER

         IF i== 1
            cDBE := "FOXDBE"
            LOOP
         ENDIF
      ENDSEQUENCE
      EXIT
   NEXT
   ErrorBlock( bError )


   IF ! Used()
      MsgBox( "Unable to open file" + CHR(10) + CHR(10)+ Name )
      RETURN
   ENDIF

   // create std dialog
   oDlg := GuiStdDialog( cTitle )
   oDlg:Close:={|| oDlg:Destroy()}

   // create browser in window
   oBrowse := GuiBrowseDb( oDlg:drawingArea )

   // create columns for all fields but exclude FoxPro specific data types
   imax    := FCount()
   aStruct := DbStruct()
   aPP     := {{ XBP_PP_COL_HA_CAPTION, "" },{XBP_PP_COL_DA_CHARWIDTH,0}}

   FOR i:=1 TO imax
      IF .NOT. aStruct[i,2] $ "BOVTXY"
         aPP[1,2] := aStruct[i,1] 
         aPP[2,2] := aStruct[i,3]
         IF aStruct[i,2] = "M"
            oColumn  := XbpMleColumn():new( oBrowse,,,, aPP )
         ELSE
            oColumn  := SLEPicColumn():new( oBrowse,,,, aPP )
         ENDIF   
         oColumn:dataLink := EditBlock( aStruct[i,1] )
         oBrowse:addColumn( oColumn )
      ENDIF
   NEXT

   // Overload resize so that the browser fills the window
   oDlg:drawingArea:resize := ;
       {|mp1,mp2,obj| obj:childList()[1]:setSize(mp2) }

   // Delegate keys for editing to the XbpGetColumn object
   oBrowse:keyboard := ;
      {|nKey, x, obj| IIf( nKey > 31 .OR. nKey < 256, ;
                           obj:getColumn( obj:colPos ):keyboard( nKey ), ;
                           NIL ) }

   // show browser and set focus on it
   oDlg:show()
   oBrowse:show()
   SetAppFocus( oBrowse )

   ModalLoop(oDlg)
RETURN


//////////////////////////////////////////////////////////////////////
//
//  GETCOL.PRG
//
//  Copyright:
//      Alaska Software, (c) 1999-2003. All rights reserved.
//      modified by James Loughner for use of  SLEPic class
//      01/20/04 Ver 1.0
//      01/23/04 Ver 1.1 fix problem when clicking to a new cell
//               and valid in curret cell fails (ie returns .F.)
//
//  Contents:
//      This example implements the SLEPicColumn class which uses
//      SLEPic objects for editing of single cells in a XbpBrowse object.
//
//  Remarks:
//
//      The following programming techniques are demonstrated:
//
//      1. How to embed an editable Xbase Part into a column object so that
//         a browser based on the XbpBrowse() class becomes editable.
//
//      2. How to move the edit control upon browser request.
//
//      3. How to use SLEPic objects as edit controls in a browser column.
//
//      Navigation whilw in edit mode
//      The up/down arrow and enter move between cell is a row
//      the Shift-UP/down arrow move between rows
//
//////////////////////////////////////////////////////////////////////



CLASS SLEPicColumn FROM XbpColumn
   PROTECTED:
//   CLASS VAR activeColumn //SHARED // The column that is currently being edited
//   CLASS VAR LastRow      //SHARED
   METHOD createXbp              // Create it upon first edit request
   METHOD setPosAndSize          // Assign pos and size for editing

   EXPORTED:

   VAR    SLEPic                 // The associated SLEPic object
   VAR    Picture                // Picture for thi column
                                 ** Overloaded methods
   METHOD init                   // Initialize object
   METHOD create                 // Request system resources
   METHOD destroy                // Release System resources
   METHOD resize                 // Resize must be reflected by SLEPic during edit
   METHOD hiliteRow              // Display the SLEPic during browsing

                                 ** This is required for editing
   VAR    active  READONLY       // Tells if we are in edit mode
   VAR    browser                // XbpBrowse object the column is contained in
   VAR    bufferLength           // Max. number of characters for editing
   VAR    lNoEdit 
   VAR    Validate
   VAR    bUpdate                // array of code blocks that update somthing on entry of SLE
   VAR    onSLEFocus             // codeblock on focus
   VAR    SLEKeyboard
   
   METHOD _LbDown                // LbDown method for :dataArea object
   METHOD READ                   // Begin editing
   METHOD keyBoard               // Called by an SLEPic object for UP / DOWN keys
   METHOD killRead               // End editing
ENDCLASS


/*
 * Initialize the object
 */
METHOD SLEPicColumn:init( oParent, oOwner, aPos, aSize, aPP, lVisible )
   ::xbpColumn:init( oParent, oOwner, aPos, aSize, aPP, lVisible )
   ::active  := .F.
   ::Picture := ""
   ::lNoEdit := .F. 
   IF oParent <> NIL .AND. oParent:isDerivedFrom( "XbpBrowse" )
      ::browser := oParent
   ENDIF   
   ::browser:Cargo := {NIL,::Browser:RowPos}
RETURN self


/*
 * Request system resources
 */
METHOD SLEPicColumn:create( oParent, oOwner, aPos, aSize, aPP, lVisible )
   LOCAL oError

   ::xbpColumn:create( oParent, oOwner, aPos, aSize, aPP, lVisible )

   /*
    * Set focus back to the browser if the column object gets input focus.
    * This assures that keyboard events are always processed by the browser.
    */
   ::dataArea:lbDown := {|aPos| ::_lbDown( aPos ) }
   ::setInputFocus   := {|x,y,obj| SetAppFocus( obj:browser ) }

   IF ::browser == NIL
      /*
       * Find the XbpBrowse object in the Parent-Child hierarchy
       *
       * NOTE: The DO WHILE loop demonstrates how to search the parent-child
       *       hierarchy upwards. The top of the Parent-Child tree is reached
       *       when the Desktop window is the parent (oParent == AppDesktop())
       */
      oParent := ::setParent()

      DO WHILE .NOT. oParent:isDerivedFrom( "XbpBrowse" )
         oParent := oParent:setParent()
         IF oParent == AppDesktop()
            oError             := Error():new()
            oError:DESCRIPTION := "Illegal parent defined for " + ::className()
            oError:operation   := ProcName()
            oError:args        := { ::setParent() }
            Eval( ErrorBlock(), oError )
         ENDIF
      ENDDO

      /*
       * An XbpBrowse object is found in the Parent-Child hierarchy
       */
      ::browser := oParent
   ENDIF
RETURN self



/*
 * Clean up memory and set object references to NIL
 */
METHOD SLEPicColumn:destroy()
   IF ::SLEPic <> NIL 
      IF ::SLEPic:Status() = XBP_STAT_CREATE
         ::SLEPic:destroy()
      ENDIF   
      ::SLEPic := NIL
   ENDIF
   ::active  := .F.
   ::browser := NIL
   ::XbpColumn:destroy()
RETURN self

/*
 * Create the SLEPic object and share the :dataLink code block
 */
METHOD SLEPicColumn:createXbp
LOCAL Temp
   ::SLEPic            := SLEPic():new( ::dataArea, ::dataArea, NIL, NIL, NIL, .F. )
   ::SLEPic:dataLink   := ::dataLink

   Temp := EVAL(::dataLink,NIL)   // get data sample
   DO CASE
      CASE VALTYPE(Temp)="L"
         ::SLEPic:bufferLength := 1 //::bufferLength
      CASE VALTYPE(Temp)="N"
         ::SLEPic:bufferLength := LEN(STR(Temp)) //::bufferLength
      CASE VALTYPE(Temp)="C"
         ::SLEPic:bufferLength := LEN(Temp) //::bufferLength
         IF EMPTY(::Picture)
            ::Picture := REPLICATE("X",LEN(Temp))
         ENDIF
      CASE VALTYPE(Temp)="D"
         ::SLEPic:bufferLength := LEN(DTOC(Temp)) //::bufferLength
   ENDCASE
   ::SLEPic:SetInputFocus  := ::onSLEFocus
   ::SLEPic:Picture  := ::Picture
   ::SLEPic:Validate := ::Validate
   ::SLEPic:KeyBoard := ::SLEKeyBoard
   ::SLEPIC:KillInputFocus := {||IF(::SLEPIC:IsValid,::KillRead(),NIL)}
   ::SLEPic:create()
RETURN


/*
 * Determine the X/Y coordinates for the SLEPic display.
 */
METHOD SLEPicColumn:setPosAndSize( nRowPos )
   LOCAL aRect, aPos, aSize

   aRect := ::dataArea:cellRect( nRowPos )
   aPos  := { aRect[1], aRect[2] }
   aSize := { aRect[3]-aRect[1], aRect[4]-aRect[2] }
   ::SLEPic:setPos ( aPos  )
   ::SLEPic:setSize( aSize )
RETURN self


/*
 * When the column is resized while Edit mode is active,
 * the edit control must be adjusted to the column width.
 */
METHOD SLEPicColumn:resize( aOldSize, aNewSize )

   IF .NOT. ::active
      RETURN ::xbpColumn:resize( aOldSize, aNewSize )
   ENDIF

   ::SLEPic:hide()
   ::xbpColumn:resize( aOldSize, aNewSize )
   ::setPosAndSize( ::browser:rowPos )
   ::SLEPic:show()

   SetAppFocus( ::SLEPic )
RETURN self


/*
 * This method is called from the XbpBrowse object when the browse cursor
 * is moved. It must be overloaded for editable column objects and is
 * used to position the edit control at the browse cursor.
 */
METHOD SLEPicColumn:hiliteRow( nRowPos, lHilite, lFrame, lRepaint )
   IF ::active
      IF lHilite .AND. !::lNoEdit
         /*
          * We are in Edit mode and the browser wants to display its
          * cursor in highlight color. Resume with editing.
          */
         ::READ( nRowPos )
      ELSE
         /*
          * We are in Edit mode and the browser wants to display its cursor
          * in normal color. Eval the :dataLink code block upon changes
          * and hide the SLEPic.
          */
         IF ::SLEPic <> NIL
             IF ::SLEPic:Ischanged()
                ::SLEPic:getData()
                ::xbpColumn:refreshRows( nRowPos, nRowPos )
             ENDIF
         ENDIF
         ::xbpColumn:hiliteRow( nRowPos, lHilite, lFrame, lRepaint )
      ENDIF
   ELSE
      /*
       * We are in non-edit mode. Pass everything on to the super class.
       */
      ::xbpColumn:hiliteRow( nRowPos, lHilite, lFrame, lRepaint )
   ENDIF
RETURN self



/*
 * Create and/or display the SLEPic object and edit the current row
 */
METHOD SLEPicColumn:READ( nRowPos, nFirstKey )
LOCAL nNewPos,nOldPos,I
   DEFAULT nRowPos TO ::browser:rowPos
   IF ::SLEPic == NIL 
      /*
       * The column has not been edited yet. Create the SLEPic object,
       * upon first editing.
       */
      ::createXbp()
   ENDIF
   IF ::browser:Cargo[1] <> NIL  .AND. ::browser:Cargo[1]:SLEPic <> NIL .AND.;
      ::browser:Cargo[1] <> self
      /*
       * The browser is clicked with the mouse while another column
       * is currently being edited. Terminate edit in that column.
       */
      IF ::browser:Cargo[1]:SLEPic:IsValid
         ::browser:Cargo[1]:killRead()
         ::Browser:RefreshAll()
      ELSE   
         ::browser:Cargo[1]:SLEPic:Show()
         SetAppFocus(::browser:Cargo[1]:SLEPic)
      ENDIF
   ENDIF

   IF ::browser:Cargo[1] = NIL 
      ::browser:Cargo[1] := self
      ::browser:Cargo[2] := ::Browser:RowPos
   ENDIF
   IF ::browser:Cargo[1]:SLEPic <> NIL .AND. ::browser:Cargo[1]:SLEPic:IsValid   
      ::browser:Cargo[1] := self
      ::browser:Cargo[2] := ::Browser:RowPos
   ENDIF


   ::active  := (::browser:Cargo[1] = self)
   ::setPosAndSize( nRowPos )

   ::SLEPic:setData()
   ::SLEPic:changed := .F.
   IF ::browser:Cargo[1]:SLEPic <> NIL .AND.!::browser:Cargo[1]:SLEPic:IsValid
      ::SLEPic:hide()
   ELSE
      ::SLEPic:show()
   ENDIF

   IF ::SLEPic:IsVisible()  // <> NIL
      SetAppFocus( ::SLEPic )
   ENDIF   
   IF ::bUpdate <> NIL
      FOR I = 1 TO LEN(::bUpdate)
         eval(::bUpdate[I])
      NEXT I
   ENDIF
   ::SLEPic:SetInputFocus  := ::onSLEFocus
   IF IsMemberVar(::SLEPic,"Picture")
      ::SLEPic:Picture  := ::Picture
      ::SLEPic:Validate := ::Validate
   ENDIF
   ::SLEPic:KeyBoard := ::SLEKeyBoard
   IF ::lNoEdit 
      ::KillREAD()
   ENDIF
   
   IF nFirstKey <> NIL .AND. SetAppFocus() = ::SLEPic
      PostAppEvent( xbeP_Keyboard, nFirstKey, NIL, ::SLEPic )
   ENDIF 
RETURN self
 

/*
 * Left button is pressed in the :dataArea of the column
 */
METHOD SLEPicColumn:_lbDown( aPos )
   LOCAL nRowPos := ::dataArea:cellFromPos( aPos )

   ::active := .T.

   IF nRowPos == ::browser:rowPos
      ::READ( nRowPos )
   ENDIF  
RETURN self



/*
 * Keyboard handling method
 *
 * When edit mode is active, the SLEPic object has input focus and
 * calls this method since the XbpColumn object is the controller
 * of the SLEPic object
 * (see method :createXbp() -> ::SLEPic:controller := self)
 *
 * - The edit mode is activated by
 *   1) the ENTER key (cell content remains)
 *   2) an editing key (the cell is cleared)
 *
 * - Esc terminates editing
 *
 */
METHOD SLEPicColumn:keyBoard( nKey )
LOCAL aRowPos, bBlock
   DO CASE
   CASE nKey == xbeK_ALT_F4
      PostAppEvent( xbeP_Close,,, SetAppWindow() )
      RETURN .T.

   CASE nKey == xbeK_ENTER
      /*
       * Edit next cell or activate edit mode.
       */
      IF ::active
         bBlock := {|o| o:down() }
      ELSE
         ::READ( ::browser:rowPos )
      ENDIF

   CASE .NOT. ::active
      /*
       * Keys processed by SLEPic activate edit mode and clear the
       * edit buffer. Other keys are ignored.
       */
      IF nKey > 31 .AND. nKey < 256
         ::READ( ::browser:rowPos, nKey )
      ENDIF

   CASE nKey == xbeK_ESC
      /*
       * Undo changes upon ESC
       */
      ::SLEPic:undo()
      ::killRead()
      bBlock := {|o| o:refreshCurrent() }

   /*
    * Keys causing browse cursor movement during editing
    *
    * NOTE: Tab keys cause xbeK_ENTER to be posted to the browser so that
    *       the Edit mode is switched on in the next or previous column.
    */
   CASE nkey == xbeK_TAB
      ::SLEPic:KillInputFocus()
      IF ::SLEPic:IsValid
        IF ::browser:colPos == ::browser:colCount
           bBlock := {|o| o:firstCol(), o:down(), ;
                          PostAppEvent( xbeP_Keyboard, xbeK_ENTER,, o ) }
        ELSE
           bBlock := {|o| o:right(), ;
                          PostAppEvent( xbeP_Keyboard, xbeK_ENTER,, o ) }
        ENDIF
      ENDIF
      ::SLEPic:SetInputFocus()
   CASE nkey == xbeK_SH_TAB
      ::SLEPic:KillInputFocus()
      IF ::SLEPic:IsValid
        IF ::browser:colPos == 1
           bBlock := {|o| o:up(), o:lastCol(), ;
                          PostAppEvent( xbeP_Keyboard, xbeK_ENTER,, o ) }
        ELSE
           bBlock := {|o| o:left(), ;
                          PostAppEvent( xbeP_Keyboard, xbeK_ENTER,, o ) }
        ENDIF
      ENDIF
      ::SLEPic:SetInputFocus()

   CASE nkey == xbeK_SH_DOWN
      ::SLEPic:KillInputFocus() // check valid
      IF ::SLEPic:IsValid
         bBlock := {|o| o:down() }
      ENDIF
      ::SLEPic:SetInputFocus()

   CASE nKey == xbeK_SH_UP
      ::SLEPic:KillInputFocus() // check valid
      IF ::SLEPic:IsValid
         bBlock := {|o| o:up() }
      ENDIF
      ::SLEPic:SetInputFocus()

   CASE nkey == xbeK_SH_RIGHT
      bBlock := {|o| o:right() }

   CASE nKey == xbeK_SH_LEFT
      bBlock := {|o| o:left() }

   CASE nkey == xbeK_PGDN
      bBlock := {|o| o:pageDown() }

   CASE nKey == xbeK_PGUP
      bBlock := {|o| o:pageUp() }

   CASE nkey == xbeK_CTRL_PGDN
      bBlock := {|o| o:goBottom() }

   CASE nKey == xbeK_CTRL_PGUP
      bBlock := {|o| o:goTop() }

   ENDCASE

   IF bBlock <> NIL
      /*
       * Invoke browser action
       */
      Eval( bBlock, ::browser )
      ::browser:forceStable()
   ENDIF

   /*
    * SLEPic needs a logical return value
    */
RETURN ( bBlock <> NIL )


/*
 * Terminate the Edit mode
 */
METHOD SLEPicColumn:killRead
   IF ::active
      ::SLEPic:hide()
      ::active       := .F.
      ::browser:Cargo[1] := NIL
      SetAppFocus( ::browser )
   ENDIF
RETURN self


//////////////////////////////////////////////////////////////////////
//
//  MLECOL.PRG
//
//  Copyright:
//      Alaska Software, (c) 1999-2003. All rights reserved.         
//  
//  Contents:
//      This example implements the XbpMleColumn class that can be used
//      for display and editing of memo fields in an XbpBrowse object.
//   
//      The Mle is provided by the MemoEditDialog class, which is also
//      implemented in this file.
//   
//  Remarks:
//     
//      The following programming techniques are demonstrated:
//     
//      1. How to subclass from the XbpGetColumn class in order to use
//         another Xbase Part than XbpGet for editing in the browser.
//   
//      2. How to exchange the :dataLink code block "secretly" so that
//         one code block is used for data display with XbpColumn and the
//         other is used for editing with XbpMle.
//   
//////////////////////////////////////////////////////////////////////



CLASS XbpMleColumn FROM SLEPicColumn
   PROTECTED:
   VAR    MemoDlg IS SLEPic IN SLEPicColumn
   VAR    _dataLink
   METHOD setPosAndSize

   EXPORTED:   
   METHOD create
   METHOD createXbp
   METHOD firstMemoLine
   METHOD hiliteRow 
   VAR MLESize
ENDCLASS


/*
 * Replace the :dataLink code block before the object is :create()ed.
 * The original :dataLink code block is preserved and used later for
 * XbpMLE. The new code block retrieves the first line of text from the
 * memo field.
 */
METHOD XbpMleColumn:create( p1, p2, p3, p4, p5, p6 )
   ::bufferLength := 40
   ::_dataLink    := ::dataLink
   ::dataLink     := {|| ::firstMemoLine() }
   ::SLEPicColumn:create( p1, p2, p3, p4, p5, p6 )
RETURN self


METHOD XbpMleColumn:firstMemoLine
RETURN MemoLine( Eval( ::_dataLink ), ::bufferLength, 1 )


/*
 * Create a MemoEditDialog instead of the XbpGet object and assign the
 * original :dataLink code block which accesses the memo field.
 *
 * Note 1: the instance variable :MemoDlg is actually declared as
 *         :xbpGet in the super class, but renamed in this class.
 *         This makes the code better readable.
 *
 * Note 2: The owner of the MemoEditDialog is the window that contains the
 *         browser. The parent of the MemoEditDialog is the same as for the
 *         browser window. If the browser window is moved, the MemoEditDialog
 *         window is moved as well.
 */
METHOD XbpMleColumn:createXbp
   LOCAL oOwner := ::setParent()
   LOCAL oParent, aSize := { 220, 180 }
   
   IF ValType(::MLESize) = "A" // array is set
      aSize := ::MLESize
   ENDIF
   DO WHILE .NOT. ( oOwner:IsDerivedFrom( "XbpDialog" ) .OR. oOwner:IsDerivedFrom( "XbpCrt" ) )
      oOwner := oOwner:setParent()
   ENDDO

   IF oOwner:isDerivedFrom( "XbpCrt" )
      oParent := AppDesktop()
   ELSE
      oParent := oOwner:setParent()
   ENDIF

   aSize[1]                := ::currentSize()[1]
   ::MemoDlg               := MemoEditDialog():new( oParent, oOwner,, aSize,, .F. )
   ::MemoDlg:controller    := self
   ::MemoDlg:dataLink      := ::_dataLink
   ::MemoDlg:moveWithOwner := .T.
   ::MemoDlg:title         := ::heading:getCell(1)
   ::MemoDlg:create()
RETURN self


/*
 * Overloaded from XbpGetColumn: only the position of the MemoEditDialog window
 * changes with the browse cursor, not the size
 *
 * Note: The position of the browse cursor cannot be directly used for
 *       positioning of the MemoEditDialog window. The coordinates must be
 *       transformed to the origin of the coordinate system that is
 *       valid for the browser window.
 */
METHOD XbpMleColumn:setPosAndSize( nRowPos )
   LOCAL aRect  := ::dataArea:cellRect( nRowPos )
   LOCAL aPos   := { aRect[1], aRect[2] + ( aRect[4]-aRect[2] ) }
   LOCAL aSize  := ::MemoDlg:currentSize()
   LOCAL oOwner := ::MemoDlg:setOwner()

   aPos[1] += ::currentPos()[1] + ::setParent():currentPos()[1]
   aPos[2] += ::footing:currentSize()[2]

   aPos[1] += ::browser:currentPos()[1]
   aPos[2] += ::browser:currentPos()[2]

   aPos[1] += oOwner:currentPos()[1]
   aPos[2] += oOwner:currentPos()[2]

   aPos[1] := Max( 0, Min( aPos[1], oOwner:setParent():currentSize()[1]-aSize[1] ) )
   aPos[2] := Max( 0, aPos[2]-aSize[2] )

   ::MemoDlg:setPos( aPos )
RETURN self


/*
 * Overloaded from XbpGetColumn: The MemoEditDialog window remains visible
 * during (de)hilighting.
 */
METHOD XbpMleColumn:hiliteRow( nRowPos, lHilite, lFrame, lRepaint )
   DEFAULT nRowPos TO ::browser:rowPos

   IF ::active
      IF lHilite
         ::read( nRowPos )
      ELSE
         IF ::MemoDlg:changed
            ::MemoDlg:getData()
            ::xbpColumn:refreshRows( nRowPos, nRowPos )
         ENDIF
      ENDIF
   ENDIF

   ::xbpColumn:hiliteRow( nRowPos, lHilite, lFrame, lRepaint )
RETURN self


******************************************************************************
/*
 * MemoEditDialog with embedded XbpMle and :dataLink code block
 */

CLASS MemoEditDialog FROM XbpDialog
   PROTECTED:
   VAR btnOk    
   VAR bntUndo

   EXPORTED:
   VAR    controller
   VAR    MLE        
   VAR    dataLink
   VAR    changed
   VAR    bufferLength 
   VAR    isValid
   ACCESS ASSIGN METHOD dataLink   
   ACCESS ASSIGN METHOD changed

   METHOD init
   METHOD create
   METHOD setData
   METHOD getData
   METHOD editBuffer
   METHOD clear
   METHOD undo
   METHOD setMarked
   METHOD keyboard
ENDCLASS


METHOD MemoEditDialog:init( oParent, oOwner, aPos, aSize, aPP, lVisible )
   LOCAL aMinSize := {218,160}

   DEFAULT oParent  TO AppDesktop(), ;
           aPos     TO {0,0}, ;
           aSize    TO {218,160}, ;
           lVisible TO .F.

   aSize[1] := Max( aSize[1], aMinSize[1] )
   aSize[2] := Max( aSize[2], aMinSize[2] )

   ::XbpDialog:init( oParent, oOwner, aPos, aSize, aPP, lVisible )
   ::XbpDialog:taskList := .F.
   ::XbpDialog:title    := "Memo"
   ::XbpDialog:minSize  := aMinSize
   ::XbpDialog:close    := {|| ::controller:killRead() }
   ::XbpDialog:setInputFocus := {|x,y,obj| SetAppFocus( IIf(obj:isVisible(),obj:MLE, obj:controller) ) }

   ::MLE                := XbpMle():new( ::drawingArea )
   ::MLE:ignoreTab      := .T.
   ::MLE:wordWrap       := .T.
   ::MLE:horizScroll    := .F.

   ::btnOk              := XbpPushButton():new( ::drawingArea, , {8,8}, {88,24} )
   ::btnOk:caption      := "~Ok"
   ::btnOk:activate     := {|| ::getData(), ::controller:killRead() }

   ::bntUndo            := XbpPushButton():new( ::drawingArea, , {112,8}, {88,24} )
   ::bntUndo:caption    := "~Undo"
   ::bntUndo:activate   := {|| ::undo()   , ::controller:killRead() }

   ::bufferLength       := 66
   ::IsValid := .T.
RETURN self


METHOD MemoEditDialog:create( oParent, oOwner, aPos, aSize, aPP, lVisible )
   ::XbpDialog:create( oParent, oOwner, aPos, aSize, aPP, lVisible )

   aPos     := {8,48}
   aSize    := ::drawingArea:currentSize()
   aSize[1] -= 16
   aSize[2] -= 56

   ::MLE:create(,, aPos, aSize )
   ::btnOk:create()
   ::bntUndo:create()

   ::drawingArea:resize := {|aOld, aNew| aNew[1] -= 16, ;
                                         aNew[2] -= 56, ;
                                         ::Mle:setSize( aNew ) }
RETURN self


/*
 * Methods for editing
 */
METHOD MemoEditDialog:setData( xValue )
   IF .NOT. ::isVisible()
      ::show()
   ENDIF

   IF PCount() == 0
      xValue := ::MLE:setData()
   ELSE
      xValue := ::MLE:setData( xValue )
   ENDIF
RETURN xValue


METHOD MemoEditDialog:getData
   LOCAL xRet := ::MLE:getData()

   ::changed := .F.
   ::controller:browser:refreshCurrent()
RETURN xRet


METHOD MemoEditDialog:editBuffer
RETURN ::MLE:editBuffer()


METHOD MemoEditDialog:clear
RETURN ::MLE:clear()


METHOD MemoEditDialog:undo
RETURN ::MLE:undo()


METHOD MemoEditDialog:setMarked( aMarked )
RETURN ::MLE:setMarked( aMarked )


METHOD MemoEditDialog:dataLink( bBlock )
   IF PCount() == 1
      ::MLE:dataLink := bBlock
   ENDIF
RETURN ::MLE:dataLink


METHOD MemoEditDialog:changed( lChanged )
   IF Valtype( lChanged ) == "L"
      ::MLE:changed( lChanged )
   ENDIF
RETURN ::MLE:changed


/*
 * xbeK_ALT_UP and xbeK_ALT_DOWN are used for browser navigation
 */
METHOD MemoEditDialog:keyboard( nKey )
   DO CASE
   CASE nKey == xbeK_ALT_F4
      ::controller:killRead()

   CASE nKey == xbeK_ALT_UP
      ::controller:keyboard( xbeK_UP )
      SetAppFocus( self )

   CASE nKey == xbeK_ALT_DOWN
      ::controller:keyboard( xbeK_DOWN )
      SetAppFocus( self )

   CASE nKey == xbeK_ALT_O
      ::getData()
      ::controller:killRead()

   CASE nKey == xbeK_ALT_U
      ::setData()
      ::controller:killRead()

   CASE .NOT. ::controller:keyboard( nKey )
      ::xbpDialog:keyboard( nKey )

   ENDCASE
RETURN self


// Browse functions

******************************************************************
* Create GUI Browser with navigation codeblocks
******************************************************************
FUNCTION GuiBrowseDB( oParent, aPos, aSize,DBArea,aPP )
   LOCAL oBrowse

   oBrowse := XbpBrowse():new( oParent,, aPos, aSize,aPP, .F. ):create()

   IF DBArea = NIL
      // navigation codeblocks for the browser
      oBrowse:skipBlock     := {|n| DbSkipper(n) }
      oBrowse:goTopBlock    := {| | DbGoTop()    }
      oBrowse:goBottomBlock := {| | DbGoBottom() }
      oBrowse:phyPosBlock   := {| | Recno()      }
   
      // Code blocks for the vertical scrollbar.
      // Note: DbPosition() returns values in the range from 0 to 100.
      oBrowse:posBlock      := {| | DbPosition()    }
      oBrowse:goPosBlock    := {|n| DbGoPosition(n) }
      oBrowse:lastPosBlock  := {| | 100             }
      oBrowse:firstPosBlock := {| | 0               }
   ELSE
      // navigation codeblocks for the browser
      oBrowse:skipBlock     := {|n| (DBArea)->(DbSkipper(n)) }
      oBrowse:goTopBlock    := {| | (DBArea)->(DbGoTop())    }
      oBrowse:goBottomBlock := {| | (DBArea)->(DbGoBottom()) }
      oBrowse:phyPosBlock   := {| | (DBArea)->(Recno())      }
   
      // Code blocks for the vertical scrollbar.
      // Note: DbPosition() returns values in the range from 0 to 100.
      oBrowse:posBlock      := {| | (DBArea)->(DbPosition())    }
      oBrowse:goPosBlock    := {|n| (DBArea)->(DbGoPosition(n)) }
      oBrowse:lastPosBlock  := {| | 100             }
      oBrowse:firstPosBlock := {| | 0               }
  ENDIF
RETURN oBrowse


******************************************************************
* Create std dialog window hidden
******************************************************************
FUNCTION GuiStdDialog( cTitle,aSize )
   LOCAL oDlg
   LOCAL aPos 

   DEFAULT cTitle TO "Standard Dialog Window"
   DEFAULT aSize  TO {600,400}
   aPos  := CenterPos( aSize, AppDesktop():currentSize() )
   
   oDlg          := XbpDialog():new(SetAppWindow():DrawingArea ,, aPos, aSize,, .F. )
   oDlg:icon     := 1
   oDlg:taskList := .T.
   oDlg:title    := cTitle
   oDlg:drawingArea:ClipChildren := .T.
   oDlg:create()
   oDlg:drawingArea:setFontCompoundName( FONT_DEFPROP_SMALL )

RETURN oDlg


FUNCTION CenterPos( aSize, aRefSize )
RETURN { Int( (aRefSize[1] - aSize[1]) / 2 ) ;
       , Int( (aRefSize[2] - aSize[2]) / 2 ) }


******************************************************************
* Get filename with filedialog
******************************************************************
FUNCTION GetFilename()
   LOCAL oDlg := XbpFiledialog():New(), cFile

   oDlg:Title := "Please select DBF file"
   oDlg:Create()
   cFile := oDlg:Open( "DATA\*.DBF" )

RETURN cFile


******************************************************************
* Data code block for fields
******************************************************************
FUNCTION EditBlock( cFieldName,cArea )
   LOCAL bBlock, cBlock
   IF (cArea)->(FieldPos( cFieldName )) <> 0
      cBlock := "{|x| IIf(x==NIL,"+cArea+"->"+cFieldName+","+cArea+"->"+"(WriteField('FIELD->"+cFieldName+"',x)) ) }"
      bBlock := &cBlock
   ENDIF
RETURN bBlock


******************************************************************
* Write value to field when record lock can be obtained
******************************************************************
FUNCTION WriteField( cField, xValue ) 
   LockRec()
   &cField := xValue
   DbUnlock()
RETURN xValue
