/*
 * Class for selecting drives and directories
   by James W. Loughner
   jwrl@worldnet.att.net
   Placed in public domain 04/29/2000

   VAR startDir     // Start Directory 
   VAR dirMethod    // optioanl Codeblock to return a list of directories
                       {|cCurDir| your_dir_function(cCurDir)}
                       see LOCALDIR method for example
   VAR FileMethod   // optional Codeblock to return a list of files
                       {|cCurDir| your_file_function(cCurDir)}
                       see LOCALFILE method for example
                       Note if you want to display only certain type of files
                       you can do it here!
   VAR SepChr       // seperation char defaults to "\"
   VAR ShowDrive    // logical to allow selection of drives defaults .T.
   VAR ShowFiles    // logical to show files defaults .F. (ie directories only)


 */

//    Example calls



/*
 * This directive calculates a centered position
 */


#include "Appevent.ch"
#include "Directry.ch"
#include "Common.ch"
#include "Font.ch"
#include "Gra.ch"
#include "Xbp.ch"


//#define TEST   // remark out to use in your program

#define  ICON_DRIVE             100
#define  ICON_OPENFOLDER        101
#define  ICON_CLOSEDFOLDER      102
#define  ICON_BMP               103
#define  ICON_JPG               104
#define  ICON_GIF               105
#define  ICON_HTML              106
#define  ICON_FILE              107

#ifdef TEST
     // define icons 

/*
 * This directive calculates a centered position
 */

#xtrans  CenterPos( <aSize>, <aRefSize> ) => ;
         { Int( (<aRefSize>\[1] - <aSize>\[1]) / 2 ) ;
         , Int( (<aRefSize>\[2] - <aSize>\[2]) / 2 ) }


PROCEDURE MAIN
LOCAL aSize,aPos,oDlg,DrawingArea,oDirs,oSel,ofield,oSelField
   aSize := {300,360}
   aPos  := CenterPos( aSize, AppDesktop():currentSize() )

   oDlg := XbpDialog():new( ,, aPos, aSize, , .T. )
   oDlg:taskList := .T.
   oDlg:maxButton := .F.
   oDlg:border   := XBPDLG_DLGBORDER
   oDlg:title    := "Select File"
   oDlg:close    := {|| Appquit()}
   oDlg:create()

   drawingArea := oDlg:drawingArea
   drawingArea:setFontCompoundName( FONT_HELV_SMALL )

   // create file tree
   oDirs  := DirSelector():new ( drawingArea, , {5 ,64}, {290,250} )
   oDirs:tabStop := .T.
   oDirs:StartDir := CurDrive()+":\"+CurDir()  // directory to show on startup
   oDirs:ShowFiles := .T.  // set to .F. to show and select only directories
   oDirs:create()

   // SLE to show selection normaly this would be on another dialog
   oSelField := XBPSLE():NEW(drawingArea,,{150,2},{100,20})
   oSelField:Bufferlength := 75
   oSelField:Create()

   oSel := XbpPushButton():new(drawingArea,,{73,2},{54,22} )
   oSel:Caption := "Select"
   oField:=oSelfield
   oSel:Activate:={||SetDir(oDlg,oDirs,oField)}
   oSel:Create()

   nEvent := xbe_None
   DO WHILE .T.
      nEvent := AppEvent( @mp1, @mp2, @oDlg )
      odlg:HandleEvent ( nEvent, mp1, mp2 ) 
   ENDDO

RETURN

FUNCTION SetDir(oDlg,oDirs,oField)
   IF oDirs:dirTree:GetData() != NIL
      oField:SetData(oDirs:dirTree:GetData():getdata())
      // oDlg:Destroy() // uncomment for real app to remove file select dialog
      SetAppfocus(oField)
   ENDIF
RETURN NIL


FUNCTION APPQUIT
   QUIT
RETURN NIL

PROCEDURE APPSYS
RETURN

#endif


CLASS DirSelector FROM XbpStatic
   PROTECTED:
   VAR aDrives
   VAR oItem    // current item

   EXPORTED:
   VAR startDir
   VAR dirTree
   VAR driveList
   VAR dirMethod
   VAR FileMethod
   VAR Root
   VAR SepChr
   VAR ShowDrive
   VAR ShowFiles
   VAR First_Time
   METHOD init, create, fillTree, curDir ,LocalDir,LocalFiles,LoadDir,ParsePath
ENDCLASS



/*
 * Initialize the object
 */
METHOD DirSelector:init( oParent, oOwner, aPos, aSize, aPresParam, lVisible )

   ::xbpStatic:init( oParent, oOwner, aPos, aSize, aPresParam, lVisible )
   ::xbpStatic:type    := XBPSTATIC_TYPE_RAISEDRECT

   ::dirTree           := XbpTreeView():new( self )
   ::dirTree:tabStop   := .T.
   ::dirTree:hasLines  := .T.
   ::dirTree:hasButtons:= .T.

   ::driveList         := XbpComboBox():new( self )
   ::driveList:tabStop := .T.
   ::driveList:type    := XBPCOMBO_DROPDOWNLIST

   ::startDir          := CurDrive() + ":\" + CurDir()
   ::ShowDrive         := .T.
   ::SepChr            := "\"
   ::ShowFiles         := .F.
   
RETURN self



/*
 * Request system resources and fill tree view with
 * data from current directory
 */
METHOD DirSelector:create( oParent, oOwner, aPos, aSize, aPresParam, lVisible )
   ::xbpStatic:create( oParent, oOwner, aPos, aSize, aPresParam, lVisible )

   aSize := ::currentSize()

  /*
   * The tree view creates a user-defined event with the selected
   * directory attached as the first message parameter
   */
   ::dirTree:create  (,, {4, 30}, {aSize[1]-8, aSize[2]-30-4 } )
   ::dirTree:itemSelected := {|oItem| ::fillTree( oItem ) }
   ::First_Time := .T.
//   ::dirTree:itemMarked   := {|oItem| PostAppEvent( xbeDS_DirChanged, oItem:getData() ) }
//   ::dirTree:rootItem:setData( ::startDir )

   IF ::ShowDrive
      ::driveList:create(,, {4, 30-4-24-96}, {aSize[1]-8, 24+96 } )

      ::aDrives := AvailDrives()
      AEval( ::aDrives, {|c| ::driveList:xbpListBox:addItem( c+":" ) } )

      IF At( ":", ::startDir ) == 2
         ::driveList:xbpListBox:setData( { AScan( ::aDrives, Left(::startDir,1) ) } )
      ELSE
         ::driveList:xbpListBox:setData( { AScan( ::aDrives, CurDrive() ) } )
      ENDIF
      ::driveList:xbpListBox:itemSelected := {|| ::fillTree() }
   ENDIF
   ::fillTree( ::dirTree:rootItem )
RETURN self



/*
 * Rebuild the tree view with data from the selected directory. Each
 * XbpTreeViewItem contains a character string of the directory
 * which is displayed by the item
 */
METHOD DirSelector:fillTree( oItem )
LOCAL cDrive,oParent,aDir,cCurDir,aItems,aTemp,oTemp,cTemp,I,J,K,tt
   IF EMPTY(::Root) .OR. oItem == NIL
      cDrive := ALLTRIM(::driveList:XbpSLE:getData())
   ELSE
      cDrive := ::Root
   ENDIF
   IF oItem == ::dirTree:rootItem .OR. oItem == NIL
     /*
      * The tree view begins with the drive letter
      */
      IF oItem == NIL     //::dirTree:rootItem
         ::dirTree:rootItem:Expand(.F.)
      ENDIF 
      aItems := ::dirTree:rootItem:getChildItems()
      AEval( aItems, {|obj| ::dirTree:rootItem:delItem(obj) } )
      oParent := ::dirTree:rootItem:addItem( cDrive, ICON_DRIVE, ICON_DRIVE, ICON_DRIVE,, cDrive )

      cCurDir := cDrive+::SepChr
      ::Root := cCurDir 
      oItem  := oParent
      ::LoadDir(oItem,cCurdir)
      oParent:Expand(.T.) 
      SetAppFocus(oParent)
      tt := ::Startdir
      IF ::First_Time 
         IF oItem!=NIL .AND.!EMPTY(::Startdir) .AND. ALLTRIM(UPPER(cCurDir)) != ALLTRIM(UPPER(::Startdir))
            oTemp := oItem
            aTemp := ::ParsePath(::Startdir)
         
            FOR I=1 TO Len(aTemp)
               aItems := oTemp:getChildItems()
            
               FOR J=1 TO LEN(aItems)
                  IF aTemp[I] = aItems[J]:GetData()
                     oTemp := aItems[J]
                     ::LoadDir(oTemp,aTemp[I])
                     EXIT      
                  ENDIF
               NEXT    
          
             NEXT I  
         ENDIF
         ::First_Time := .F.
      ENDIF   
   ELSE
      cCurDir := oItem:GetData()
      IF EMPTY(cCurDir)
         cCurDir := ::Root
      ENDIF
      ::LoadDir(oItem,cCurdir)
   ENDIF
RETURN self  

METHOD ParsePath(cDir)
LOCAL aRtn :={},nPtr:= 1
   IF Substr(cDir,LEN(cDir),1)!= ::SepChr
      cDir := cDir+::SepChr
   ENDIF  
   DO WHILE nPtr>0 
      nPtr:=AT(::SepChr,cDir,nPtr)+1
      AADD(aRtn,left(cDir,nPtr-1))
      IF nPtr >= LEN(cDir)
         EXIT
      ENDIF   
   ENDDO
RETURN aRtn

METHOD DirSelector:LoadDir(oItem,cCurDir)
LOCAL oTemp,aDir,Dispbmp,I
   IF EMPTY(oItem:GetChildItems())
      IF EMPTY(::DirMethod)
         aDir := ::Localdir(cCurDir)
      ELSE
         aDir := Eval(::DirMethod,cCurDir)
      ENDIF
      FOR I:= 1 TO LEN(aDir)
          oTemp := oItem:addItem( aDir[I]   , ;
                                  ICON_CLOSEDFOLDER, ;
                                  ICON_CLOSEDFOLDER, ;
                                  ICON_OPENFOLDER  , ;
                                  NIL              , ;
                                  cCurDir +IF(Substr(cCurDir,LEN(cCurDir),1)=::SepChr,"" ,::SepChr)+ aDir[I] )

      NEXT
      IF ::ShowFiles
         IF EMPTY(::FileMethod)
            aDir := ::LocalFiles(cCurDir)
         ELSE
            aDir := Eval(::FileMethod,cCurDir)
         ENDIF
         FOR I:= 1 TO LEN(aDir)
             DO CASE
                CASE ".BMP"$UPPER(aDir[I])
                   Dispbmp := ICON_BMP
                CASE ".JPG"$UPPER(aDir[I])
                   Dispbmp := ICON_JPG
                CASE ".GIF"$UPPER(aDir[I])
                   Dispbmp := ICON_GIF
                CASE ".HTM"$UPPER(aDir[I])
                   Dispbmp := ICON_HTML

             OTHERWISE
                Dispbmp := ICON_FILE
             ENDCASE
             oTemp := oItem:addItem( aDir[I]   , ;
                                     Dispbmp, ;
                                     Dispbmp, ;
                                     Dispbmp, ;
                                     NIL              , ;
                                     cCurDir + IF(Substr(cCurDir,LEN(cCurDir),1)=::SepChr,"" ,::SepChr) + aDir[I] )

         NEXT
      ENDIF
      oItem:Expand(.T.)
   ENDIF

RETURN self 

METHOD DirSelector:LocalDir(cDir)
LOCAL aRtn:={},aDir,I,cSep
   IF Substr(cDir,LEN(cDir),1)=::SepChr 
     cSep := ""
   ELSE
     cSep := ::SepChr
   ENDIF
   aDir := DIRECTORY(cDir+cSep,"D")
   FOR I = 1 TO LEN(aDir)
      IF "D" $ aDir[I,F_ATTR]        .AND. ;
         .NOT. aDir[I,F_NAME] == "." .AND. ;
         .NOT. aDir[I,F_NAME] == ".."

         AADD(aRtn,ALLTRIM(aDir[I,F_NAME]))
      ENDIF
   NEXT
   aRtn := aSort(aRtn,,,{|aX,aY| Upper(aX[1]) < Upper(aY[1]) })

RETURN aRtn

METHOD DirSelector:LocalFiles(cDir)
LOCAL aRtn:={},aDir,I,cSep
   IF Substr(cDir,LEN(cDir),1)=::SepChr
     cSep := ""
   ELSE
     cSep := ::SepChr
   ENDIF
   aDir := DIRECTORY(cDir+cSep+"*.*")
   FOR I = 1 TO LEN(aDir)
      AADD(aRtn,ALLTRIM(aDir[I,F_NAME]))
   NEXT
   aRtn := ASORT(aRtn)
RETURN aRtn

/*
 * Return the currently selected directory
 */
METHOD DirSelector:curDir
RETURN ::oItem:getData()



#define DRIVE_NOT_READY  21

FUNCTION AvailDrives()
   LOCAL cDrive := CurDrive()
   LOCAL aDrives:= { cDrive }
   LOCAL cCurdir:= Curdir()
   LOCAL bError := ErrorBlock( {|oErr| DriveError( oErr, aDrives ) } )


   FOR i:=3 TO 26
      BEGIN SEQUENCE
        Curdrive( Chr(64+i) ) 
        CurDir()
        IF AScan( aDrives, CurDrive() ) == 0
           AAdd( aDrives, Curdrive() )
        ENDIF
      ENDSEQUENCE
   NEXT

   CurDrive( cDrive )
   Curdir(cdrive+":\"+cCurdir)
   ErrorBlock( bError )   
RETURN ASort( aDrives )



/*
 * Handle runtime error when a drive is not ready
 */
STATIC PROCEDURE DriveError( oError, aDrives )
   IF oError:osCode == DRIVE_NOT_READY
      AAdd( aDrives, oError:args[1] )
   ENDIF
   BREAK
RETURN




