/*
      JimsDialog
      by James W. Loughner
      Version 2.3
      
      Class to replace XBPDialog in FormDesigner Code
      Sets all Field Assoc SLE's to Clipper Key and basic Pictures
      Optionally attach object to DB Notifiaction to react to changes 
        in Rec pointer
      Control of db pointer via obj methods.
      Remembers the data pointer to the record being displayed and reacts
      appropriatly to any dbf changes.

     Ver 2.0 Adds translation of captions
     Ver 2.1 extends translation to all Xbp objects
             XlateStr Function to translate other strings that are not part of a XbpDialog
             now includes JimsCode.ch to simplify translation in other functions
     Ver 2.2 auto loads aCaptionobjs array when translating which simplifies the translation call
     Ver 2.3 Optional use of SLEPic class instead of PicSLE functions requires linking of SLEPic
             if an object is derived from SLEPic and has a DataLink the correct picture will
             automaticly be applied based on the type field linked to.   
     note requires linking of the functions in "PICSLE.PRG"
     see Set_SLEPic method!
*/

#include "Gra.ch"
#include "Xbp.ch"
#include "Common.ch"
#include "Appevent.ch"
#include "Font.ch"
#include "class.ch"
#include "Dmlb.ch"
#include "Appedit.ch"
#include "dbstruct.ch"
#include "Nls.ch"
#include "appevent.ch"
#include "dll.ch"
// note these define the largest string that can be translated
#define MAXBASE  75
#define MAXTRANSLATE 100

#define REFRESH_EVENT xbeP_User+100
// flags for PlaySound
#define SND_SYNC            0
#define SND_ASYNC           1
#define SND_FILENAME        131072
#define SND_PURGE           64

STATIC LanguageDB := ""

CLASS JimsDialog From XBPDialog
   HIDDEN:
      METHOD Translator
      METHOD SetCenter
      VAR IsTrans
   EXPORTED:
      METHOD init
      METHOD create
      METHOD skip 
      METHOD setData 
      METHOD getData 
      METHOD DbGoTop 
      METHOD DBGoBottom 
      METHOD AddRec 
      METHOD DelRec 
      METHOD Notify
      METHOD Translate
      METHOD UnTranslate
      METHOD Set_SLEPic // set picture to type of field ie Date is entered in data format
                       // also adds Clipper navigation keys (Note only effects objects in
                       // edcontrol array. Also if you want complex pictures like PHONE etc. 
                       // you can set there keyborad ivars in the utilization code. )
                       // This method is automaticly called in the create method
      METHOD destroy  
      METHOD Show
      VAR PendingSave
      VAR DbArea    // Data Area Alias (Must be set before create()if there is one)
      VAR Use_Notify // Automaticly use Notification (default .T.) Can be used
                    // to either sync windows or create unsynced windows with the 
                    // same class 
      VAR OKExit    // code block to deterime exit of window ie validate the data 
                    // must return .T. or .F. (optional)
      VAR GetHook  // Code block to update other (non sle/mle types / other 
                   // data areas etc.) executed after DBarea is updated in
                   // ::getdata method (Optional)              
      VAR SetHook  // Code block to update other (non sle/mle types / other 
                   // data areas etc.) executed after screen is updated in
                   // ::setdata method (Optional)              
      VAR AddHook   // code block to init other parts when apppending a record
                    // an example might be if you have a SLE that must contain a
                    // value from a parent DB. Or populate a list. etc. (optional)
      VAR EdControls // Note: this must contain the values of EditControls array from 
                     // the FD code. This is because EditControls is defined in the 
                     // impimentation level and that is derived from JimsDialog so
                     // Jimsdialog knows nothing about EditControls.
                     // (Must be set before create())
                     // note you could add other controls not in the field list
                     // but the FD only adds the fields. not other SLE's etc.
                     
      VAR ConfirmDelete // code block to confirm delete if not used delete occures 
                        // without conformation (optional) Return .T. to delete
      VAR CurrentRec  // contains Current record number 
      CLASS VAR TranslationDB // name of a data base that will translate captions (static,
                        // 3state,pushbutton,checkbox,tabpage,Radiobutton classes)
                        // DB structure is;
                        //  {{"BASE","C",MAXBASE,0},{"TRANSLATE","C",MAXTRANSLATE,0}}                                 
                        // Where BASE is the base language ie origianl in caption
                        // and TRANSLATE contains the translation to the new language
                        // if left blank or file or record not found the default caption 
                        // will display otherwise the TRANSLATE field from the record that
                        // matches the origianl caption will display (optional) 
                        // When first displayed with this option a record is added to the
                        // language dbf. The TRANSLATE field will be blank. You can then add 
                        // translations for the BASE text via DBU or an other DBF editor. 
                        // note: be sure to use or recreate the indexes
      VAR aCaptionObjs  // Array of all the caption objects in the dialog (required for translation)                     
      VAR lCenter       // Default .T. Will cause the dialog to be centered in the Parent
      VAR isProtected 
      VAR RecoverFlag 
      VAR BGColor   
      VAR ISMain
      METHOD Protect
      CLASS VAR Siblings SHARED // Array of all objects of this type
ENDCLASS

METHOD JimsDialog:init( oParent, oOwner, aPos, aSize, aPP, lVisible )
   * Execute Init method of the super class
   ::Xbpdialog:Init( oParent, oOwner, aPos, aSize, aPP, lVisible )
   ::XbpDialog:Close := {|| ::Destroy()}// go ahead and set the close action
   ::Use_Notify := .F. //default to NonSYNC type windows
   ::Currentrec := 1
   ::TranslationDB := ""
   ::aCaptionObjs := {}
   ::lCenter := .T. 
   ::IsTrans := .F.
   ::PendingSave := .F.   
   ::isProtected  := .F.
   IF Valtype(::Siblings) != "A"
      ::Siblings := {}
   ENDIF   
   AADD(::Siblings,Self) // add this instance to the array
//   ::SetEventMask( 3, .T. )
   ::RecoverFlag := .F.  
   ::EdControls := {} 
   ::Icon := 3  // pritnpresssmall  
   ::BGColor := GRAMAkeRGBColor({0XF1,0XF1,0XF1}) 
   ::MoveWithOwner := .T. 
   ::IsMain := .F. 
   ::TaskList := .F. 
   ::ClipSiblings := .T.
RETURN Self   


METHOD JimsDialog:create( oParent, oOwner, aPos, aSize, aPP, lVisible )
LOCAL I
   IF !::IsMain
      ::MaxButton := .F.
      ::MinButton := .F.
      ::TaskList := .F.
   ELSE
      ::TaskList := .T.
   ENDIF
   * Execute Create method of the super class
//   IF !EMPTY(LanguageDB) .AND. EMPTY(::TranslationDB).AND. !::IsTrans
//      ::TranslationDB := LanguageDb  
//   ENDIF   

   ::Xbpdialog:create( oParent, oOwner, aPos, aSize, aPP, .F. ) 
   IF !::IsMain
      ::MaxSize := ::CurrentSize()
      ::MinSize := ::CurrentSize()
   ENDIF   
   IF ::DBArea != NIL
      ::Currentrec := (::DBarea)->(RECNO())
      IF ::EdControls != NIL
         //* set @K process to all SLE's in  EdControls which is normaly the FD EditContorls array
         AEval ( ::EdControls, { | oXbp | ::Set_SLEPic(oXbp) } )
      ENDIF  
      IF ::Use_Notify
        (::Dbarea)->(DBRegisterClient(self)) //register the object with the dbf
      ENDIF 
   ENDIF 
   ::DrawingArea:SetColorBG(::BGColor) 
//   SetClipperStyle(::DrawingArea)
//   ::SetFontCompoundName(FONT_HELV_SMALL)
   IF ::EdControls != nil
//      SetAppFocus(::EdControls[1])
   ENDIF   
   IF ::lCenter
      ::SetCenter()
   ENDIF
   IF ::isProtected
      FOR I = 1 TO LEN(::EdControls)
         DO CASE
            CASE ::EdControls[I]:IsDerivedfrom("SLEPIC")
               ::EdControls[I]:PROTECT(.T.)  
            CASE ::EdControls[I]:IsDerivedfrom("COMBOBOXSLE")
               ::EdControls[I]:PROTECT(.T.)  
            CASE ::EdControls[I]:IsDerivedfrom("XBPCHECKBOX")
               ::EdControls[I]:DISABLE()  
         ENDCASE
      NEXT I
   ENDIF
RETURN Self   

METHOD JimsDialog:Show()
LOCAL aChild,I
   ::XbpDialog:Show()
   aChild := ::DrawingArea:ChildList()
   FOR I = 1 TO LEN(aChild)
      IF aChild[I]:IsDerivedFrom("ShadowButton")
         aChild[I]:Paint()
      ENDIF
      IF aChild[I]:IsDerivedFrom("SLEPopUp")
         aChild[I]:oPush:Paint()
      ENDIF
      IF aChild[I]:IsDerivedFrom("ComboBoxSLE")
         aChild[I]:oPush:Paint()
      ENDIF
   NEXT I
RETURN Self

METHOD JimsDialog:Protect(lVal)
LOCAL I
   IF lVal = NIL
      lVal := .T.
   ENDIF   
   IF lVAl
      FOR I = 1 TO LEN(::EdControls)
         DO CASE
            CASE ::EdControls[I]:IsDerivedfrom("SLEPIC")
               ::EdControls[I]:PROTECT(.T.)  
            CASE ::EdControls[I]:IsDerivedfrom("COMBOBOXSLE")
               ::EdControls[I]:PROTECT(.T.)  
            CASE ::EdControls[I]:IsDerivedfrom("XBPCHECKBOX")
               ::EdControls[I]:DISABLE()  
         ENDCASE
      NEXT I 
      ::isProtected := .T.
   ELSE
      FOR I = 1 TO LEN(::EdControls)
         DO CASE
            CASE ::EdControls[I]:IsDerivedfrom("SLEPIC")
               ::EdControls[I]:PROTECT(.F.)  
            CASE ::EdControls[I]:IsDerivedfrom("COMBOBOXSLE")
               ::EdControls[I]:PROTECT(.F.)  
            CASE ::EdControls[I]:IsDerivedfrom("XBPCHECKBOX")
               ::EdControls[I]:ENABLE()  
         ENDCASE
      NEXT I 
      ::isProtected := .F.
   ENDIF    
RETURN Self

METHOD JimsDialog:skip( nSkip ) 
LOCAL lTemp := ::Use_Notify
   IF ::DBArea != NIL
      ::Use_Notify := .F.
      (::DBArea)->(DBGoto(::CurrentRec))
      IF ::getData()                      // Write record current record 
         (::DBarea)->(DbSkip( nSkip ))    // move record pointer
         IF (::DBarea)->(EOF())   
            (::DBarea)->(DbSkip( -1 ))    // move back to lastrec
         ENDIF 
         ::Currentrec := (::DBarea)->(RECNO()) //remember record number
         ::setData()                      // Read record new record
      ENDIF 
      ::Use_Notify := lTemp
   ENDIF   
RETURN self 


METHOD JimsDialog:DbGoTop() 
LOCAL lTemp := ::Use_Notify
   IF ::DBArea != NIL
      ::Use_Notify := .F.
      IF ::getData()                      // Write record 
         (::DBarea)->(DbGoTop())  // Change record pointer 
         ::Currentrec := (::DBarea)->(RECNO())
         ::setData()                      // Read record 
      ENDIF 
      ::Use_Notify := lTemp
   ENDIF
RETURN self 


METHOD JimsDialog:DbGoBottom() 
LOCAL lTemp := ::Use_Notify
   IF ::DBArea != NIL
      ::Use_Notify := .F.
      IF ::getData()                      // Write record 
         (::DBarea)->(DbGoBottom())  // Change record pointer 
         ::Currentrec := (::DBarea)->(RECNO())
         ::setData()                      // Read record 
      ENDIF 
      ::Use_Notify := lTemp
   ENDIF 
RETURN self 


METHOD JimsDialog:setData() 
LOCAL lTemp := ::Use_Notify, Currec :=(::DBarea)->(RecNo()) 
   IF ::DBArea != NIL
      ::Use_Notify := .F.
      (::DBarea)->(DbSuspendNotifications())
      (::DBarea)->(DBGoto(Currec))
      (::DBarea)->(DbResetNotifications())
      AEval( ::edControls, {|o| o:setData() } ) // set the screen
      IF ValType(::SetHook) = "B"
         Eval(::SetHook, Self)
      ENDIF
      ::Use_Notify := lTemp
      (::DBarea)->(DbSuspendNotifications())
      (::DBarea)->(DBGoto(Currec))
      (::DBarea)->(DbResetNotifications())
   ENDIF
RETURN self 


METHOD JimsDialog:getData() 
LOCAL lOk:=.F.,lChanged := .T. 
LOCAL lTemp := ::Use_Notify
LOCAL aList := (::DBarea)->(DbClientList()),I,oMain 
   IF ::DBArea != NIL
      ::Use_Notify := .F.
   
      // Set not exit conditions here
      IF ValType(::OKExit) = "B"
         lChanged := Eval(::OKExit,Self) 
      ENDIF
      IF lChanged                            // Data is changed 
         (::DBarea)->(DbSuspendNotifications())
         IF ::PendingSave   // And AddRec has occured
            #ifdef DEMO
            IF (::DBarea)->(LastRec()) < 750
               (::DBarea)->(DbAppend()) // Append & Lock record 
               ::Currentrec := (::DBarea)->(RECNO())
               (::DBarea)->(DbUnlock())
            ELSE
               oMain := SetAppWindow(Self)
               MSGBOX("Demo Limit Exceded Call MTO at 304-546-5013")
               SetAppWindow():Destroy()
               APPQuit()
            ENDIF       
            #else
            IF ::RecoverFlag 
               Set Deleted OFF
               (::DBArea)->(DBSeek("XXXX"))
               IF !(::DBArea)->(FOUND())
                  (::DBarea)->(DbAppend()) // Append & Lock record 
                  ::Currentrec := (::DBarea)->(RECNO())
                  (::DBarea)->(DbUnlock())
               ENDIF
               SET DELETED ON
            ELSE
               (::DBarea)->(DbAppend()) // Append & Lock record 
               ::Currentrec := (::DBarea)->(RECNO())
               (::DBarea)->(DbUnlock())
            ENDIF
            #endif
         ENDIF
         ::PendingSave := .F.
         (::DBarea)->(DBGoto(::Currentrec)) // be sure we are sitting on the right record
         lOk := (::DBarea)->(DbRLock()) // Lock record 
         IF lOk                           // Record is locked 
            AEval( ::edControls, {|o| o:getData() } ) //Write data
            IF ValType(::GetHook) = "B"
               Eval(::GetHook, Self)
            ENDIF
            (::DBarea)->(DbUnlock())    // Unlock record 
         ELSE
            oMain := SetAppWindow(Self)    
            MsgBox( "Record is currently locked! Not updated." ) 
            SetAppWindow(oMain)
         ENDIF 
         // Notify other objects a change has occured
         FOR I=1 TO LEN(aList)
            IF aList[I] != Self
               aList[I]:notify(xbeDBO_Notify,DBO_TABLE_UPDATE)
            ENDIF   
         NEXT I
         (::DBarea)->(DbResetNotifications())
      ENDIF 
//      SetAppFocus(::EdControls[1])
      ::Use_Notify := lTemp
   ENDIF   
RETURN lOk 

METHOD JimsDialog:Addrec() 
LOCAL lTemp := ::Use_Notify
   IF ::DBArea != NIL
      ::Use_Notify := .F.
      IF .T. //::Getdata()                 // Save current record 
//         (::DBarea)->(DbAppend()) // Append & Lock record 
         (::DBarea)->(DbSuspendNotifications())
         (::DBarea)->(DBGoBottom())
         DO WHILE !(::DBarea)->(EOF()) // make sure
            (::DBarea)->(DBSkip(1)) // move to phantom record
         ENDDO   
//         ::Currentrec := (::DBarea)->(RECNO())
         AEval( ::edControls, {|o| o:SetData() } ) // read data 
         (::DBarea)->(DBGOTO(::CurrentRec)) // return to current record
         (::DBarea)->(DbResetNotifications())
         IF ValType(::AddHook)="B"
            Eval(::AddHook,Self)
         ENDIF
         ::PendingSave := .T.
//         (::DBarea)->(DbUnlock())    // Unlock record  
      ENDIF 
      ::Use_Notify := lTemp
   ENDIF   
RETURN Self 

METHOD JimsDialog:DelRec() 
LOCAL lOk:=.T.,I
LOCAL lTemp := ::Use_Notify
   IF ::DBArea != NIL
      ::Use_Notify := .F.
   
      IF Valtype(::Confirmdelete)="B"
         lOk := Eval(::Confirmdelete,Self)
      ENDIF 
      IF ConfirmBox(Self,"Delete this record?","WARNING!!",XBPMB_YESNO,XBPMB_QUESTION)= XBPMB_RET_NO
         lOk := .F.
      ENDIF   
      IF lOk
         IF (::DBarea)->(DbRlock())
            FOR I = 1 TO (::DBarea)->(Fcount())
                DO CASE
                   CASE ValType((::DBarea)->(FieldGet(I))) = "C"
                      (::DBarea)->(FieldPut(I,""))
                   CASE ValType((::DBarea)->(FieldGet(I))) = "N"
                        (::DBarea)->(FieldPut(I,0))                      
                   CASE ValType((::DBarea)->(FieldGet(I))) = "D"
                      (::DBarea)->(FieldPut(I,CTOD("")))
                   CASE ValType((::DBarea)->(FieldGet(I))) = "L"
                      (::DBarea)->(FieldPut(I,.F.))
                   CASE ValType((::DBarea)->(FieldGet(I))) = "M"
                      (::DBarea)->(FieldPut(I,""))
                ENDCASE
            NEXT I
            (::DBarea)->(DbDelete())
            (::DBarea)->(DBUnlock())    
            (::DBarea)->(DbSkip( -1 ))    // move record pointer
            ::Currentrec := (::DBarea)->(RECNO()) //remember record number
            IF ::edControls != NIL
               AEval( ::edControls, {|o| o:setData() } )
            ENDIF   
         ENDIF     
      ENDIF   
      ::Use_Notify := lTemp
   ENDIF   
RETURN Self 


METHOD JimsDialog:Notify(nEvent,mp1,mp2)
LOCAL lOk
   IF ::DBArea != NIL
      (::DBarea)->(DbSuspendNotifications())
      IF ::Use_Notify 
         DO CASE
            CASE mp1 == DBO_MOVE_PROLOG // record pointer is about to be moved save data if changed
               IF !( AScan( ::edControls, {|o| o:changed } ) == 0 )
                  lOk := (::DBarea)->(DbRLock()) // Lock record 
                  IF lOk                           // Record is locked 
                     AEval( ::edControls, {|o| o:getData() } ) //Write data
                     IF ValType(::GetHook) = "B"
                        Eval(::GetHook, Self)
                     ENDIF
                     (::DBarea)->(DbUnlock())    // Unlock record 
                  ENDIF 
               ENDIF 
               ::SetData()
            CASE mp1 == DBO_MOVE_DONE .OR. ;       // skip is done
                 mp1 == DBO_GOBOTTOM  .OR. ;
                 mp1 == DBO_TABLE_UPDATE .OR.;
                 mp1 == DBO_TABLE_DELETED  .OR. ;
                 mp1 == DBO_TABLE_APPEND .OR.;
                 mp1 == DBO_GOTOP
               ::Currentrec := (::DBarea)->(RECNO()) 
               ::SetData()
         ENDCASE
      ELSEIF ::CurrentRec = (::DBarea)->(RECNO()) 
         // only update screen if deleted or edited
         DO CASE
            CASE mp1 == DBO_TABLE_UPDATE
               AEval( ::edControls, {|o| o:setData() } ) 
            CASE mp1 == DBO_TABLE_DELETED
               ::Currentrec := (::DBarea)->(RECNO()) 
               ::SetData()
         ENDCASE
      ENDIF   
      (::DBarea)->(DbResetNotifications())
   ENDIF   
RETURN Self

METHOD JimsDialog:Set_SLEPic(o)
LOCAL aStruc:={},cCB,cField:="",cTemp,I
   aStruc := (::DBarea)->(DBStruct())
   // First get the field name from the datalink code block
   // note this might not work if you are using custom datalinks
   // should be of the form:
   // {|x| IIf( PCOUNT()==0, Trim( (::DATA)->PHONE ), (::DATA)->PHONE := x ) }
   // note the last '>' and the space after the field name
   cCB := VAR2Char(o:datalink)
   cTemp := SUBSTR(cCB,RAT(">",cCB)+1)
   FOR I = 1 TO LEN(cTemp)
       IF cTemp[I] = " " .OR. cTemp[I] = ":"
          EXIT
       ENDIF
       cField += cTemp[I] 
   NEXT I
   // now set a picture type
   I := ASCAN(aStruc, {|X| X[1]==cField} )   
   IF I != 0
      DO CASE
         CASE aStruc[I,DBS_TYPE]="C" // genaric character
              IF o:isDerivedfrom("SLEPic")
                 o:Bufferlength := AStruc[I,DBS_LEN]
                 o:Picture := REPLICATE("X",AStruc[I,DBS_LEN]) 
              ELSEIF !o:isDerivedfrom("XBpCombobox")
                 o:Keyboard := {|nKey,x,obj| SLEAtK(nKey,obj)}
              ENDIF   
         CASE aStruc[I,DBS_TYPE]="N" // genaric float
              IF o:isDerivedfrom("SLEPic")
                 o:Bufferlength := AStruc[I,DBS_LEN]
                 o:DataType := "N"  
                 IF AStruc[I,DBS_DEC]>0
                    o:Picture := REPLICATE("9",AStruc[I,DBS_LEN]-1-AStruc[I,DBS_DEC])+"."+REPLICATE("9",AStruc[I,DBS_DEC]) 
                 ELSE
                    o:Picture := REPLICATE("9",AStruc[I,DBS_LEN]) 
                 ENDIF   
              ELSE
                 o:Keyboard := {|nKey,x,obj| SLENumeric(nKey,obj,AStruc[I,DBS_LEN],AStruc[I,DBS_DEC])}
              ENDIF   
         CASE aStruc[I,DBS_TYPE]="L" // logical
              DO CASE
                 CASE o:isDerivedfrom("SLEPic" ) //'Y','N' char input
                    o:Bufferlength := 1
                    o:Picture := "Y"
                 CASE o:isDerivedfrom("XBPSLE" ) //'Y','N' char input
                    o:Keyboard := {|nKey,x,obj| SLELogical(nKey,obj)}
                 CASE o:isDerivedfrom("XBPCHECKBOX" ) // checkbox input
                    o:Keyboard := {|nKey,x,obj| SLEMove(nKey,obj)}
              ENDCASE   
         CASE aStruc[I,DBS_TYPE]="D" // date also do validation here
              IF o:isDerivedfrom("SLEPic")
                 DO CASE
                    CASE Setlocale(NLS_IDATE) = "0"  // mm/dd/yy 
                       IF Setlocale(NLS_ICENTURY) = "0"  //YY
                          o:Bufferlength := 8
                          o:Picture := "99/99/99"
                       ELSE                            //YYYY
                          o:Bufferlength := 10
                          o:Picture := "99/99/9999"
                       ENDIF
                    CASE Setlocale(NLS_IDATE) = "1"  // dd/mm/yy 
                       IF Setlocale(NLS_ICENTURY) = "0"  //YY
                          o:Bufferlength := 8
                          o:Picture := "99/99/99"
                       ELSE                           //YYYY
                          o:Bufferlength := 10
                          o:Picture := "99/99/9999"
                       ENDIF
                    CASE Setlocale(NLS_IDATE) = "2"  // yy/mm/dd 
                       IF Setlocale(NLS_ICENTURY) = "0"  //YY
                          o:Bufferlength := 8
                          o:Picture := "99/99/99"
                       ELSE                            //YYYY
                          o:Bufferlength := 10
                          o:Picture := "9999/99/99"
                       ENDIF
                 ENDCASE          
                 o:HoldFormat := .T.
              ELSE
                 o:Keyboard := {|nKey,x,obj| SLEDate(nKey,obj)}
                 o:KillInputFocus := {|x,y,obj| IF(SLEvalidateDate(obj),,SetAppFocus(obj))} 
              ENDIF   
         CASE aStruc[I,DBS_TYPE]="M" // memo

         OTHERWISE     
              o:Keyboard := {|nKey,x,obj| SLEAtK(nKey,obj)}
      ENDCASE   
   ELSE
//      o:Keyboard := {|nKey,x,obj| SLEAtK(nKey,obj)}
   ENDIF
RETURN Self

METHOD JimsDialog:Destroy()
LOCAL I,oMain
   IF SELECT(::DBarea)>0 .AND. ::Use_Notify
     (::DBarea)->(DBDeRegisterClient(Self))
   ENDIF  
   I := aScan(::Siblings,Self)  // find this instance in array
   IF I > 0
      aDel(::Siblings,I)  // remove self from array
      aSize(::Siblings,Len(::Siblings)-1)
   ELSE
//      oMain := SetAppWindow(Self)
//      AADD(::Siblings,Self)
//      MSGBOX("Error Self not found in siblings")
//      SetAppWindow(oMain)       
      Return Self   // stops call ing destroy twice
   ENDIF  

   IF VALType(oMainWindow) = "O" .AND. oMainWindow != SetAppWindow()   // make sure SetAppWindow points to the main window
      SetAppWindow(oMainWindow)
   ENDIF   

   ::DbArea        := NIL
   ::OKExit        := NIL
   ::GetHook       := NIL
   ::SetHook       := NIL
   ::EdControls    := NIL
   ::ConfirmDelete := NIL
   ::CurrentRec    := NIL 
   ::XBPDialog:destroy()// we assume this will take care of all children also!
RETURN Self

METHOD JimsDialog:Translate()
   IF !EMPTY(::TranslationDB)
      IF FILE(::TranslationDB+IF(".DBF"$UPPER(::TranslationDB),"",".DBF"))
         IF !::IsTrans
            LanguageDB := ::TranslationDB 
            ::aCaptionObjs := GetCaptionObj(::DrawingArea)
            ::Translator()  
         ENDIF
      ELSE
         DbCreate(::TranslationDB,{{"BASE","C",MAXBASE,0},{"TRANSLATE","C",MAXTRANSLATE,0}})
         IF !::IsTrans
            LanguageDB := ::TranslationDB 
            ::aCaptionObjs := GetCaptionObj(::DrawingArea)
            ::Translator() 
         ENDIF
      ENDIF
   ELSEIF ::IsTrans 
      ::UnTranslate()
      LanguageDB := ::TranslationDB 
   ENDIF
   Self:InvalidateRect()
RETURN Self

METHOD JimsDialog:Translator()
LOCAL PrevArea,Fname,fname2,I,J,cMulti,Temp,lHadToOpen,aTemp,test,PrevOrd
   PrevArea := SELECT()
   aTemp := OpenXlateDB(::TranslationDB)
   Fname := aTemp[1]
   lHadToOpen := aTemp[2]
   SELECT (Fname)
   PrevOrd := Ordsetfocus("BASE")
   IF !EMPTY(::GetTitle())     // take care of self (dialog)
      SEEK ::GetTitle()
      IF !EOF() 
         IF !EMPTY((Fname)->TRANSLATE)
            ::SetTitle( ALLTRIM((Fname)->TRANSLATE) )
         ENDIF
      ELSEIF !::IsTrans
         (Fname)->(DBAppend())
         REPLACE (Fname)->BASE WITH ::GetTitle()
         (Fname)->(DBUNLOCK())
      ENDIF
      // deal with menus
      
      IF ::MenuBar():NumItems() > 0
         XlateMenu(::MenuBar())         
      ENDIF
   ENDIF      
   FOR I = 1 TO LEN(::aCaptionObjs)
      IF ::aCaptionObjs[I]:IsDerivedFrom("XbpDialog")  // special handling XbpDialog child
         SEEK ::aCaptionObjs[I]:GetTitle()
         IF !EOF() 
            IF !EMPTY((Fname)->TRANSLATE)
               ::aCaptionObjs[I]:SetTitle( ALLTRIM((Fname)->TRANSLATE) )
            ENDIF
         ELSEIF !::IsTrans
            (Fname)->(DBAppend()) 
            REPLACE (Fname)->BASE WITH ::aCaptionObjs[I]:GetTitle()
            (Fname)->(DBUNLOCK())
         ENDIF
         LOOP
      ENDIF   
      IF ::aCaptionObjs[I]:IsDerivedFrom("XbpMultiCellGroup")  // special handling for XbpQuickBrowse
         FOR J = 1 TO ::aCaptionObjs[I]:MaxCol
            IF ValType(::aCaptionObjs[I]:GetCell(1,J))="C"   // should be the heading cell and we only want charcter type
               (Fname)->(DBSEEK(::aCaptionObjs[I]:GetCell(1,J))) 
               IF !(Fname)->(EOF())
                  IF !EMPTY((Fname)->TRANSLATE)
                     ::aCaptionObjs[I]:SetCell(1,J,ALLTRIM((Fname)->TRANSLATE))
                  ENDIF
               ELSEIF !::IsTrans
                  (Fname)->(DBAPPEND())
                  REPLACE (Fname)->BASE WITH ::aCaptionObjs[I]:GetCell(1,J)
                  (Fname)->(DBUNLOCK())
               ENDIF   
            ENDIF    
         NEXT I
         LOOP
      ENDIF
      IF ::aCaptionObjs[I]:IsDerivedFrom("XbpCellGroup") // special for XbpBrowse
         (Fname)->(DBSEEK( ::aCaptionObjs[I]:GetCell(1)))
      ELSE
         (Fname)->(DBSEEK( ::aCaptionObjs[I]:Caption))  //Everything else should have a caption
      ENDIF   
      IF !(Fname)->(EOF())
         IF !EMPTY((Fname)->TRANSLATE)
            IF ::aCaptionObjs[I]:IsDerivedFrom("XbpCellGroup")
               IF ::aCaptionObjs[I]:type = XBPCOL_TYPE_TEXT // Forget it if its not a text type
                  ::aCaptionObjs[I]:SetCell(1,ALLTRIM((Fname)->TRANSLATE))
               ENDIF
            ELSE         
               ::aCaptionObjs[I]:SetCaption( ALLTRIM((Fname)->TRANSLATE) )
            ENDIF   
         ENDIF
      ELSEIF !::IsTrans
         (Fname)->(DBAPPEND())
         IF ::aCaptionObjs[I]:IsDerivedFrom("XbpCellGroup")
            IF ::aCaptionObjs[I]:type = XBPCOL_TYPE_TEXT // Forget it if its not a text type
               Test := ::aCaptionObjs[I]
               REPLACE (Fname)->BASE WITH ::aCaptionObjs[I]:GetCell(1)
            ENDIF
         ELSE         
            REPLACE (Fname)->BASE WITH ::aCaptionObjs[I]:Caption
         ENDIF   
         (Fname)->(DBUNLOCK())
      ENDIF   
   NEXT I
   ::IsTrans := .T.
   IF lHadToOpen
      (Fname)->(DBCloseArea()) 
   ELSE
      (Fname)->(OrdSetFocus(PrevOrd))   
   ENDIF   
   SELECT (PrevArea)
   ::InvalidateRect()
RETURN Self


METHOD JimsDialog:UnTranslate()
LOCAL PrevArea,Fname,fname2,I,J,cMulti,lHadToOpen,aTemp
   IF EMPTY(LanguageDB)
      RETURN Self
   ENDIF
   PrevArea := SELECT()
   aTemp := OpenXlateDB(LanguageDB)
   Fname := aTemp[1]
   lHadToOpen := aTemp[2]
   SELECT (Fname)
   Ordsetfocus("TRANS")
   IF !EMPTY(::GetTitle())     // take care of self (dialog)
      SEEK ALLTRIM(::GetTitle())
      IF !EOF() 
         IF !EMPTY((Fname)->BASE)
            ::SetTitle( ALLTRIM((Fname)->BASE) )
         ENDIF
      ENDIF
      IF ::MenuBar():NumItems() > 0
         XlateMenu(::MenuBar(),.T.)  // force untrans         
      ENDIF
   ENDIF      
   FOR I = 1 TO LEN(::aCaptionObjs)
      IF ::aCaptionObjs[I]:IsDerivedFrom("XbpDialog")  // special handling XbpDialog child
         SEEK ::aCaptionObjs[I]:GetTitle()
         IF !EOF() 
            IF !EMPTY((Fname)->BASE)
               ::aCaptionObjs[I]:SetTitle( ALLTRIM((Fname)->BASE) )
            ENDIF
         ENDIF
         LOOP
      ENDIF   
      IF ::aCaptionObjs[I]:IsDerivedFrom("XbpMultiCellGroup")  // special handling for XbpQuickBrowse
         FOR J = 1 TO ::aCaptionObjs[I]:MaxCol
            IF ValType(::aCaptionObjs[I]:GetCell(1,J))="C"   // should be the heading cell and we only want charcter type
               (Fname)->(DBSEEK(::aCaptionObjs[I]:GetCell(1,J))) 
               IF !(Fname)->(EOF())
                  IF !EMPTY((Fname)->BASE)
                     ::aCaptionObjs[I]:SetCell(1,J,ALLTRIM((Fname)->BASE))
                  ENDIF
               ENDIF   
            ENDIF    
         NEXT I
         LOOP
      ENDIF
      IF ::aCaptionObjs[I]:IsDerivedFrom("XbpCellGroup") // special for XbpBrowse
         (Fname)->(DBSEEK( ::aCaptionObjs[I]:GetCell(1)))
      ELSE
         (Fname)->(DBSEEK( ::aCaptionObjs[I]:Caption))  //Everything else should have a caption
      ENDIF   
      IF !(Fname)->(EOF())
         IF !EMPTY((Fname)->BASE)
            IF ::aCaptionObjs[I]:IsDerivedFrom("XbpCellGroup")
               IF ::aCaptionObjs[I]:type = XBPCOL_TYPE_TEXT
                  ::aCaptionObjs[I]:SetCell(1,ALLTRIM((Fname)->BASE))
               ENDIF
            ELSE         
               ::aCaptionObjs[I]:SetCaption( ALLTRIM((Fname)->BASE) )
            ENDIF   
         ENDIF
      ENDIF   
   NEXT I
   ::IsTrans := .F.
   IF lHadToOpen
      (Fname)->(DBCloseArea()) 
   ELSE
      (Fname)->(OrdSetFocus(PrevOrd))   
   ENDIF   
   SELECT (PrevArea) 
   ::InvalidateRect()
RETURN Self

METHOD JimsDialog:SetCenter()
  ::SetPos({INT((::SetParent():currentSize()[1]-::currentSize()[1])/2),INT((::SetParent():currentSize()[2]-::currentSize()[2])/2)})  
RETURN Self

// End Class Code

// Start Functions

FUNCTION GetCaptionObj(o)
LOCAL aRtn:={},aTemp,I,aSub,J
   aTemp := o:ChildList()
   FOR I = 1 TO LEN(aTemp)
      DO CASE
         CASE aTemp[I]:IsDerivedFrom("XbpStatic").AND. aTemp[I]:ClassName()!="XbpBrowse" .AND. !aTemp[I]:IsDerivedFrom("XbpBrowse").AND. aTemp[I]:ClassName()!="XbpQuickBrowse" .AND. !aTemp[I]:IsDerivedFrom("XbpQuickBrowse")
            AADD(aRtn,aTemp[I])
            aSub := GetCaptionObj(aTemp[I])
            FOR J := 1 TO LEN(aSub)
               AADD(aRtn,aSub[J])
            NEXT J
         CASE aTemp[I]:IsDerivedFrom("Xbp3State")
            AADD(aRtn,aTemp[I])
         CASE aTemp[I]:IsDerivedFrom("XbpCheckBox")
            AADD(aRtn,aTemp[I])
         CASE aTemp[I]:IsDerivedFrom("XbpPushButton")
            AADD(aRtn,aTemp[I])
         CASE aTemp[I]:IsDerivedFrom("XbpRadioButton")
            AADD(aRtn,aTemp[I])
         CASE aTemp[I]:IsDerivedFrom("XbpTabPage")
            AADD(aRtn,aTemp[I])
            aSub := GetCaptionObj(aTemp[I])
            FOR J := 1 TO LEN(aSub)
               AADD(aRtn,aSub[J])
            NEXT J
         CASE aTemp[I]:IsDerivedFrom("XbpDialog")
            AADD(aRtn,aTemp[I])
            aSub := GetCaptionObj(aTemp[I]:drawingArea)
            FOR J := 1 TO LEN(aSub)
               AADD(aRtn,aSub[J])
            NEXT J 
            // get any menu items 
         CASE aTemp[I]:IsDerivedFrom("XbpBrowse").OR. aTemp[I]:ClassName() = "XbpBrowse"
            FOR J = 1 TO aTemp[I]:ColCount
               IF !EMPTY(aTemp[I]:GetColumn(J):heading:GetCell(1))
                  AADD(aRtn,aTemp[I]:GetColumn(J):heading)
               ENDIF             
               IF !EMPTY(aTemp[I]:GetColumn(J):Footing:GetCell(1))
                  AADD(aRtn,aTemp[I]:GetColumn(J):Footing)
               ENDIF             
            NEXT J
         CASE aTemp[I]:IsDerivedFrom("XbpQuickBrowse")
            IF !EMPTY(aTemp[I]:heading)
               AADD(aRtn,aTemp[I]:heading)
            ENDIF             
         CASE aTemp[I]:IsDerivedFrom("AppBrowse")
            aSub := GetCaptionObj(aTemp[I])
            FOR J := 1 TO LEN(aSub)
               AADD(aRtn,aSub[J])
            NEXT J
         CASE aTemp[I]:IsDerivedFrom("AppEdit")
            aSub := GetCaptionObj(aTemp[I])
            FOR J := 1 TO LEN(aSub)
               AADD(aRtn,aSub[J])
            NEXT J
      ENDCASE
   NEXT I
RETURN aRtn 


FUNCTION SetXlateDB(cXlate)
   IF cXlate = NIL
      cXlate := ""
   ENDIF 
   LanguageDB := cXlate 
RETURN NIL 

STATIC FUNCTION OpenXlateDB(LangDB)
LOCAL PrevArea,Fname,Fname2,lHadToOpen := .F.,aRtn[2]
      Fname := IF(".DBF"$ UPPER(LangDB),LEFT(LangDB,len(LanguageDB)-4),LangDB)
      Fname2 := Fname+IF(ordbagext()=".NTX","2","")
      IF SELECT(Fname) = 0
         IF !FILE(LangDB+IF(".DBF"$UPPER(LangDB),"",".DBF"))
            DbCreate(LangDB,{{"BASE","C",MAXBASE,0},{"TRANSLATE","C",MAXTRANSLATE,0}}) 
         ENDIF
         DBuseArea(.T.,,Fname,Fname,.T.)           
         lHadToOpen := .T.
      ENDIF
      SELECT (Fname)
      IF FILE(Fname+OrdBagExT()) 
         IF lHadToOpen
            IF ordbagext()=".NTX"
               SET INDEX TO (Fname), (Fname2)
            ELSE
               SET INDEX TO (Fname)
            ENDIF   
         ENDIF
      ELSE
         DBCloseArea()
         DBuseArea(.T.,,Fname,Fname,.F.)
         INDEX ON BASE TAG BASE TO (Fname)
         INDEX ON TRANSLATE TAG TRANS TO (Fname2)
         DBCloseArea()
         DBuseArea(.T.,,Fname,Fname,.T.)
         IF ordbagext()=".NTX"
            SET INDEX TO (Fname), (Fname2)
         ELSE
            SET INDEX TO (Fname)
         ENDIF   
      ENDIF
      aRtn[1] := Fname
      aRtn[2] := lHadToOpen
RETURN aRtn

FUNCTION XlateMenu(oMenu,lUntrans)
LOCAL PrevArea,Fname,cRtn,I,aTemp,lHadToOpen := .F.,PrevOrd,cStr
   IF lUntrans = NIL
      lUntrans := .F.
   ENDIF   
   IF !EMPTY(LanguageDB)
      PrevArea := SELECT()
      aTemp := OpenXlateDB(LanguageDB)
      Fname := aTemp[1]
      lHadToOpen := aTemp[2]
      SELECT (Fname)
      IF !lUntrans
         PrevOrd := (Fname)->(Ordsetfocus("BASE"))
      ELSE
         PrevOrd := (Fname)->(Ordsetfocus("TRANS"))
      ENDIF   
      FOR I = 1 TO oMenu:NumItems()
         cStr := oMenu:getItem(I)[1] // get first item from item array
         IF ValType(cStr) = "O" // must be a XbpMenu 
            XlateMenu(cStr,lUntrans) // recurse
            LOOP   
         ENDIF
         IF ValType(cStr) = "C"
            IF oMenu:IsDerivedFrom("XbpMenu") .AND. I = 1
               (Fname)->(DBSeek(ALLTRIM(oMenu:Title))) 
               IF !(Fname)->(EOF())
                  IF !lUntrans
                     IF !EMPTY((Fname)->TRANSLATE)
                        oMenu:SetTitle(ALLTRIM((Fname)->TRANSLATE))
                     ENDIF
                  ELSE
                     IF !EMPTY((Fname)->BASE)
                        oMenu:SetTitle(ALLTRIM((Fname)->BASE))
                     ENDIF
                  ENDIF
               ELSEIF !lUntrans
                  (Fname)->(DBAPPEND())
                  REPLACE (Fname)->BASE WITH oMenu:Title
                  (Fname)->(DBUNLOCK())
               ENDIF
            ENDIF 
            
            (Fname)->(DBSeek(ALLTRIM(cStr))) 
            IF !(Fname)->(EOF())
               IF !lUntrans
                  IF !EMPTY((Fname)->TRANSLATE)
                     aTemp := oMenu:getItem(I)
                     aTemp[1] := ALLTRIM((Fname)->TRANSLATE)
                     oMenu:SetItem(I,aTemp)
                  ENDIF
               ELSE
                  IF !EMPTY((Fname)->BASE)
                     aTemp := oMenu:getItem(I)
                     aTemp[1] := ALLTRIM((Fname)->BASE)
                     oMenu:SetItem(I,aTemp)
                  ENDIF
               ENDIF
            ELSEIF !lUntrans
               (Fname)->(DBAPPEND())
               REPLACE (Fname)->BASE WITH cStr
               (Fname)->(DBUNLOCK())
            ENDIF
         ENDIF
      NEXT I
      IF lHadToOpen
         (Fname)->(DBCloseArea()) 
      ELSE
         (Fname)->(OrdSetFocus(PrevOrd))   
      ENDIF   
      SELECT (PrevArea)
   ENDIF
RETURN NIL

FUNCTION XlateStr(cStr)
LOCAL PrevArea,Fname,Fname2,cRtn,lHadToOpen := .F.,PrevOrd,aTemp
   cRtn := cStr
   IF !EMPTY(LanguageDB) .AND. !EMPTY(cStr)
      PrevArea := SELECT()
      aTemp := OpenXlateDB(LanguageDB)
      Fname := aTemp[1]
      lHadToOpen := aTemp[2]
      PrevOrd := Ordsetfocus("BASE")
      (Fname)->(DBSeek(ALLTRIM(cStr))) 
      IF !(Fname)->(EOF())
         IF !EMPTY((Fname)->TRANSLATE)
            cRtn := ALLTRIM((Fname)->TRANSLATE)
         ENDIF
      ELSE
         (Fname)->(DBAPPEND())
         REPLACE (Fname)->BASE WITH cStr
         (Fname)->(DBUNLOCK())
      ENDIF
      IF lHadToOpen
         (Fname)->(DBCloseArea())
      ELSE 
         (Fname)->(OrdSetFocus(PrevOrd))  
      ENDIF   
      SELECT (PrevArea)
   ENDIF
RETURN cRtn

FUNCTION XlateMsgBox(cMessage,cTitle)
RETURN MsgBox(XlateStr(cMessage),XlateStr(cTitle))

FUNCTION XlateConfirmBox(oOwner,cMessage,cTitle,nButtons,nStylenStartBtn )
RETURN ConfirmBox(oOwner,XlateStr(cMessage),XlateStr(cTitle),nButtons,nStylenStartBtn )


// Simple browse 
CLASS JimsBrowse FROM XbpBrowse
   EXPORTED:
      METHOD Notify,Init,Skipper,RefreshAll,Wheel,KeyBoard,create,HandleEvent,destroy
      VAR DBArea,NotifyToggal,IsStable,FocusChange,AutoUpdate,AutoThread
ENDCLASS 

METHOD JimsBrowse:Init(oParent, oOwner, aPos, aSize, aPP, lVisible)
   ::XbpBrowse:Init(oParent, oOwner, aPos, aSize, aPP, lVisible)
   ::DBArea := Alias() // set to current area as default
   ::NotifyToggal := .T.  //turn on notify action
   //Set default codeblocks
   // Navigation code blocks for the browser    
   ::skipBlock     := {|n| ::Skipper(n) } 
   ::goTopBlock    := {| | (::DBarea)->(DbSuspendNotifications()),::NotifyToggal:=.F.,(::DBArea)->(DbGoTop()),::NotifyToggal:=.T. ,(::DBarea)->(DbResetNotifications())   } 
   ::goBottomBlock := {| | (::DBarea)->(DbSuspendNotifications()),::NotifyToggal:=.F.,(::DBArea)->(DbGoBottom()),::NotifyToggal:=.T. ,(::DBarea)->(DbResetNotifications()) } 

   ::phyPosBlock   := {| | (::DBArea)->(Recno())      } 

   // Navigation code blocks for the vertical scroll bar 
   ::posBlock      := {| | (::DBArea)->(DbPosition())    } 
   ::goPosBlock    := {|n| ::NotifyToggal:=.F.,(::DBArea)->( DbGoPosition(n)),::NotifyToggal:=.T. } 
   ::lastPosBlock  := {| | 100       } 
   ::firstPosBlock := {| | 1               } 
   ::StableBlock   := {| | ::IsStable := .T.} 
   ::hitTopBlock   := {| | ::Gotop()}
   ::IsStable := .T. 
   ::AutoUpdate := .F.
RETURN Self 

METHOD JimsBrowse:Create(oParent, oOwner, aPos, aSize, aPP, lVisible)
   ::XbpBrowse:Create(oParent, oOwner, aPos, aSize, aPP, lVisible)
   IF ::AutoUpdate
      ::AutoThread := BrowseThread():New() 
      ::AutoThread:oBrowse := Self
      ::AutoThread:Start()
   ENDIF   
       
RETURN Self

METHOD JimsBrowse:HandleEvent(nEvent,mp1,mp2)
   IF nEvent = xbeM_Wheel
      ::Wheel(mp1,mp2)
   ENDIF
   IF nEvent = REFRESH_EVENT
      ::RefreshAll()
   ELSE
      ::XbpBrowse:HandleEvent(nEvent,mp1,mp2)
   ENDIF    
RETURN SELF
   
METHOD JimsBrowse:RefreshAll()
   (::DBarea)->(DbSuspendNotifications())
   ::NotifyToggal := .F.
   ::XbpBrowse:RefreshAll()
   ::NotifyToggal := .T.  
   (::DBarea)->(DbResetNotifications())
RETURN Self

METHOD JimsBrowse:Notify(nEvent,mp1,mp2)     
   (::DBarea)->(DbSuspendNotifications())
   DO CASE
      CASE mp1 == DBO_MOVE_DONE .OR. ;       // skip is done
           mp1 == DBO_GOBOTTOM  .OR. ;
           mp1 == DBO_TABLE_UPDATE .OR.;
           mp1 == DBO_TABLE_DELETED  .OR. ;
           mp1 == DBO_TABLE_APPEND .OR.;
           mp1 == DBO_GOTOP
         IF ::NotifyToggal
            ::RefreshAll()           
            ::ForceStable()
         ENDIF   
   ENDCASE 
   (::DBarea)->(DbResetNotifications())   
RETURN Self

METHOD JimsBrowse:Skipper(n)
LOCAL Rtn,lDidtog:=.F.,I
   IF ::NotifyToggal
      ::NotifyToggal := .F.
      lDidTog := .T.         
   ENDIF   
   (::DBarea)->(DbSuspendNotifications())
   Rtn := (::DBArea)->(DbSkipper(n))
   IF (::DBArea)->(BOF())
      Rtn := 0
   ENDIF  
   IF (::DBArea)->(EOF())
      Rtn := 0
   ENDIF
   (::DBarea)->(DbResetNotifications())
   IF lDidTog
      ::NotifyToggal := .T.
   ENDIF   
RETURN Rtn

/*METHOD JimsBrowse:Wheel(aPos,aWheel)
   SetAppFocus(Self)    
   ::XbpBrowse:Wheel(aPos,aWheel)
   ::XbpBrowse:RefreshAll()
RETURN Self  */

METHOD JimsBrowse:Wheel(aPos,aWheel)
  ::XbpBrowse:Wheel(aPos,aWheel)
  IF aWheel[2] > 0 .AND. ::RowPos <= 5 .AND. (::DBArea)->(DbPosition())<=1
     ::XbpBrowse:RefreshAll()
  ENDIF 
RETURN Self

METHOD JimsBrowse:Keyboard(nKey)
   DO CASE
      CASE nKey = xbeK_ALT_E .AND. ::FocusChange!=NIL
         SetAppFocus(::FocusChange)
      OTHERWISE
         ::XbpBrowse:Keyboard(nKey)
   ENDCASE

   IF nKey = xbeK_UP .AND. ::RowPos <= 2 .AND. (::DBArea)->(DbPosition())<=1
      PostAppEvent(xbeP_Keyboard,xbeK_CTRL_PGUP,,Self)
   ENDIF 
           
RETURN Self

METHOD JimsBrowse:destroy()
   IF !EMPTY(::AutoThread)
      ::AutoThread:terminate()
   ENDIF
   ::XbpBrowse:destroy()
RETURN Self      

CLASS BrowseThread FROM Thread 
   PROTECTED: 
      VAR terminated 
   EXPORTED: 
      VAR workCounter,oBrowse 

      INLINE METHOD init 
         ::Thread:init() 
         ::terminated  := .F. 
         ::workCounter := 0 
      RETURN self 

      INLINE METHOD terminate 

      RETURN ( ::terminated := .T. ) 

      METHOD execute, checkTermination 
ENDCLASS       

// This method is executed in the thread 
METHOD BrowseThread:execute 
   DO WHILE .T.
      // program code for thread 
      SLEEP(3000)
      ::checkTermination() 
      IF !::terminated
         PostAppEvent(REFRESH_EVENT,,,::oBrowse)
      ENDIF   
   ENDDO 
RETURN self 

** This method checks whether or not the thread must 
** terminate and cancels it, if necessary 

METHOD BrowseThread:checkTermination 
   IF ::terminated 
      ::terminated := .F. 
      ::quit() 
   ENDIF 
RETURN self 

PROCEDURE Set_ViewMode(o,lVal,oDefFocus)// note pass drawingarea 
LOCAL I,aList 
   aList := o:ChildList()
   IF lVal 
      FOR I = 1 TO LEN(aList)
         DO CASE
            CASE aList[I]:IsDerivedFrom("SLEPic")
               aList[I]:Protect(.T.,oDefFocus)
            CASE aList[I]:IsDerivedFrom("XbpCheckBox")
               aList[I]:SetInputFocus := {|x,y,o|SetAppFocus(oDefFocus)}
            CASE aList[I]:IsDerivedFrom("XbpMLE")
               aList[I]:SetInputFocus := {|x,y,o|SetAppFocus(oDefFocus)}
            CASE aList[I]:IsDerivedFrom("XbpComboBox")
               aList[I]:SetInputFocus := {|x,y,o|SetAppFocus(oDefFocus)}
            CASE aList[I]:IsDerivedFrom("XbpRadioButton")
               aList[I]:SetInputFocus := {|x,y,o|SetAppFocus(oDefFocus)}
            CASE aList[I]:IsDerivedFrom("XbpStatic")
               Set_ViewMode(aList[I],lVal,oDefFocus)
         ENDCASE   
      NEXT I
   ELSE
      FOR I = 1 TO LEN(aList)
         DO CASE
            CASE aList[I]:IsDerivedFrom("SLEPic")
               aList[I]:Protect(.F.)
            CASE aList[I]:IsDerivedFrom("XbpCheckBox")
               aList[I]:SetInputFocus := NIL
            CASE aList[I]:IsDerivedFrom("XbpMLE")
               aList[I]:SetInputFocus := NIL
            CASE aList[I]:IsDerivedFrom("XbpComboBox")
               aList[I]:SetInputFocus := NIL
            CASE aList[I]:IsDerivedFrom("XbpRadioButton")
               aList[I]:SetInputFocus := NIL
             CASE aList[I]:IsDerivedFrom("XbpStatic")
               Set_ViewMode(aList[I],lVal)
        ENDCASE   
      NEXT I
   ENDIF
RETURN


PROCEDURE SetClipperStyle(o)  // note pass drawingarea
LOCAL I 
   FOR I = 1 TO LEN(o:ChildList())
      IF o:ChildList()[I]:IsDerivedFrom("XbpCheckBox")
         o:ChildList()[I]:KeyBoard := {|nKey,x,o| SLEMove(nKey,o)}
         o:ChildList()[I]:setFontCompoundName("9.Arial bold") 
      ENDIF  
      IF o:ChildList()[I]:IsDerivedFrom("XbpPushButton")
         o:ChildList()[I]:KeyBoard := {|nKey,x,o| ButtonMove(nKey,o)} 
//         o:ChildList()[I]:setFontCompoundName("6.Arial bold") 
      ENDIF  
      IF o:ChildList()[I]:IsDerivedFrom("XbpRadioButton")
         o:ChildList()[I]:KeyBoard := {|nKey,x,o| SLEMove(nKey,o)} 
         o:ChildList()[I]:setFontCompoundName("9.Arial bold") 
      ENDIF  
      IF o:ChildList()[I]:IsDerivedFrom("SLEPic")
         o:ChildList()[I]:XbpSLE:setFontCompoundName("10.Lucida Console") 
      ENDIF  
      IF o:ChildList()[I]:IsDerivedFrom("XbpMLE")
         o:ChildList()[I]:setFontCompoundName("10.Lucida Console") 
      ENDIF  
      IF o:ChildList()[I]:IsDerivedFrom("SLEPopUp")
//         o:ChildList()[I]:oPush:setFontCompoundName("9.Arial bold") 
      ENDIF  
      IF o:ChildList()[I]:IsDerivedFrom("XbpListBox")
         o:ChildList()[I]:setFontCompoundName("10.Lucida Console") 
      ENDIF  
      IF o:ChildList()[I]:IsDerivedFrom("ComboBoxSLE")
         o:ChildList()[I]:XbpListBox:setFontCompoundName("10.Lucida Console") 
//         o:ChildList()[I]:oPush:setFontCompoundName("9.Arial bold") 
      ENDIF  
      IF o:ChildList()[I]:IsDerivedFrom("XbpBrowse")
         o:setFontCompoundName("10.Lucida Console") 
      ENDIF  
      IF o:ChildList()[I]:IsDerivedFrom("XbpStatic")
         // It's a static it may have children that are checkboxes so recurse
         o:ChildList()[I]:setFontCompoundName("9.Arial bold") 
         SetCLipperStyle(o:ChildList()[I])
      ENDIF  
      IF o:ChildList()[I]:IsDerivedFrom("XbpTabPage")
         // It's a TABPAGE it may have children that are checkboxes so recurse
         o:ChildList()[I]:setFontCompoundName("9.Arial bold") 
         SetCLipperStyle(o:ChildList()[I])
      ENDIF  
   NEXT I    
RETURN 

// Set datalinks by use of DB_ALIAS_FIELD  naming convention
Function SetDataLinks(o,cPub,Helpbase,HelpPefix)  // dialog obj ie self
LOCAL I,J,K,L,aIvars,cAlias,cField,bLink,aRtn:={},cType:="" 
   aIvars := o:ClassDescribe(CLASS_DESCR_MEMBERS) 
   FOR I = 1 TO LEN(aIvars)
      cAlias := ""
      cField := ""
      cType:="" 
      IF LEFT(aIvars[I,CLASS_MEMBER_NAME],3) = "DB_"
         FOR J = 4 TO LEN(aIvars[I,CLASS_MEMBER_NAME])
            IF aIvars[I,CLASS_MEMBER_NAME][J] = "_"
               EXIT
            ENDIF
            cAlias += aIvars[I,CLASS_MEMBER_NAME][J]    
         NEXT J
         FOR K = J+1 TO LEN(aIvars[I,CLASS_MEMBER_NAME])
            IF aIvars[I,CLASS_MEMBER_NAME][K] = "_"
               EXIT
            ENDIF
            cField += aIvars[I,CLASS_MEMBER_NAME][K]
         NEXT K   
         FOR L = K+1 TO LEN(aIvars[I,CLASS_MEMBER_NAME])
            IF aIvars[I,CLASS_MEMBER_NAME][L] = "_"
               EXIT
            ENDIF
            cTYPE += UPPER(aIvars[I,CLASS_MEMBER_NAME][L])
         NEXT L  
         cAlias := STRTRAN(cAlias,"0",cPub)
         IF LEN(cType) > 1 
            cType := LEFT(cType,1)
         ENDIF
         IF !EMPTY(cAlias) .AND. !EMPTY(cField) 
          /*ADVTSR	CCNO	ADVCODE      fields to encrypt
            CLADVT	CCNO	PHONE
            GLADCJ	CCNO	CODE
            GLCLCJ	CCNO	CODE
            GLMICJ	CCNO	CODE
            GLMECJ	CCNO	CODE
            GLSUCJ	CCNO	CODE 
            GLCARD   CCNO  ID
            SUBIMP	CCNO	ID 
            SULST?	CCNO	ID 
            SUVEND   CCNO  VENCODE*/
            DO CASE
               CASE  cField = "CCNO" 
                  DO CASE
                     CASE cAlias $ "ADVTSR"
                        bLink := &("{|x| IF(x=NIL,DECRYPT("+cAlias+"->"+cField+","+cAlias+"->ADVCODE"+") , "+cAlias+"->"+cField+" := ENCRYPT(x,"+cAlias+"->ADVCODE"+"))}") 
                     CASE cAlias $ "CLADVT"
                        bLink := &("{|x| IF(x=NIL,DECRYPT("+cAlias+"->"+cField+","+cAlias+"->PHONE"+") , "+cAlias+"->"+cField+" := ENCRYPT(x,"+cAlias+"->PHONE"+"))}") 
                     CASE cAlias $ "GLADCJ GLCLCJ GLMICJ GLSUCJ GLMECJ"
                        bLink := &("{|x| IF(x=NIL,DECRYPT("+cAlias+"->"+cField+","+cAlias+"->CODE"+") , "+cAlias+"->"+cField+" := ENCRYPT(x,"+cAlias+"->CODE"+"))}") 
                     CASE LEFT(cAlias,5) $ "SULST GLCARD SUBIMP"
                        bLink := &("{|x| IF(x=NIL,DECRYPT("+cAlias+"->"+cField+","+cAlias+"->ID"+") , "+cAlias+"->"+cField+" := ENCRYPT(x,"+cAlias+"->ID"+"))}") 
                     CASE cAlias $ "SUVEND"
                        bLink := &("{|x| IF(x=NIL,DECRYPT("+cAlias+"->"+cField+","+cAlias+"->VENCODE"+") , "+cAlias+"->"+cField+" := ENCRYPT(x,"+cAlias+"->VENCODE"+"))}") 
                     OTHERWISE
                        bLink := &("{|x| IF(x=NIL,"+cAlias+"->"+cField+" , "+cAlias+"->"+cField+" := x)}") 
                  ENDCASE      
               OTHERWISE
                  bLink := &("{|x| IF(x=NIL,"+IF(cType="C","TRIM(","")+cAlias+"->"+cField+IF(cType="C",")","")+" , "+cAlias+"->"+cField+" := x)}") 
            ENDCASE   
            o:&(aIvars[I,CLASS_MEMBER_NAME]):DataLink := bLink  
            AADD(aRtn , {o:&(aIvars[I,CLASS_MEMBER_NAME]),aIvars[I,CLASS_MEMBER_NAME]})
         ENDIF
      ENDIF
//      o:&(aIvars[I,CLASS_MEMBER_NAME]):helpLink := MagicHelpLabel():New(HelpBase+I,HelpPrefix+aIvars[I,CLASS_MEMBER_NAME]) 
   NEXT I    
RETURN aRtn

FUNCTION ParentDialog(o)
LOCAL oParent:=NIL
    DO WHILE EMPTY(oParent)
   IF o = NIL 
      MSGBOX("NIL")
   ENDIF 
       IF !o:IsDerivedFrom("XbpDialog")
          IF o:SetParent():IsDerivedFrom("XbpDialog")
             oParent := o:SetParent()
          ENDIF
       ELSE
          oParent := o
       ENDIF
       o := o:SetParent()  
    ENDDO 
RETURN oParent

FUNCTION Encrypt(cPlain,cKey)
LOCAL nKey := LEN(cKey),i,j,cRtn:=""
   IF !EMPTY(cPlain)
      FOR i = 1 TO LEN(cPlain)
           IF (ASC(cPlain[i])+ASC(cKey[MOD(i,nKey)+1])) > 127
                cRtn += CHR((ASC(cPlain[i])+ASC(cKey[MOD(i,nKey)+1])) -95)   //95 = 127-32
           ELSE
                cRtn += CHR((ASC(cPlain[i])+ASC(cKey[MOD(i,nKey)+1])) )
           ENDIF
      NEXT i
   ENDIF
RETURN cRtn

FUNCTION Decrypt(cEncript,cKey)
LOCAL nKey := LEN(cKey),i,j,cRtn:=""
   IF !EMPTY(cEncript)
      FOR i = 1 TO LEN(cEncript)
           IF (ASC(cEncript[i])-ASC(cKey[MOD(i,nKey)+1])) < 32
                cRtn += CHR((ASC(cEncript[i])-ASC(cKey[MOD(i,nKey)+1])) +95)// 95 = 127-32
           ELSE
                cRtn += CHR((ASC(cEncript[i])-ASC(cKey[MOD(i,nKey)+1])) )
           ENDIF
      NEXT i
   ENDIF
RETURN cRtn

#ifdef FANCYBUTTON
   #define XbpPushButton ShadowButton
#endif   

PROCEDURE MPMMSGBOX(cMSG,cTitle)
LOCAL oDlg,aMsg:={},I,J,oBut,oMsg,X:=0,Y:=45,H,W,oDummy
   AADD(aMsg,"")
   J := 1
   FOR I = 1 TO LEN(cMsg)
      IF cMsg[I] = CHR(13)
         AADD(aMsg,"")
         ++J  
         LOOP
      ENDIF
      aMsg[J] += cMsg[I]   
   NEXT I 
   oDlg := MSGDialog():New(oMainWindow:DrawingArea,SetAppWindow())
//   oDlg := MSGDialog():New(AppDeskTop(),SetAppWindow())
   oDlg:AlwaysOnTop := .T.  
   oDlg:Border := XBPDLG_NO_BORDER 
   IF VALTYPE(cTitle) = "C" 
      oDlg:Title := cTitle
   ENDIF  
   oDlg:Create()
   oDlg:DrawingArea:SetColorBG(GRA_CLR_WHITE)
   oDlg:DrawingArea:SetColorFG(GRA_CLR_RED)
   // calc size
   FOR I = LEN(aMsg) TO 1 Step -1
      oMsg := XbpStatic():New(oDlg:DrawingArea,,{10,Y})
      oMsg:AutoSize := .T.
      oMsg:Type := XBPSTATIC_TYPE_TEXT 
      oMsg:Options := XBPSTATIC_TEXT_CENTER+XBPSTATIC_TEXT_VCENTER
      oMsg:Caption  := aMsg[I]
      oMsg:Create()
      X := MAX(oMsg:CurrentSize()[1],X)
      Y += oMsg:CurrentSize()[2]+3
   NEXT I
   X := MAX(X,150)
   Y := MAX(Y,80)
   oDlg:SetPosAndSize({(oMainWindow:DrawingArea:CurrentSize()[1]-X-20)/2,(oMainWindow:DrawingArea:CurrentSize()[2]-Y-30)/2},{X+20,Y+35})
   oDummy := XbpSLE():New(oDlg:drawingArea,,{-30,-30},{1,1}):Create()
   oBut := XbpPushButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()[1]-75)/2,3},{75,25})
   oBut:Caption := "OK" 
   oBut:Activate := {||oDlg:Destroy()} 
   oBut:Create()
   oBut:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut)}  
   IF oMainWindow != SetAppWindow()
      oDlg:SetModalState(XBP_DISP_APPMODAL) 
   ENDIF   
   oDlg:KEY1 := {xbeK_ALT_O,oBut:Activate}
   SetAppFocus(oBut)
   IF FILE("MSGBOXSOUND.WAV")
      DllCall("WINMM.DLL",DLL_STDCALL,"PlaySoundA","MSGBOXSOUND.WAV",0, SND_FILENAME + SND_ASYNC)
   ENDIF   
   oDlg:DrawingArea:InvalidateRect()
   ModalLoop(oDlg)
//   WindowsLoop(oDlg)
   SetAppFocus(SetAppWindow())

RETURN


FUNCTION MPMCONFIRMBOX(oOwner,cMSG,cTitle,nButtons,nStyle,nStartBtn)
LOCAL oDlg,aMsg:={},I,J,oBut1,oBut2,oBut3,oMsg,X:=0,Y:=40,H,W,XOffset:=0,Icon,nRtn := 0,oFocus,oDummy
   IF oOwner = NIL
      oOwner := SetAppWindow()
   ENDIF
   IF nStartBtn = NIL
      nStartBtn := XBPMB_DEFBUTTON1
   ENDIF
   IF nButtons = NIL
      nButtons := XBPMB_OKCANCEL
   ENDIF
   AADD(aMsg,"")
   J := 1
   FOR I = 1 TO LEN(cMsg)
      IF cMsg[I] = CHR(13)
         AADD(aMsg,"")
         ++J  
         LOOP
      ENDIF
      aMsg[J] += cMsg[I]   
   NEXT I   
   // icon selection nstyle
   DO CASE
      CASE BAnd(nStyle,XBPMB_QUESTION)!=0 
         Icon := XBPSTATIC_SYSICON_ICONQUESTION
         XOffset := 35
      CASE BAnd(nStyle,XBPMB_WARNING)!=0
         Icon := XBPSTATIC_SYSICON_ICONWARNING
         XOffset := 35
      CASE BAnd(nStyle,XBPMB_INFORMATION)!=0
         Icon := XBPSTATIC_SYSICON_ICONINFORMATION
         XOffset := 35
      CASE BAnd(nStyle,XBPMB_CRITICAL)!=0
         Icon := XBPSTATIC_SYSICON_ICONERROR
         XOffset := 35
      OTHERWISE  // no icon
         Icon :=NIL
         XOffset := 15
   ENDCASE
   oDlg := MSGDialog():New(oMainWindow:DrawingArea,oOwner)
//   oDlg := MSGDialog():New(AppDeskTop(),oOwner)
   oDlg:AlwaysOnTop := .T.  
   oDlg:Border := XBPDLG_NO_BORDER 
   IF VALTYPE(cTitle) = "C" 
      oDlg:Title := cTitle
   ENDIF
   oDlg:Create()
   oDlg:DrawingArea:SetColorBG(GRA_CLR_WHITE)
   oDlg:DrawingArea:SetColorFG(GRA_CLR_DARKGREEN)
   // calc size
   FOR I = LEN(aMsg) TO 1 Step -1
      oMsg := XbpStatic():New(oDlg:DrawingArea,,{10+Xoffset,Y})
      oMsg:AutoSize := .T.
      oMsg:Type := XBPSTATIC_TYPE_TEXT 
      oMsg:Options := XBPSTATIC_TEXT_CENTER+XBPSTATIC_TEXT_VCENTER
      oMsg:Caption  := aMsg[I]
      oMsg:Create()
      X := MAX(oMsg:CurrentSize()[1],X)
      Y += oMsg:CurrentSize()[2]+3
   NEXT I
   DO CASE
      CASE nButtons = XBPMB_OK 
         X := MAX(X,150)
      CASE nButtons = XBPMB_OKCANCEL .OR. nButtons = XBPMB_RETRYCANCEL .OR. nButtons = XBPMB_YESNO   // two butons
         X := MAX(X,250)
      CASE nButtons = XBPMB_ABORTRETRYIGNORE .OR. nButtons = XBPMB_YESNOCANCEL    // three butons
         X := MAX(X,300)
      OTHERWISE
         X := MAX(X,250)
         nButtons := XBPMB_OKCANCEL    // default
   ENDCASE      
   Y := MAX(Y,80) 
   IF Icon != NIL
      oStat := XbpStatic():New(oDlg:drawingArea,,{2,(Y+20)/2},{30,30})
      oStat:Type := XBPSTATIC_TYPE_SYSICON
      oStat:Caption := Icon
      oStat:Create()
   ENDIF
   oDlg:SetPosAndSize({(oMainWindow:DrawingArea:CurrentSize()[1]-X-20)/2,(oMainWindow:DrawingArea:CurrentSize()[2]-Y-30)/2},{X+20+XOffset,Y+35})
   oDummy := XbpSLE():New(oDlg:drawingArea,,{-30,-30},{1,1}):Create()
   DO CASE 
      CASE nButtons = XBPMB_OK
         oBut1 := MSGButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()[1]-75)/2,3},{75,25})
         oBut1:Caption := "OK" 
         oBut1:Activate := {||nRtn := XBPMB_RET_OK,oDlg:Destroy()} 
         oBut1:TabStop  := .T. 
         oBut1:Create()
         oBut1:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut1)}  
         IF nStartBtn = XBPMB_DEFBUTTON1
            oFocus := oBut1
         ENDIF 
         oDlg:KEY1 := {{xbeK_ALT_O,ASC("O"),ASC("o")},oBut1:Activate}
         
      CASE nButtons = XBPMB_OKCANCEL 
         oBut1 := MSGButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()[1]/3-75/2),3},{75,25})
         oBut1:Caption := "OK" 
         oBut1:Activate := {||nRtn := XBPMB_RET_OK,oDlg:Destroy()} 
         oBut1:TabStop  := .T. 
         oBut1:Create()
         oBut1:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut1)}  
         IF nStartBtn = XBPMB_DEFBUTTON1
            oFocus := oBut1
         ENDIF
         oBut2 := MSGButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()[1]*2/3-75/2),3},{75,25})
         oBut2:Caption := "Cancel" 
         oBut2:TabStop  := .T. 
         oBut2:Activate := {||nRtn := XBPMB_RET_CANCEL,oDlg:Destroy()} 
         oBut2:Create()
         oBut2:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut2)}  
         IF nStartBtn = XBPMB_DEFBUTTON2
            oFocus := oBut2
         ENDIF
         oDlg:KEY1 := {{xbeK_ALT_O,ASC("O"),ASC("o")},oBut1:Activate}
         oDlg:KEY2 := {{xbeK_ALT_C,ASC("C"),ASC("c")},oBut2:Activate}
      CASE nButtons = XBPMB_YESNO
         oBut1 := MSGButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()[1]/3-75/2),3},{75,25})
         oBut1:Caption := "YES" 
         oBut1:Activate := {||nRtn := XBPMB_RET_YES,oDlg:Destroy()} 
         oBut1:TabStop  := .T. 
         oBut1:Create()
         oBut1:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut1)}  
         IF nStartBtn = XBPMB_DEFBUTTON1
            oFocus := oBut1
         ENDIF
         oBut2 := MSGButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()[1]*2/3-75/2),3},{75,25})
         oBut2:Caption := "No" 
         oBut2:Activate := {||nRtn := XBPMB_RET_NO,oDlg:Destroy()} 
         oBut2:TabStop  := .T. 
         oBut2:Create()
         oBut2:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut2)}  
         IF nStartBtn = XBPMB_DEFBUTTON2
            oFocus := oBut2
         ENDIF
         oDlg:KEY1 := {{xbeK_ALT_Y,ASC("Y"),ASC("y")},oBut1:Activate}
         oDlg:KEY2 := {{xbeK_ALT_N,ASC("N"),ASC("n")},oBut2:Activate}
      CASE nButtons = XBPMB_RETRYCANCEL    
         oBut1 := MSGButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()/3[1]-75/2),3},{75,25})
         oBut1:Caption := "Retry" 
         oBut1:Activate := {||nRtn := XBPMB_RET_RETRY,oDlg:Destroy()} 
         oBut1:TabStop  := .T. 
         oBut1:Create()
         oBut1:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut1)}  
         IF nStartBtn = XBPMB_DEFBUTTON1
            oFocus := oBut1
         ENDIF
         oBut2 := MSGButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()[1]*2/3-75/2),3},{75,25})
         oBut2:Caption := "Cancel" 
         oBut2:Activate := {||nRtn := XBPMB_RET_CANCEL,oDlg:Destroy()} 
         oBut2:TabStop  := .T. 
         oBut2:Create()
         oBut2:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut2)}  
         IF nStartBtn = XBPMB_DEFBUTTON2
            oFocus := oBut2
         ENDIF
         oDlg:KEY1 := {{xbeK_ALT_R,ASC("R"),ASC("r")},oBut1:Activate}
         oDlg:KEY2 := {{xbeK_ALT_C,ASC("C"),ASC("c")},oBut2:Activate}
      CASE nButtons = XBPMB_ABORTRETRYIGNORE 
         oBut1 := MSGButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()[1]/4-75/2),3},{75,25})
         oBut1:Caption := "Abort" 
         oBut1:Activate := {||nRtn := XBPMB_RET_ABORT,oDlg:Destroy()} 
         oBut1:TabStop  := .T. 
         oBut1:Create()
         oBut1:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut1)}  
         IF nStartBtn = XBPMB_DEFBUTTON1
            oFocus := oBut1
         ENDIF
         oBut2 := MSGButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()[1]-75)/2,3},{75,25})
         oBut2:Caption := "Retry" 
         oBut2:Activate := {||nRtn := XBPMB_RET_RETRY,oDlg:Destroy()} 
         oBut2:TabStop  := .T. 
         oBut2:Create()
         oBut2:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut2)}  
         IF nStartBtn = XBPMB_DEFBUTTON2
            oFocus := oBut2
         ENDIF
         oBut3 := MSGButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()[1]*3/4-75/2),3},{75,25})
         oBut3:Caption := "Ignore" 
         oBut3:Activate := {||nRtn := XBPMB_RET_IGNORE,oDlg:Destroy()} 
         oBut3:TabStop  := .T. 
         oBut3:Create()
         oBut3:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut3)}  
         IF nStartBtn = XBPMB_DEFBUTTON3
            oFocus := oBut3
         ENDIF
         oDlg:KEY1 := {{xbeK_ALT_A,ASC("A"),ASC("a")},oBut1:Activate}
         oDlg:KEY2 := {{xbeK_ALT_R,ASC("R"),ASC("r")},oBut2:Activate}
         oDlg:KEY3 := {{xbeK_ALT_I,ASC("I"),ASC("i")},oBut2:Activate}
      CASE nButtons = XBPMB_YESNOCANCEL    
         oBut1 := MSGButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()[1]/4-75/2),3},{75,25})
         oBut1:Caption := "Yes" 
         oBut1:Activate := {||nRtn := XBPMB_RET_YES,oDlg:Destroy()} 
         oBut1:TabStop  := .T. 
         oBut1:Create()
         oBut1:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut1)}  
         IF nStartBtn = XBPMB_DEFBUTTON1
            oFocus := oBut1
         ENDIF
         oBut2 := MSGButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()[1]-75)/2,3},{75,25})
         oBut2:Caption := "No" 
         oBut2:Activate := {||nRtn := XBPMB_RET_NO,oDlg:Destroy()} 
         oBut2:TabStop  := .T. 
         oBut2:Create()
         oBut2:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut2)}  
         IF nStartBtn = XBPMB_DEFBUTTON2
            oFocus := oBut2
         ENDIF
         oBut3 := MSGButton():New(oDlg:DrawingArea,,{(oDlg:DrawingArea:CurrentSize()[1]*3/4-75/2),3},{75,25})
         oBut3:Caption := "Cancel" 
         oBut3:Activate := {||nRtn := XBPMB_RET_CANCEL,oDlg:Destroy()} 
         oBut3:TabStop  := .T. 
         oBut3:Create()
         oBut3:KeyBoard := {|nKey,x,o| ButtonMove2(nKey,oBut3)}  
         IF nStartBtn = XBPMB_DEFBUTTON3
            oFocus := oBut3
         ENDIF
         oDlg:KEY1 := {{xbeK_ALT_Y,ASC("Y"),ASC("y")},oBut1:Activate}
         oDlg:KEY2 := {{xbeK_ALT_N,ASC("N"),ASC("n")},oBut2:Activate}
         oDlg:KEY3 := {{xbeK_ALT_C,ASC("C"),ASC("c")},oBut3:Activate}
   ENDCASE
   
   IF oFocus != NIL
      SetAppFocus(oFocus)
   ELSE
      SetAppFocus(oBut)
   ENDIF
//   IF oMainWindow != SetAppWindow()
      oDlg:SetModalState(XBP_DISP_APPMODAL) 
//   ENDIF   
   IF FILE("CONFIRMBOXSOUND.WAV")
      DllCall("WINMM.DLL",DLL_STDCALL,"PlaySoundA","CONFIRMBOXSOUND.WAV",0, SND_FILENAME + SND_ASYNC)
   ENDIF   
   oDlg:DrawingArea:InvalidateRect() 
//   PostAppEvent(xbeP_Paint,{0,0,0,0},,oDlg:DrawingArea)
   ModalLoop(oDlg)
//   WindowsLoop(oDlg)
   SetAppFocus(SetAppWindow())

RETURN nRtn

CLASS MSGDialog FROM XbpDialog  // mange keypress for msgbox/confirmbox objects
EXPORTED: 
   VAR KEY1,KEY2,KEY3
   INLINE METHOD KeyBoard(nKey)
      DO CASE 
         CASE ::KEY1 != NIL .AND. nKey $ ::KEY1[1]
             Eval(::KEY1[2])
         CASE ::KEY2 != NIL .AND. nKey $ ::KEY2[1]
             Eval(::KEY2[2])
         CASE ::KEY3 != NIL .AND. nKey $ ::KEY3[1]
             Eval(::KEY3[2])
         OTHERWISE
             ::XbpDialog:KeyBoard(nKey)
      ENDCASE
   RETURN Self 
ENDCLASS

CLASS MSGButton FROM XbpPushButton  // mange keypress for msgbox/confirmbox objects
EXPORTED: 
   VAR KEY1,KEY2,KEY3
   INLINE METHOD KeyBoard(nKey)
      DO CASE 
         CASE ::SetParent():SetParent():KEY1 != NIL .AND. nKey $ ::SetParent():SetParent():KEY1[1]
             Eval(::SetParent():SetParent():KEY1[2])
         CASE ::SetParent():SetParent():KEY2 != NIL .AND. nKey $ ::SetParent():SetParent():KEY2[1]
             Eval(::SetParent():SetParent():KEY2[2])
         CASE ::SetParent():SetParent():KEY3 != NIL .AND. nKey $ ::SetParent():SetParent():KEY3[1]
             Eval(::SetParent():SetParent():KEY3[2])
         OTHERWISE
             ::XbpPushButton:KeyBoard(nKey)
      ENDCASE
   RETURN Self 
ENDCLASS

PROCEDURE ModalLoop(oSel)
LOCAL nEvent,mp1,mp2,oXbp
   DO WHILE oSel:Status() = XBP_STAT_CREATE
      nEvent := AppEvent( @mp1, @mp2, @oXbp )
      oXbp:HandleEvent( nEvent, mp1, mp2 )
   ENDDO
RETURN


//EOF
/////