/*
         PRINTIT.PRG
*/

/*
   METRIC FUNCTIONS:

   FUNCTION MetSay(nTop,cVar,cPict,oPS,oFont,nTab,nPct,lNewLine) --> nTop
   FUNCTION MetCenter(nTop,cVar,oPS,oFont,nPct,lNewLine,nStart,nWidth) --> nTop
   FUNCTION MetRight(nTop,cVar,nVal,cPict,oPS,oFont,nTab,nPct,lNewLine) --> nTop
   FUNCTION MetPad(nTop,cVar,cPict,oPS,oFont,nTab,nRight,cPad,nPct,lNewLine) --> nTop
   FUNCTION MetLine(nTop,nPct,oFont,nLines) --> nTop
   FUNCTION MetUL(nTop,cChar,cPict,oPS,oFont,nTab,nPct,lNewLine) --> nTop
   FUNCTION PageCheck(nTop,oPS,oFont,nLines,nPct,nLimit) --> lNewPage
   FUNCTION MetPic(cFile,aPos,aTargetSize,oPS) --> NIL
   FUNCTION MetAge(nTop,aVal,oPS,oFont) --> nTop

   PAGE DIMENSIONS: NOTE: 1" = 25.4 mm or 254 mm/10 (the standard unit of measurement)
      Provided by the PaperSize method of the printer object. PaperSize returns
      an 8-element array:
         1 Paper width   in 1/10 mm (= nXsize)
         2 Paper height  in 1/10 mm (= nYsize)
         3 Left margin   in 1/10 mm (= nLeftMargin)
         4 Bottom margin in 1/10 mm (= nBottomMargin)
         5 Right margin  in 1/10 mm (= nRightMargin)
         6 Top margin    in 1/10 mm (= nTopMargin)
         7 Width in pixels
         8 Height in pixels
         The four margin settings define a non-printable area around the edges of
         the paper, so the available width is nXsize-nLeftMargin-nRightMargin, and
         the available height is nYsize-nTopMargin-nBottomMargin. These are set as
         static variables by the StartPrinter() function and can be called with the
         SetTopMargin(), SetBottomMargin(), SetLeftMargin() and SetRightMargin()
         functions.

   PAGE COORDINATES:
      The 0,0 point is at the bottom left corner of the page. The X coordinate
      increases from left to right. The Y coordinate DECREASES from top to bottom.

   COMMON PARAMETERS:
      nTop     = Position of the top of the line to be printed in 1/10 mm. Must always
                 be passed from the calling program. The starting value for nTop in the
                 calling program is typically nTopMargin. Remember to reset nTop after
                 every Printer_Eject.
      cVar     = Variable to be printed (can be any variable type, except in MetRight()
                 where it must be a character variable).
      cPict    = A picture clause to be used with a numeric variable; defaults to
                 "999,999.99"
      oPS      = The default presentation space, defined in StartPrinter().
      oFont    = The font to be used for the specific line; defaults to the last font
                 specified.
      nTab     = The horizontal starting position of the line in INCHES from the
                 left margin.
      nPct     = The amount of extra space left below the line being printed expressed
                 as a percentage of the current line (e.g. 50 would allow an additional
                 half line of space, 100 would allow a full additional line);
                 defaults to 0, or the standard amount of space allowed by the font
                 specification.
      lNewLine = If .T., returns the Y position (nTop) of the next line to be printed,
                 otherwise returns the Y position of the current line; default varies
                 by function.

   SPECIFIC FUNCTIONS:

   FUNCTION MetSay(nTop,cVar,cPict,oPS,oFont,nTab,nPct,lNewLine) --> nTop
      Standard function for left-to-right printing, starting at nTab INCHES from the
         left margin.
      lNewLine defaults .F.

   FUNCTION MetCenter(nTop,cVar,oPS,oFont,nPct,lNewLine,nStart,nWidth) --> nTop
      Centers a line in a defined space. nStart and nWidth can be used to define an
         imaginary column within which the printed line will be centered, where nStart
         = the starting horizontal position of the column in INCHES from left edge of
         the page, and nWidth = the width in INCHES of the column.
      If nStart and nWidth aren't defined, the line is centered on the full page.
      There's no picture function included, so numerics should be sent as strings or
         transforms; otherwise they'll be displayed as unformatted strings.
      lNewLine defaults .T.

   FUNCTION MetRight(nTop,cVar,nVal,cPict,oPS,oFont,nTab,nPct,lNewLine) --> nTop
      Prints a right-justified line ending at nTab. Required when aligning columns
         of figures using a proportional font.
      cVar is a character string the prints to the left of the numeric value contained
         in nVal. If there's no nVal, MetRight() prints the right-justified text
         contained in cVar.
      nTab defaults to the right edge of the page.
      lNewLine defaults .T.

   FUNCTION MetPad(nTop,cVar,cPict,oPS,oFont,nTab,nRight,cPad,nPct,lNewLine) --> nTop
      Prints a left-justified line starting at nTab and padded with a specified
         character to a fixed right-hand point.
      cPad is the padding character; defaults to "."
      nRight is the right-hand end of the line and must be specified in INCHES
      lNewLine defaults .F.

   FUNCTION MetLine(nTop,nPct,oFont,nLines) --> nTop
      Skips nLines blank lines and returns the Y value of the next line to be printed.
      nLines defaults to 1.

   FUNCTION MetUL(nTop,cChar,cPict,oPS,oFont,nTab,nPct,lNewLine) --> nTop
      Used with MetRight() to print an underscore below a column of figures to be
         followed by a total.
      cPict defines the width of the underscore.
      cChar is the character to be used for the underscore; defaults to "-".
      nTab is the right edge of the underscore; defaults to the right edge of the page.
      lNewLine defaults .T.

   FUNCTION PageCheck(nTop,oPS,oFont,nLines,nPct,nLimit) --> lNewPage
      Determines is sufficient space remains available on the page to print nLines
         at the specified point size and leading.
      nLines defaults to 1.
      nLimit defines the number of lines that must remain available after nLines are
         printed; defaults to 0.
      Returns .T. if sufficient space is NOT available and a new page is required.

   FUNCTION MetPic(cFile,aPos,aTargetSize,oPS) --> NIL
      Prints the graphic specified by cFile in an imaginary box (the print area)
         defined by aTargetSize at the coordinates specified in aPos.
      aPos = X,Y of upper left corner of print area in 1/10 mm
      aTargetSize = Width,Depth of print area in INCHES
      According to Alaska, this should work with any standard graphic file, but I've
         only been able to make it work with .JPGs.

   FUNCTION MetAge(nTop,aVal,oPS,oFont) --> nTop
      A specialized function to print the aging lines at the bottom of a bill.
      aVal contains the values to be printed for current, 30, 60, 90 and 120+ days.
      The default picture for all values is "999,999.99"
      Returns the Y position of the next line.

   nTab CONVERSIONS FOR DOS PROGRAMS USING A FIXED WIDTH (Courier) FONT
      12-point type (specified by PV10P1) prints 10 characters per inch (CPI):
         Multiply the X position by 0.1. If the DOS command was @ LN,45 SAY...,
         set nTab to 4.5.
      10-point type (specified by PV12P1) prints 12 CPI:
         Multiply the X position by 0.0833. If the DOS command was @ LN,45 SAY...,
         set nTab to 3.75.
       8-point type (specified by PV17P1) actually prints 15 CPI:
         Multiply the X position by 0.0667. If the DOS command was @ LN,45 SAY...,
         set nTab to 3.

*/

#include "Gra.ch"
#include "Xbp.ch"
#include "Common.ch"
#include "Appevent.ch"
#INCLUDE "font.ch"
#include "XBPDEV.CH"

#define SPACECHR  "L"
//#define LINESPACING 5
#define LINESPACING 12
#define WFACTOR 1
//#define HFACTOR 0.82
#define HFACTOR 1
STATIC oCurrFont
STATIC oCurrRefFont
STATIC oCurroPS
STATIC aLastPos
STATIC PrnAdjust := 0
STATIC SegPage
STATIC oPreview
STATIC oDlgPreview
STATIC aPaperSize
STATIC RtnScreen
STATIC PaperLand := .F.
// used with metric functions
STATIC nRightMargin := 0  // right margin
STATIC nLeftMargin := 0  // left margin
STATIC nTopMargin := 0  // top margin
STATIC nBottomMargin := 0  // bottom margin

// Get/Set functions
FUNCTION SetRightMargin(nVal)
LOCAL nRtn := nRightMargin
   IF nVal != NIL
      nRightMargin := nVal
   ENDIF
RETURN nRtn

FUNCTION SetTopMargin(nVal)
LOCAL nRtn := nTopMargin
   IF nVal != NIL
      nTopMargin := nVal
   ENDIF
RETURN nRtn

FUNCTION SetLeftMargin(nVal)
LOCAL nRtn := nLeftMargin
   IF nVal != NIL
      nLeftMargin := nVal
   ENDIF
RETURN nRtn

FUNCTION SetBottomMargin(nVal)
LOCAL nRtn := nBottomMargin
   IF nVal != NIL
      nBottomMargin := nVal
   ENDIF
RETURN nRtn

FUNCTION SetFont(oFont)
LOCAL oRtn := oCurrFont
   IF oFont != NIL
      oCurrFont := oFont
   ENDIF
RETURN oRtn

FUNCTION SetReference(oFont)
LOCAL oRtn := oCurrRefFont
   IF oFont != NIL
      oCurrRefFont := oFont
   ENDIF
RETURN oRtn

FUNCTION SetoPS(oPS)
LOCAL oRtn := oCurroPS
   IF oPS != NIL
      oCurroPS := oPS
   ENDIF
RETURN oRtn

FUNCTION SetAdjust()
LOCAL cText,nRtn := PrnAdjust
   IF PrnAdjust = 0 .AND. FILE("PRNADJUST.TXT")
      cText := MEMOREAD("PRNADJUST.TXT")
      PrnAdjust := VAL(Alltrim(cText))
   ENDIF
RETURN nRtn

FUNCTION PrnVar(aPos,cVar,oPS,oFont,oRefFont) // note aPos is in Y,X order to ease mod of @say syntex
LOCAL aTopRight,a_size, aPrint,aSize2

      SetAdjust()
      IF oFont = NIL
         oFont := oCurrFont
      ELSE
//         oCurrFont := oFont
      ENDIF
      IF PaperLand
         aPos[1]+=4
         aPos[2]+=5
      ENDIF

      IF oRefFont = NIL
         oRefFont := oCurrRefFont
      ELSE
         oCurrRefFont := oRefFont
      ENDIF
      IF oPS = NIL
         oPS := oCurroPS
      ELSE
         oCurroPS := oPS
      ENDIF
      IF oRefFont = NIL .OR. oFont = NIL .OR. oPS = NIL
         MSGBOX("ERROR no defined font in printit! "+STR(aPos[1],3)+','+STR(aPos[2],3))
         RETURN NIL
      ENDIF
      DO CASE
        CASE VALTYPE(cVar)=="N"
          cVar := STR(cVar)
        CASE VALTYPE(cVar)=="D"
          cVar := DTOC(cVar)
        CASE VALTYPE(cVar)=="C"
        CASE VALTYPE(cVar)=="L"
          IF cVar
             cVar := "Y"
          ELSE
             cVar := "N"
          ENDIF
        OTHERWISE
          MSGBOX("ERROR incorrect data type in printit! "+STR(aPos[1],3)+','+STR(aPos[2],3))
      ENDCASE
      GraSetFont(oPS , oRefFont )                // reference font for spacing
      IF !oRefFont:Fixed
         a_Size := GraQueryTextBox( oPS, SPACECHR )  //spaceing based on 'L'
      ELSE
         a_Size := {oRefFont:Width * WFACTOR,oRefFont:Height * HFACTOR}
      ENDIF
      GraSetFont(oPS , oFont )                // Select font

      aTopRight :={aPaperSize[1]-(aPaperSize[3]+(aPaperSize[1]-aPaperSize[5])),aPaperSize[2] - (aPaperSize[4]+(aPaperSize[2]-aPaperSize[6]))}
      IF !oRefFont:Fixed
         aPrint := {aPos[2]*a_Size[3,1],aTopRight[2]-(aPos[1]*(a_Size[1,2]+LINESPACING+PrnAdjust))}
      ELSE
         aPrint := {aPos[2]*a_Size[1],aTopRight[2]-(aPos[1]*(a_Size[2]+PrnAdjust))}
      ENDIF
          // Print string
      GraStringAt( oPS, aPrint, cVar )
      oCurrRefFont := oRefFont
//      oCurrFont := oFont

RETURN NIL

// Metric left-justified line
FUNCTION MetSay(nTop,cVar,cPict,oPS,oFont,nTab,nPct,lNewLine)
LOCAL aPoints, aPrint, nLeft, nDepth, nBottom, nLead
LOCAL nRtn := nTop

      IF oFont = NIL
         oFont := oCurrFont
      ELSE
         oCurrFont := oFont
      ENDIF

      IF oPS = NIL
         oPS := oCurroPS
      ELSE
         oCurroPS := oPS
      ENDIF

      IF oFont = NIL .OR. oPS = NIL
         MSGBOX("ERROR no defined font in MetSay "+LTRIM(STR(nTop,4)))
         RETURN nRtn
      ENDIF

      IF cPict = NIL
         cPict := "999,999.99"
      ENDIF

      IF nTab = NIL
         nTab := 0
      ENDIF

      IF nPct = NIL
         nPct := 0 // 0% default leading
      ENDIF

      IF lNewLine = NIL
         lNewLine := .F.
      ENDIF

      DO CASE
         CASE VALTYPE(cVar)=="N"
            cVar := TRANSFORM(cVar,cPict)
         CASE VALTYPE(cVar)=="D"
            cVar := SHORTDATE(cVar)
         CASE VALTYPE(cVar)=="C"
         CASE VALTYPE(cVar)=="L"
            IF cVar
               cVar := "Y"
            ELSE
               cVar := "N"
            ENDIF
         OTHERWISE
//            MPMMSGBOX("ERROR incorrect data type in MetSay "+LTRIM(STR(nTop,4)))
            cVar := '>>>DATA-TYPE ERROR<<<'
//            RETURN nRtn
      ENDCASE
      GraSetFont(oPS , oFont )  // reference font for spacing and printing

      aPoints := GraQueryTextBox( oPS, cVar )  //size of actual string

      nLead := ROUND((nPct/100)*(aPoints[3,2]-aPoints[2,2]),0)

      nDepth     := aPoints[1,2]-aPoints[2,2]
      nLeft      := nLeftMargin+nTab*254
      nBottom    := nTop-nDepth
      aPrint     := {nLeft,nBottom}
      // Print string
      GraStringAt( oPS, aPrint, cVar )

      nLead := ROUND((nPct/100)*nDepth,0)

      IF lNewLine
         nRtn := nTop-(nDepth+nLead) //returns new Y coordinate for next line
      ELSE
         nRtn := nTop //stays on the same line
      ENDIF

RETURN nRtn

// Metric line feed proportional to the font in use
FUNCTION MetLine(nTop,nPct,oFont,nLines)
LOCAL aPoints, nDepth, nBottom, nLead, oPS := oCurroPS
LOCAL nRtn := nTop

      IF oFont = NIL
         oFont := oCurrFont
      ELSE
         oCurrFont := oFont
      ENDIF

      IF oFont = NIL .OR. oPS = NIL
         MSGBOX("ERROR no defined font in MetLine "+LTRIM(STR(nTop,4)))
         RETURN nRtn
      ENDIF

      IF nLines = NIL
         nLines := 1
      ENDIF

      IF nPct = NIL
         nPct := 0 // 0% default leading
      ENDIF

      GraSetFont(oPS , oFont )  // reference font for spacing and printing

      aPoints := GraQueryTextBox( oPS, 'X' )  //size of actual string

      nLead := ROUND((nPct/100)*(aPoints[3,2]-aPoints[2,2]),0)

      nDepth    := aPoints[1,2]-aPoints[2,2]
      nBottom   := nTop-nDepth

      nLead := ROUND((nPct/100)*nDepth,0)

      nRtn := nTop-nLines*(nDepth+nLead) //returns new Y coordinate for next line

RETURN nRtn

// Metric centered line
FUNCTION MetCenter(nTop,cVar,oPS,oFont,nPct,lNewLine,nStart,nWidth)
LOCAL aPoints, aPrint, nLeft, nDepth, nBottom, nStrWidth, nPageWidth, nLead
LOCAL nRtn := nTop

      IF oFont = NIL
         oFont := oCurrFont
      ENDIF

      IF oPS = NIL
         oPS := oCurroPS
      ELSE
         oCurroPS := oPS
      ENDIF

      IF oFont = NIL .OR. oPS = NIL
         MSGBOX("ERROR no defined font in MetCenter "+LTRIM(STR(nTop,4)))
         RETURN nRtn
      ENDIF

      IF lNewLine = NIL
         lNewLine := .T.
      ENDIF

      IF nStart = NIL
         nStart := nLeftMargin
      ELSE
         nStart := nStart*254
      ENDIF

      nPageWidth := nRightMargin-nLeftMargin

      IF nWidth = NIL
         nWidth := nPageWidth
      ELSE
         nWidth := nWidth*254
      ENDIF

      IF nPct = NIL
         nPct := 0 // 0% default leading
      ENDIF

      DO CASE
         CASE VALTYPE(cVar)=="N"
            cVar := ALLTRIM(STR(cVar)) // since it's centered, we don't want extra space
         CASE VALTYPE(cVar)=="D"
            cVar := SHORTDATE(cVar)
         CASE VALTYPE(cVar)=="C"
//            cVar := ALLTRIM(cVar) // since it's centered, we don't want extra space
         CASE VALTYPE(cVar)=="L"
            IF cVar
               cVar := "Y"
            ELSE
               cVar := "N"
            ENDIF
         OTHERWISE
//            MPMMSGBOX("ERROR incorrect data type in MetCenter "+LTRIM(STR(nTop,4)))
            cVar := '>>>DATA-TYPE ERROR<<<'
//            RETURN nRtn
      ENDCASE
      GraSetFont(oPS , oFont )  // reference font for spacing and printing

      aPoints := GraQueryTextBox( oPS, cVar )  //size of actual string

      nLead := ROUND((nPct/100)*(aPoints[3,2]-aPoints[2,2]),0)

      nStrWidth  := aPoints[3,1]-aPoints[2,1] //lower left to upper right to allow for italics
      nDepth     := aPoints[1,2]-aPoints[2,2]
      nLeft      := nStart+ROUND((nWidth-nStrWidth)/2,0)
      nBottom    := nTop-nDepth
      aPrint     := {nLeft,nBottom}
      // Print string
      GraStringAt( oPS, aPrint, cVar )

      IF lNewLine
         nRtn := nTop-(nDepth+nLead) //returns new Y coordinate for next line
      ELSE
         nRtn := nTop //stays on the same line
      ENDIF

RETURN nRtn

// Metric left-justified line padded with leader characters to fixed right hand point
FUNCTION MetPad(nTop,cVar,cPict,oPS,oFont,nTab,nRight,cPad,nPct,lNewLine)
LOCAL aPoints, aPrint, nLeft, nDepth, nBottom, nLead
LOCAL nRtn := nTop

      IF oFont = NIL
         oFont := oCurrFont
      ELSE
         oCurrFont := oFont
      ENDIF

      IF oPS = NIL
         oPS := oCurroPS
      ELSE
         oCurroPS := oPS
      ENDIF

      IF oFont = NIL .OR. oPS = NIL
         MSGBOX("ERROR no defined font in MetPad "+LTRIM(STR(nTop,4)))
         RETURN nRtn
      ENDIF

      IF nRight = NIL
         MSGBOX("ERROR no right boundary defined in MetPad "+LTRIM(STR(nTop,4)))
         RETURN nRtn
      ENDIF

      IF cPict = NIL
         cPict := "999,999.99"
      ENDIF

      IF nTab = NIL
         nTab := 0
      ENDIF

      IF nPct = NIL
         nPct := 0 // 0% default leading
      ENDIF

      IF lNewLine = NIL
         lNewLine := .F.
      ENDIF

      IF cPad = NIL
         cPad := '.'
      ENDIF

      DO CASE
         CASE VALTYPE(cVar)=="N"
            cVar := TRANSFORM(cVar,cPict)
         CASE VALTYPE(cVar)=="D"
            cVar := SHORTDATE(cVar)
         CASE VALTYPE(cVar)=="C"
         CASE VALTYPE(cVar)=="L"
            IF cVar
               cVar := "Y"
            ELSE
               cVar := "N"
            ENDIF
         OTHERWISE
//            MPMMSGBOX("ERROR incorrect data type in MetPad "+LTRIM(STR(nTop,4)))
            cVar := '>>>DATA-TYPE ERROR<<<'
//            RETURN nRtn
      ENDCASE
      GraSetFont(oPS , oFont )  // reference font for spacing and printing

      aPoints := GraQueryTextBox( oPS, cVar )  //size of actual string
      DO WHILE aPoints[3,1]-aPoints[2,1]<=(nRight-nTab)*254
         cVar += cPad
         aPoints := GraQueryTextBox( oPS, cVar )  //size of actual string
      ENDDO

      nLead := ROUND((nPct/100)*(aPoints[3,2]-aPoints[2,2]),0)

      nDepth     := aPoints[1,2]-aPoints[2,2]
      nLeft      := nLeftMargin+nTab*254
      nBottom    := nTop-nDepth
      aPrint     := {nLeft,nBottom}
      // Print string
      GraStringAt( oPS, aPrint, cVar )

      nLead := ROUND((nPct/100)*nDepth,0)

      IF lNewLine
         nRtn := nTop-(nDepth+nLead) //returns new Y coordinate for next line
      ELSE
         nRtn := nTop //stays on the same line
      ENDIF

RETURN nRtn

// Metric right-justified line -- used for aligning columns with proportional fonts
FUNCTION MetRight(nTop,cVar,nVal,cPict,oPS,oFont,nTab,nPct,lNewLine)
LOCAL aPoints, aPrint, nLeft, nDepth, nBottom, nStrWidth, nPageWidth, nLead
LOCAL nRtn := nTop, cValue

      IF oFont = NIL
         oFont := oCurrFont
      ENDIF

      IF oPS = NIL
         oPS := oCurroPS
      ELSE
         oCurroPS := oPS
      ENDIF

      IF oFont = NIL .OR. oPS = NIL
         MSGBOX("ERROR no defined font in MetRight "+LTRIM(STR(nTop,4)))
         RETURN nRtn
      ENDIF

      IF cPict = NIL
         IF nVal = NIL
            cPict := ""
         ELSE
            cPict := "999,999.99"
         ENDIF
      ENDIF

      IF nVal = NIL
         cValue := ''
      ELSE
         cValue := ALLTRIM(TRANSFORM(nVal,cPict))
      ENDIF

      IF cVar = NIL
         cVar := ''
      ENDIF

      IF lNewLine = NIL
         lNewLine := .T.
      ENDIF

      IF nTab = NIL
         nTab := nRightMargin
      ELSE
         nTab := nTab*254
      ENDIF

      nPageWidth := nRightMargin-nLeftMargin

      IF nPct = NIL
         nPct := 0 // 0% default leading
      ENDIF

      IF VALTYPE(cVar) != "C" .OR. (VALTYPE(nVal) != 'N' .AND. nVal != NIL)
//         MPMMSGBOX("ERROR incorrect data type in MetRight "+LTRIM(STR(nTop,4)))
         cVar := '>>>DATA-TYPE ERROR<<<'
//         RETURN nRtn
      ENDIF

      GraSetFont(oPS , oFont )  // reference font for spacing and printing

      IF EMPTY(cValue)
         aPoints := GraQueryTextBox( oPS, cPict )  //size of actual string
      ELSE
         aPoints := GraQueryTextBox( oPS, cValue )  //size of actual string
      ENDIF

      nLead := ROUND((nPct/100)*(aPoints[3,2]-aPoints[2,2]),0)

      nStrWidth := aPoints[3,1]-aPoints[2,1] //lower left to upper right to allow for italics
      nDepth    := aPoints[1,2]-aPoints[2,2]
      nLeft     := nTab-nStrWidth
      nBottom   := nTop-nDepth
      aPrint    := {nLeft,nBottom}
      // Print string
      IF .NOT. EMPTY(cValue)
         GraStringAt( oPS, aPrint, cValue )
      ENDIF

      IF .NOT. EMPTY(cVar)
         IF EMPTY(cPict)
            aPoints   := GraQueryTextBox( oPS, cVar )  //size of actual string
         ELSE
            aPoints   := GraQueryTextBox( oPS, cVar+' '+cPict )  //size of actual string
         ENDIF
         nStrWidth := aPoints[3,1]-aPoints[2,1] //lower left to upper right to allow for italics
         nLeft     := MAX(0,nTab-nStrWidth)
         aPrint    := {nLeft,nBottom}
         GraStringAt( oPS, aPrint, cVar )
      ENDIF

      IF lNewLine
         nRtn := nTop-(nDepth+nLead) //returns new Y coordinate for next line
      ELSE
         nRtn := nTop //stays on the same line
      ENDIF

RETURN nRtn

// Metric underline -- for undelining before a total
FUNCTION MetUL(nTop,cChar,cPict,oPS,oFont,nTab,nPct,lNewLine)
LOCAL aPoints, aPrint, nLeft, nDepth, nBottom, nStrWidth, nPageWidth, nLead
LOCAL nRtn := nTop, nTarget, cVar

      IF oFont = NIL
         oFont := oCurrFont
      ENDIF

      IF oPS = NIL
         oPS := oCurroPS
      ELSE
         oCurroPS := oPS
      ENDIF

      IF oFont = NIL .OR. oPS = NIL
         MSGBOX("ERROR no defined font in MetUL "+LTRIM(STR(nTop,4)))
         RETURN nRtn
      ENDIF

      IF cPict = NIL
         cPict := "999,999.99"
      ENDIF

      IF lNewLine = NIL
         lNewLine := .T.
      ENDIF

      IF cChar = NIL
         cChar := '-'
      ENDIF

      IF nTab = NIL
         nTab := nRightMargin
      ELSE
         nTab := nTab*254
      ENDIF

      nPageWidth := nRightMargin-nLeftMargin

      IF nPct = NIL
         nPct := 0 // 0% default leading
      ENDIF

      GraSetFont(oPS , oFont )  // reference font for spacing and printing

      aPoints := GraQueryTextBox( oPS, cPict )  //size of actual string

      nLead := ROUND((nPct/100)*(aPoints[3,2]-aPoints[2,2]),0)

      nTarget := aPoints[3,1]-aPoints[2,1] //lower left to upper right to allow for italics
      nDepth  := aPoints[1,2]-aPoints[2,2]

      cVar := REPLICATE(cChar,LEN(cPict)-2)
      nStrWidth := 0
      DO WHILE nStrWidth < nTarget
         cVar += cChar
         aPoints   := GraQueryTextBox( oPS, cVar )  //size of actual string
         nStrWidth := aPoints[3,1]-aPoints[2,1] //lower left to upper right to allow for italics
      ENDDO

      nLeft    := MAX(0,nTab-nStrWidth)
      nBottom  := nTop-nDepth
      aPrint   := {nLeft,nBottom}
      // Print string
      GraStringAt( oPS, aPrint, cVar )

      IF lNewLine
         nRtn := nTop-(nDepth+nLead) //returns new Y coordinate for next line
      ELSE
         nRtn := nTop //stays on the same line
      ENDIF

RETURN nRtn

// Check available page depth for metric printing
FUNCTION PageCheck(nTop,oPS,oFont,nLines,nPct,nLimit)
LOCAL aPoints, nDepth, nBottom, nLead
LOCAL lRtn := .F.

      IF oFont = NIL
         oFont := oCurrFont
      ENDIF

      IF oPS = NIL
         oPS := oCurroPS
      ELSE
         oCurroPS := oPS
      ENDIF

      IF oFont = NIL .OR. oPS = NIL
         MSGBOX("ERROR no defined font in PageCheck "+LTRIM(STR(nTop,4)))
         RETURN lRtn
      ENDIF

      IF nLines = NIL
         nLines := 1
      ENDIF

      GraSetFont(oPS , oFont )  // reference font for spacing and printing

      aPoints := GraQueryTextBox( oPS, 'XXX' ) // dummy character string

      IF nPct = NIL
         nLead := 0
      ELSE
         nLead := ROUND((nPct/100)*(aPoints[3,2]-aPoints[2,2]),0)
      ENDIF

      nDepth     := aPoints[1,2]-aPoints[2,2]
      nBottom    := nTop-(nLines*nDepth+(nLines-1)*nLead)

      IF nLimit = NIL
         nLimit := nBottomMargin
      ELSE
         nLimit := nBottomMargin+nLimit*nDepth+(nLimit-1)*nLead
      ENDIF

      IF nBottom<nLimit
         lRtn := .T. // a new page is required
      ENDIF

RETURN lRtn

// Picture function for metric printing
FUNCTION MetPic(cFile,aPos,aTargetSize,oPS)
// aPos = X,Y of upper left corner in 1/10 mm
// aTargetSize = Width,Depth in inches
LOCAL oBitmap,aRect,a_Size,aTopRight,lTest,oRefFont
      IF oPS = NIL
         oPS := oCurroPS
      ELSE
         oCurroPS := oPS
      ENDIF

      oBitmap := XbpBitmap():new():create( oPS )
      aRect   := { aPos[1], aPos[2]-aTargetSize[2]*254, aPos[1]+aTargetSize[1]*254, aPos[2] }
      oBitmap:loadFile( cFile )          // Load bitmap
      lTest := oBitmap:draw( oPS, aRect )  // Spool print job
      oBitmap:destroy()
RETURN NIL

// Metric aging block at bottom of bills
FUNCTION MetAge(nTop,aVal,oPS,oFont)
LOCAL aPoints, aPrint, nLeft, nDepth, nBottom, nStrWidth, nPageWidth
LOCAL nRtn := nTop, nX[5], cCheckStr, I, nXbar, nXunit

      IF oFont = NIL
         oFont := oCurrFont
      ENDIF

      IF oPS = NIL
         oPS := oCurroPS
      ELSE
         oCurroPS := oPS
      ENDIF

      IF oFont = NIL .OR. oPS = NIL
         MSGBOX("ERROR no defined font in MetAge "+LTRIM(STR(nTop,4)))
         RETURN nRtn
      ENDIF

      nPageWidth := nRightMargin-nLeftMargin

      GraSetFont(oPS , oFont )  // reference font for spacing and printing

      aPoints := GraQueryTextBox( oPS, '999,999.99' )
      nXunit  := aPoints[3,1]-aPoints[2,1] //width of a single element

      aPoints := GraQueryTextBox( oPS, ' | ' )
      nXbar   := aPoints[3,1]-aPoints[2,1] //width of the spacer

      cCheckStr := '999,999.99 | 999,999.99 | 999,999,999 | 999,999.99 | 999,999.99'
      aPoints   := GraQueryTextBox( oPS, cCheckStr )  //size of full string
      nStrWidth := aPoints[3,1]-aPoints[2,1] //lower left to upper right to allow for italics
      nDepth    := aPoints[1,2]-aPoints[2,2]
      nLeft     := ROUND((nPageWidth-nStrWidth)/2,0)
      nBottom   := nTop-nDepth

      // Calculate the right edge of each element
      nX[1] := nLeft + nXunit
      FOR I=2 TO 5
         nX[I] := nX[I-1] + nXbar + nXunit
      NEXT

      // Print the first line
      aPoints   := GraQueryTextBox( oPS, 'Current' )  //size of full string
      nStrWidth := aPoints[3,1]-aPoints[2,1] //lower left to upper right to allow for italics
      aPrint    := {nX[1]-nStrWidth,nBottom}
      GraStringAt( oPS, aPrint, 'Current |' )

      aPoints   := GraQueryTextBox( oPS, '30 Days' )  //size of full string
      nStrWidth := aPoints[3,1]-aPoints[2,1] //lower left to upper right to allow for italics
      aPrint    := {nX[2]-nStrWidth,nBottom}
      GraStringAt( oPS, aPrint, '30 Days |' )

      aPoints   := GraQueryTextBox( oPS, '60 Days' )  //size of full string
      nStrWidth := aPoints[3,1]-aPoints[2,1] //lower left to upper right to allow for italics
      aPrint    := {nX[3]-nStrWidth,nBottom}
      GraStringAt( oPS, aPrint, '60 Days |' )

      aPoints   := GraQueryTextBox( oPS, '90 Days' )  //size of full string
      nStrWidth := aPoints[3,1]-aPoints[2,1] //lower left to upper right to allow for italics
      aPrint    := {nX[4]-nStrWidth,nBottom}
      GraStringAt( oPS, aPrint, '90 Days |' )

      aPoints   := GraQueryTextBox( oPS, '120+ Days' )  //size of full string
      nStrWidth := aPoints[3,1]-aPoints[2,1] //lower left to upper right to allow for italics
      aPrint    := {nX[5]-nStrWidth,nBottom}
      GraStringAt( oPS, aPrint, '120+ Days' )

      nBottom -= nDepth

      FOR I=1 TO 4
         aPoints   := GraQueryTextBox( oPS, LTRIM(TRANSFORM(aVal[I],'999,999.99')) )  //size of full string
         nStrWidth := aPoints[3,1]-aPoints[2,1] //lower left to upper right to allow for italics
         aPrint    := {nX[I]-nStrWidth,nBottom}
         GraStringAt( oPS, aPrint, LTRIM(TRANSFORM(aVal[I],'999,999.99'))+' |' )
      NEXT
      aPoints   := GraQueryTextBox( oPS, LTRIM(TRANSFORM(aVal[5],'999,999.99')) )  //size of full string
      nStrWidth := aPoints[3,1]-aPoints[2,1] //lower left to upper right to allow for italics
      aPrint    := {nX[5]-nStrWidth,nBottom}
      GraStringAt( oPS, aPrint, LTRIM(TRANSFORM(aVal[5],'999,999.99')) )

      nRtn := nBottom //returns new Y coordinate for next line

RETURN nRtn

FUNCTION PrnMemo(aPos,cVar,oPS,oFont,nWidth) // note aPos is in Y,X order to ease mod of @say syntex
LOCAL aTopRight,a_size,a_size2, aPrint,aSize2,nline:= 0,TotLine,cLine,oRefFont

      SetAdjust()
      IF oFont = NIL
         oFont := oCurrFont
      ELSE
//         oCurrFont := oFont
      ENDIF
      IF PaperLand
         aPos[1]+=4
         aPos[2]+=5
      ENDIF

      oRefFont := oCurrRefFont
      IF oRefFont = NIL .OR. oFont = NIL
         MSGBOX("ERROR no defined font in printit! "+STR(aPos[1],3)+','+STR(aPos[2],3))
         RETURN NIL
      ENDIF
      DO CASE
        CASE VALTYPE(cVar)=="N"
          cVar := STR(cVar)
        CASE VALTYPE(cVar)=="D"
          cVar := DTOC(cVar)
        CASE VALTYPE(cVar)=="C"
        CASE VALTYPE(cVar)=="L"
          IF cVar
             cVar := "Y"
          ELSE
             cVar := "N"
          ENDIF
        OTHERWISE
          MSGBOX("ERROR incorrect data type in printit! "+STR(aPos[1],3)+','+STR(aPos[2],3))
      ENDCASE
      GraSetFont(oPS , oRefFont )                // reference font for spacing

      IF !oRefFont:Fixed
         a_Size := GraQueryTextBox( oPS, SPACECHR )  //spaceing based on 'L'
      ELSE
         a_Size := {oRefFont:Width * WFACTOR,oRefFont:Height * HFACTOR}
      ENDIF


      GraSetFont(oPS , oFont )                // Select font
      a_Size2 := GraQueryTextBox( oPS, SPACECHR )

      TotLine := MLCOUNT(cVar,nWidth)
      aTopRight :={aPaperSize[1]-(aPaperSize[3]+(aPaperSize[1]-aPaperSize[5])),aPaperSize[2] - (aPaperSize[4]+(aPaperSize[2]-aPaperSize[6]))}
      DO WHILE nLine < TotLine
         IF !oRefFont:Fixed
            aPrint := {aPos[2]*a_Size[3,1],aTopRight[2]-(aPos[1]*(a_Size[1,2]+LINESPACING+PrnAdjust))-nLine*a_Size2[1,2]}
         ELSE
            aPrint := {aPos[2]*a_Size[1],aTopRight[2]-(aPos[1]*(a_Size[2]+LINESPACING+PrnAdjust))-nLine*a_Size2[2]}
         ENDIF
         // Print string
         cLine := MemoLine(cVar,nWidth,nLine+1)
         GraStringAt( oPS, aPrint, cLine )
         ++nLine
      ENDDO

RETURN nLine+aPos[1]+1

FUNCTION PrVLine(aFrom,aTo,oPs,oRefFont)
LOCAL aTopRight,a_size, aPosFrom,aPosTo,aSize2

      SetAdjust()
      IF oPS = NIL
         oPS := oCurroPS
      ELSE
         oCurroPS := oPS
      ENDIF
      IF oRefFont = NIL
         oRefFont := oCurrRefFont
      ELSE
         oCurrRefFont := oRefFont
      ENDIF
      GraSetFont(oPS , oRefFont )                // reference font for spacing

      IF !oRefFont:Fixed
         a_Size := GraQueryTextBox( oPS, SPACECHR )  //spaceing based on 'L'
      ELSE
         a_Size := {oRefFont:Width * WFACTOR,oRefFont:Height * HFACTOR}
      ENDIF

      aTopRight :={aPaperSize[1]-(aPaperSize[3]+(aPaperSize[1]-aPaperSize[5])),aPaperSize[2] - (aPaperSize[4]+(aPaperSize[2]-aPaperSize[6]))}
      IF !oRefFont:Fixed
         aPosFrom := {aFrom[2]*a_Size[3,1],aTopRight[2]-(aFrom[1]*(a_Size[1,2]+LINESPACING+PrnAdjust))}
         aPosTo := {aTo[2]*a_Size[3,1],aTopRight[2]-(aTo[1]*(a_Size[1,2]+LINESPACING+PrnAdjust))}
      ELSE
         aPosFrom := {aFrom[2]*a_Size[1],aTopRight[2]-(aFrom[1]*(a_Size[2]+PrnAdjust))}
         aPosTo := {aTo[2]*a_Size[1],aTopRight[2]-(aTo[1]*(a_Size[2]+PrnAdjust))}
      ENDIF
      GRaLine( oPS,aPosFrom,aPosTo)
      ++aPosFrom[1]
      ++aPosTo[1]
      GRaLine( oPS,aPosFrom,aPosTo)
RETURN NIL

FUNCTION PrnVarCenter(nPos,cVar,oPS,oFont,oRefFont)
LOCAL aTopRight,a_size, aPrint,Size2
      SetAdjust()
      IF oFont = NIL
         oFont := oCurrFont
      ELSE
//         oCurrFont := oFont
      ENDIF

      IF oRefFont = NIL
         oRefFont := oCurrRefFont
      ELSE
//         oCurrRefFont := oRefFont
      ENDIF
      IF oPS = NIL
         oPS := oCurroPS
      ELSE
         oCurroPS := oPS
      ENDIF
      IF oRefFont = NIL .OR. oFont = NIL .OR. oPS = NIL
         MSGBOX("ERROR no defined font in printit! Center"+STR(nPos,3))
         RETURN NIL
      ENDIF
      DO CASE
        CASE VALTYPE(cVar)=="N"
          cVar := STR(cVar)
        CASE VALTYPE(cVar)=="D"
          cVar := DTOC(cVar)
        CASE VALTYPE(cVar)=="C"
        CASE VALTYPE(cVar)=="L"
          IF cVar
             cVar := "Y"
          ELSE
             cVar := "N"
          ENDIF
        OTHERWISE
          MSGBOX("ERROR incorrect data type in printit Center! "+STR(nPos,3))
      ENDCASE
      GraSetFont(oPS , oRefFont )                // reference font for spacing

      IF !oRefFont:Fixed
         a_Size := GraQueryTextBox( oPS, SPACECHR )  //spaceing based on 'L'
      ELSE
         a_Size := {oRefFont:Width * WFACTOR,oRefFont:Height * HFACTOR}
      ENDIF
//      a_Size := GraQueryTextBox( oPS, cVar )  //spaceing based on actual string
      Size2 := FontHMetric(cVar,oFont)      // probably not needed
      GraSetFont(oPS , oFont )                // Select font

      aTopRight :={aPaperSize[1]-(aPaperSize[3]+(aPaperSize[1]-aPaperSize[5])),aPaperSize[2] - (aPaperSize[4]+(aPaperSize[2]-aPaperSize[6]))}
      IF !oRefFont:Fixed
         aPrint := {(aTopRight[1]-Size2*a_Size[3,1])/2,aTopRight[2]-(nPos*(a_Size[1,2]+LINESPACING+PrnAdjust))}
      ELSE
         aPrint := {(aTopRight[1]-Size2*a_Size[1])/2,aTopRight[2]-(nPos*(a_Size[2]+PrnAdjust))}
      ENDIF
      // Print string
      GraStringAt( oPS, aPrint, cVar )
 //     oCurrRefFont := oRefFont
//      oCurrFont := oFont

RETURN NIL

FUNCTION PrnPic(cFile,aPos,aTargSize,oPS)
LOCAL oBitmap,aRect,a_Size,aTopRight,lTest,oRefFont
      IF oPS = NIL
         oPS := oCurroPS
      ELSE
         oCurroPS := oPS
      ENDIF
      oRefFont := oCurrRefFont
      IF !oRefFont:Fixed
         a_Size := GraQueryTextBox( oPS, SPACECHR )  //spaceing based on 'L'
      ELSE
         a_Size := {oRefFont:Width * WFACTOR,oRefFont:Height * HFACTOR}
      ENDIF
      aTopRight :={aPaperSize[1]-(aPaperSize[3]+(aPaperSize[1]-aPaperSize[5])),aPaperSize[2] - (aPaperSize[4]+(aPaperSize[2]-aPaperSize[6]))}

      oBitmap    := XbpBitmap():new():create( oPS )
      IF !oRefFont:Fixed
         aRect      := { aPos[2]*a_Size[3,1], aTopRight[2]-aPos[1]*(a_Size[1,2]+LINESPACING+PrnAdjust), aPos[2]*a_Size[3,1]+aTargSize[1], aTopRight[2]-aPos[1]*(a_Size[1,2]+LINESPACING+PrnAdjust)+aTargSize[2] }
      ELSE
         aRect      := { aPos[2]*a_Size[1], aTopRight[2]-aPos[1]*(a_Size[2]+PrnAdjust), aPos[2]*a_Size[1]+aTargSize[1], aTopRight[2]-aPos[1]*(a_Size[2]+LINESPACING+PrnAdjust)+aTargSize[2] }
      ENDIF
      oBitmap:loadFile( cFile )          // Load bitmap
      lTest := oBitmap:draw( oPS, aRect )  // Spool print job
      oBitmap:destroy()
RETURN NIL

FUNCTION PrnBOX(aPos,aBoxSize,nThickness,oPS,oRefFont) //aBoxSize and aPos in refernce spacing
LOCAL I,aLeftBottom,aRightTop,a_Size,aTopPage
      SetAdjust()
      IF Valtype(nThickness) != "N"
         nThickness = 1
      ENDIF
      IF oPS = NIL
         oPS := oCurroPS
      ELSE
         oCurroPS := oPS
      ENDIF
      IF PaperLand
         aPos[1]+=4
         aPos[2]+=5
      ENDIF
      IF oRefFont = NIL
         oRefFont := oCurrRefFont
      ELSE
         oCurrRefFont := oRefFont
      ENDIF
      IF oRefFont = NIL .OR. oPS = NIL
         MSGBOX("ERROR no defined font in printit BOX! "+STR(nPos,3))
         RETURN NIL
      ENDIF
      GraSetFont(oPS , oRefFont )                // reference font for spacing

      IF !oRefFont:Fixed
         a_Size := GraQueryTextBox( oPS, SPACECHR )  //spaceing based on 'L'
      ELSE
         a_Size := {oRefFont:Width * WFACTOR,oRefFont:Height * HFACTOR}
      ENDIF
      aTopPage :={aPaperSize[1]-(aPaperSize[3]+(aPaperSize[1]-aPaperSize[5])),aPaperSize[2] - (aPaperSize[4]+(aPaperSize[2]-aPaperSize[6]))}
      IF !oRefFont:Fixed
         aLeftBottom := {aPos[2]*a_Size[3,1],aTopPage[2]-(aPos[1]*(a_Size[1,2]+LINESPACING+PrnAdjust))} // remember X,Y reversed in aPos for Clipper compat
         aRightTop := {aLeftBottom[1]+a_Size[3,1]*aBoxSize[2],aLeftBottom[2]+a_Size[3,2]*aBoxSize[1]}
      ELSE
         aLeftBottom := {aPos[2]*a_Size[1],aTopPage[2]-(aPos[1]*(a_Size[2]+PrnAdjust))} // remember X,Y reversed in aPos for Clipper compat
         aRightTop := {aLeftBottom[1]+a_Size[1]*aBoxSize[2],aLeftBottom[2]+a_Size[2]*aBoxSize[1]}
      ENDIF
      FOR I = 1 TO nThickness
         GraBox(oPS,aLeftBottom,aRightTop)
         ++aLeftBottom[1]
         ++aLeftBottom[2]
         --aRightTop[1]
         --aRightTop[2]
      NEXT I
RETURN NIL

FUNCTION PrnDataBOX(aPos,aBoxSize,cPrompt,cData,nThickness,oPS,oFontPrompt,oFont,oRefFont) //aBoxSize and aPos in refernce spacing
LOCAL aRtn
   // Draw the box
   PrnBOX(aPos,aBoxSize,nThickness,oPS,oRefFont)
   PrnVar({aPos[1]-aBoxSize[1]+1 ,aPos[2]+1},cPrompt,oPS,oFontPrompt,oRefFont)
   PrnVar({aPos[1] ,aPos[2]+1},Alltrim(cData),oPS,oFont,oRefFont)

RETURN NIL  // number of refernce spacing used (Y,X) format


FUNCTION FontHMetric(cStr,oFont) // return graphic length ratio of str to SPACECHR
STATIC oRefFont,nRtn,a_Size,prevStr:=""
LOCAL oPS,a_Size2
   IF cStr = NIL
      RETURN 0
   ENDIF
   IF PrevStr == cStr
      RETURN nRtn
   ELSE
      prevStr := cStr
   ENDIF
   IF EMPTY(oRefFont)
      oRefFont := oCurrRefFont
   ENDIF
   oPS := oCurroPS
   IF empty(a_Size) .OR. oRefFont != oCurrRefFont
      oRefFont := oCurrRefFont
      GraSetFont(oPS , oRefFont )                // reference font for spacing
      a_Size  := GraQueryTextBox( oPS, SPACECHR )  //spaceing based on 'L'
   ENDIF
   GraSetFont(oPs , oFont )
   a_Size2 := GraQueryTextBox( oPS, cStr )  //spaceing based on 'cStr'
   nRtn := a_Size2[3,1]/a_Size[3,1]    //ratio
   GraSetFont(oPS , oCurrFont )
RETURN nRtn

FUNCTION FontVMetric(cStr,oFont) // return graphic length ratio of str to SPACECHR
LOCAL nRtn,oRefFont,oPS,a_Size,a_Size2
   IF cStr = NIL
      RETURN 0
   ENDIF
   oRefFont := oCurrRefFont
   oPS := oCurroPS
   GraSetFont(oPS , oRefFont )                // reference font for spacing
   a_Size  := GraQueryTextBox( oPS, SPACECHR )  //spaceing based on 'L'
   GraSetFont(oPs , oFont )
   a_Size2 := GraQueryTextBox( oPS, cStr )  //spaceing based on 'cStr'
   nRtn := a_Size2[3,2]/a_Size[3,2]    //ratio
   GraSetFont(oPS , oCurrFont )
RETURN nRtn

FUNCTION PrinterPS( cPrinterObjectName,cIsLabel )
      LOCAL oPS, oDC := XbpPrinter():New()

      oDC:Create( cPrinterObjectName )
      IF cIslabel = "Label"
         IF oDC:SetFormSize(XBPPRN_FORM_USER) = NIL

         ENDIF
      ENDIF

      oPS := XbpPresSpace():New()
      oPS:Create( oDC, oDC:paperSize(), GRA_PU_LOMETRIC )
RETURN oPS

PROCEDURE DestroyDevice( oPS )
      LOCAL oDC := oPS:device()

      IF oDC <> NIL
         oPS:configure()
         oDC:destroy()
      ENDIF
RETURN

FUNCTION StartPrinter(cfont,nSize,cPrinterObjectName,cIsLabel,lPreview,o,nOrient,nCopies,cDefault)
LOCAL oPS, oPrinter,aPageSize,I,aPrinters //aPaperSize STATIC
     // PRINT FUNCTION
     IF nOrient = NIL
        nOrient := XBPPRN_ORIENT_PORTRAIT
     ENDIF
     IF lPreview = NIL
        lPreview := .F.
     ENDIF
     IF nOrient = XBPPRN_ORIENT_PORTRAIT
        PaperLand := .F.
     ELSE
        PaperLand := .T.
     ENDIF
     IF nCopies = NIL
        nCopies := 1
     ENDIF
     IF cDefault = NIL
        cDefault := ''
     ENDIF
     IF cPrinterObjectName = NIL
        IF .NOT. EMPTY(cDefault) .AND. FILE(cDefault)
           oPrinter := XbpPrinter():new()
           oPrinter:create()
           oPrinter:devName := MemoRead(cDefault)
           oPrinter:SetNumCopies(nCopies)
        ELSE
           oDlg := XbpPrintDialog():new()
           oDlg:create()
           oDlg:numCopies := nCopies
           oPrinter := oDlg:display()
           oDlg:destroy()
        ENDIF
        IF oPrinter = NIL
           MsgBox("Printing aborted!")
           Return {}
        ENDIF
        DO WHILE ActiveReports > 0      // spool reports we only allow one report to print at a time
           SLEEP(100)
        ENDDO
        oPrinter:setOrientation(nOrient)
        oPS := XbpPresSpace():New()
        aPaperSize  := oPrinter:paperSize()
        IF !lPreview
           oPS:Create( oPrinter, aPaperSize, GRA_PU_LOMETRIC )
        ELSE
           oPS:mode := XBPPS_MODE_HIGH_PRECISION
           aPageSize := { aPaperSize[5]-aPaperSize[3], aPaperSize[6]-aPaperSize[4] }
         // Link presentation space with printer device context
         // The PS is created with the metric unit 1/10 mm
           oPS:create( oPrinter, aPageSize , GRA_PU_LOMETRIC)
        ENDIF
     ELSE
        DO WHILE ActiveReports > 0      // spool reports we only allow one report to print at a time
           SLEEP(100)
        ENDDO
        oPS := PrinterPS(ALLTRIM(cPrinterObjectName),cIsLabel)
     ENDIF
     oFont := XbpFont():new(oPS)              // Create XbpFont object

     oFont:familyName := cFont // Describe font
     oFont:nominalPointSize := nSize
     oFont:generic := .T.
     oFont:create()                          // Create font
     // Activate spooling
     IF lPreview
        oPreview := XbpPreview():New()
        oPreview:oPresSpace := oPS
        oPreview:Printer := oPrinter
        // Start first page
        oPreview:nPage := 1
        oDlgPreview := Preview(o,oPreview,nOrient)
        BusyPointer(oDlgPreview:DrawingArea,.T.)
        GraSegDrawMode( oPS, GRA_DM_RETAIN )
        AADD(oPreview:segments,GraSegOpen( oPS ))
        BusyPointer(oDlgPreview:DrawingArea,.F.) //NEED TO GET RID OF IT
        RtnScreen := o
     ELSE
        oPS:device():startDoc()
     ENDIF
     oCurroPS := oPS
     // Set default values for margins for metric functions
     aPaperSize  := oPrinter:paperSize()
     SetRightMargin(aPaperSize[5]) // right margin
     SetLeftMargin(aPaperSize[3]) // left margin
     SetTopMargin(aPaperSize[6]) // top margin
     SetBottomMargin(aPaperSize[4]) // bottom margin

     ++ActiveReports
RETURN {oPS,oFont,oPrinter}

FUNCTION Printer_EJECT(oPs)
   IF oPS = NIL
      oPS := oCurroPS
   ENDIF
   IF EMPTY(oPreview)
      oPS:device():NewPage()
   ELSE
      GraSegClose( oPS )
      //Render current
      oPreview:drawPage()
      //Start new page
      AADD(oPreview:segments,GraSegOpen( oPS ))
      oPreview:nPage := oPreview:nPage + 1
      ++oPreview:nPagePos
   ENDIF
RETURN NIL

FUNCTION End_Doc(oPs)
   IF oPS = NIL
      oPS := oCurroPS
   ENDIF
   IF EMPTY(oPreview)
      oPS:device():EndDoc()
   ENDIF
RETURN NIL

FUNCTION EndPrinter(oPS)
LOCAL nEvent, mp1, mp2
LOCAL I,J
     // End spooling
  IF EMPTY(oPreview)
     oPS:device():endDoc()
     DestroyDevice( oPS )
  ELSE
     GraSegClose( oPS ) //End Current Page
     GraSegDrawMode( oPS, GRA_DM_DRAWANDRETAIN )
//     BusyPointer(oDlgPreview:DrawingArea,.T.) Don't want it in the display
     oPreview:nPagePos := 1
     oPreview:DrawPage()
     nEvent := 0
     DO WHILE oDlgPreview:Status() = XBP_STAT_CREATE
        nEvent := AppEvent( @mp1, @mp2, @oXbp )
        oXbp:handleEvent( nEvent, mp1, mp2 )
     ENDDO
     oPreview := NIL
     IF !EMPTY(RtnScreen)
        SetAppFocus(RtnScreen)
     ENDIF
  ENDIF
  --ActiveReports
RETURN NIL

FUNCTION PicSay(xVal,cPic)
   IF !EMPTY(cPic)
      xVal := Transform(xVal,cPic)
   ENDIF
RETURN xVal
