///////////////////////////////////////////////////////////////////////////////
//
//  XCALC.PRG
//
//  Contents :
//      Calculator example program
//
//  Remarks:
//    This software is published under the GNU General Public License (GPL)
//  
//      http://www.opensource.org/licenses/gpl-license.php
//  
//    NO WARRANTY
//  
//    BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
//    FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW. EXCEPT
//    WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER
//    PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND, 
//    EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE 
//    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
//    PURPOSE. THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE 
//    PROGRAM IS WITH YOU. SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME 
//    THE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
//  
//    IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
//    WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR 
//    REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR 
//    DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL 
//    DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THE PROGRAM 
//    (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
//    INACCURATE OR LOSSES SUSTAINED BY YOU OR THIRD PARTIES OR A FAILURE
//    OF THE PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS), EVEN IF SUCH 
//    HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE POSSIBILITY OF 
//    SUCH DAMAGES.
//
//  Author and copyright
//    (c) Hannes Ziegler 2005
//
//        www.knowleXbase.com
//
///////////////////////////////////////////////////////////////////////////////

#include "Gra.ch"
#include "Xbp.ch"
#include "Common.ch"
#include "Appevent.ch"
#include "Leddisp.ch"

#xtrans NTrim( <x> )       => LTrim(Str(<x>))


******************************************************************************
* This class is derived from the implementation-level class of the form.
* Instance variables are declared in the _XppCalculator class.
******************************************************************************
CLASS XppCalculator FROM _XppCalculator
   EXPORTED:
   VAR nOperand1
   VAR nOperand2
   VAR cOperator
   VAR cBuffer
   VAR nValue
   VAR nResult
   VAR nMemory
   VAR lClear
   VAR aButtons 
   VAR oSLE
//   VAR oHelp
//   VAR aTooltips

   METHOD init,destroy
   METHOD create
   METHOD _keyBoard
   METHOD enter
   METHOD calculate
ENDCLASS

******************************************************************************
* Initialize form
******************************************************************************
METHOD XppCalculator:INIT( oParent, oOwner, aPos, aSize, aPP, lVisible )

   ::nOperand1 := 0
   ::nOperand2 := 0
   ::nResult   := 0
   ::cOperator := ""
   ::cBuffer   := ""
   ::lClear    :=.F.
   ::nValue    := 0
   ::nMemory   := 0    
   
   ::MoveWithOwner := .T.
   ::Close := {||::Destroy()}

   * Execute method of the super class
   ::_XppCalculator:init( oParent, oOwner, aPos, aSize, aPP, lVisible )
   ::_XppCalculator:setDisplayFocus := {|| SetAppFocus( ::drawingArea ) }
   ::_XppCalculator:minButton   := .F.
   ::_XppCalculator:maxButton   := .F.
   ::_XppCalculator:border      := XBPDLG_RECESSEDBORDERTHICK_FIXED

   ::drawingArea:bitmap         :=  BMP_BACKGROUND
   ::drawingArea:keyboard       :=  {|nKey| ::_keyBoard( nKey ) }

   ::ledDisplay :withExponent   := .T.

   ::aButtons := { ;
      { ::btnClearAll  , BMP_BTN_CLEAR      , xbeK_CTRL_C, "Clear all = Ctrl+C"               }, ;
      { ::btnClearEntry, BMP_BTN_CLEARENTRY , xbeK_SPACE , "Clear entry = Space bar"          }, ;
      { ::btnBackSpace , BMP_BTN_BACKSPACE  , xbeK_BS    , "Delete last digit = Back space"   }, ;
      { ::btn7         , BMP_BTN_7          , Asc( "7" ) , "Digit = 7"                        }, ;
      { ::btn8         , BMP_BTN_8          , Asc( "8" ) , "Digit = 8"                        }, ;
      { ::btn9         , BMP_BTN_9          , Asc( "9" ) , "Digit = 9"                        }, ;
      { ::btn4         , BMP_BTN_4          , Asc( "4" ) , "Digit = 4"                        }, ;
      { ::btn5         , BMP_BTN_5          , Asc( "5" ) , "Digit = 5"                        }, ;
      { ::btn6         , BMP_BTN_6          , Asc( "6" ) , "Digit = 6"                        }, ;
      { ::btn1         , BMP_BTN_1          , Asc( "1" ) , "Digit = 1"                        }, ;
      { ::btn2         , BMP_BTN_2          , Asc( "2" ) , "Digit = 2"                        }, ;
      { ::btn3         , BMP_BTN_3          , Asc( "3" ) , "Digit = 3"                        }, ;
      { ::btn0         , BMP_BTN_0          , Asc( "0" ) , "Digit = 0"                        }, ;
      { ::btnDot       , BMP_BTN_DOT        , Asc( "." ) , "Decimal point = ."                }, ;
      { ::btnSign      , BMP_BTN_SIGN       , Asc( "#" ) , "Reverse sign = #"                 }, ;
      { ::btnEqual     , BMP_BTN_EQUAL      , Asc( "=" ) , "Calculate = Equal sign or Return" }, ;
      { ::btnPlus      , BMP_BTN_PLUS       , Asc( "+" ) , "Add = +"                          }, ;
      { ::btnMinus     , BMP_BTN_MINUS      , Asc( "-" ) , "Subtract = -"                     }, ;
      { ::btnMultiply  , BMP_BTN_MULTIPLY   , Asc( "*" ) , "Multiply = *"                     }, ;
      { ::btnDivide    , BMP_BTN_DIVIDE     , Asc( "/" ) , "Divide = /"                       }, ;
      { ::btnMemStore  , BMP_BTN_MEM_STORE  , xbeK_ALT_S , "Store value in memory = Alt+S"    }, ;
      { ::btnMemPlus   , BMP_BTN_MEM_PLUS   , xbeK_ALT_A , "Add value to memory = Alt+A"      }, ;
      { ::btnMemRecall , BMP_BTN_MEM_RECALL , xbeK_ALT_R , "Recall memory value = Alt+R"      }, ;
      { ::btnMemClear  , BMP_BTN_MEM_CLEAR  , xbeK_ALT_C , "Clear memory = Alt+C"             }, ;
      { ::btnOneDiv    , BMP_BTN_ONEDIVX    , Asc( "r" ) , "One divided by x = R"             }, ;
      { ::btnPower2    , BMP_BTN_POWER2     , Asc( "s" ) , "Square = S"                       }, ;
      { ::btnRoot      , BMP_BTN_ROOT       , Asc( "q" ) , "Square root = Q"                  }, ;
      { ::btnPowerY    , BMP_BTN_POWERY     , Asc( "p" ) , "Power of = P"                     }, ;
      { ::btnLogN      , BMP_BTN_LOGN       , Asc( "n" ) , "Natural logarithm = N"            }, ;
      { ::btnEPower    , BMP_BTN_EPOWERX    , Asc( "e" ) , "Natural anti-logarithm = E"       }, ;
      { ::btnLog10     , BMP_BTN_LOG10      , Asc( "l" ) , "Logarithm base 10 = L"            }, ;
      { ::btn10Power   , BMP_BTN_10POWERX   , Asc( "t" ) , "Anti-logarithm base 10 = T"       }  ;
   }

/*   ::aTooltips := Array( Len( ::aButtons ) )

   AEval( ::aButtons, {|a,i| ::aTooltips[i] := { a[2], a[4] } } )

   ::oHelp := MagicHelp( ::aToolTips ):new() */

   AEval( ::aButtons, {|a,i| ;
            a[1]:pointerFocus := .F.  , ;
            a[1]:caption      := a[2] , ;
            a[1]:cargo        := i    , ;
            a[1]:activate     := {|x,y,o| ::enter( ::aButtons[o:cargo,3] ) }, ;
            a[1]:helpLink     := NIL ;
        } )

RETURN self


******************************************************************************
* Request system resources
******************************************************************************
METHOD XppCalculator:create( oParent, oOwner, aPos, aSize, aPP, lVisible )

   * Execute method of the super class
   ::_XppCalculator:create( oParent, oOwner, aPos, aSize, aPP, lVisible )

   * Display the form
   ::show()

   ::LedDisplay:setData( ::nValue )

//   ::oHelp:start()
RETURN self


******************************************************************************
* Animate the pushbuttons when a key is pressed
******************************************************************************
METHOD XppCalculator:_keyboard( nKey )
   LOCAL i:= AScan( ::aButtons, {|a| a[3] == nKey } )

   IF i > 0
      // press button down -> synchronous event handling
      ::aButtons[i,1]:handleEvent( xbeM_LbDown, {1,1}, NIL )

      // hold it depressed
      Sleep( 10 )

      // Release button -> asynchronous event handling
      // The lbUp event results in xbeP_Activate since the
      // mouse pointer is simulated at position {1,1} inside the button
      PostAppEvent( xbeM_LbUp  , {1,1},, ::aButtons[i,1] )
   ELSE
      ::enter( nKey )
   ENDIF
RETURN

******************************************************************************
* Handle user input
******************************************************************************
METHOD XppCalculator:enter( nKey )
   LOCAL lUpdate := .T., cChr := "", oError
   LOCAL bError  := ErrorBlock( {|e| Break(e) } )

   IF nKey < 255
      cChr := Chr( nKey )
   ENDIF

   BEGIN SEQUENCE
      DO CASE
      CASE nKey == xbeK_ALT_X
         PostAppEvent( xbeP_Close )

      CASE nKey == xbeK_CTRL_C
         ::cBuffer   := ""
         ::cOperator := ""
         ::nOperand1 := 0
         ::nOperand2 := 0
         ::nValue    := 0
         ::nResult   := 0
         ::LedDisplay:minus := .F.

      CASE nKey == xbeK_ALT_S
         ::nMemory := ::nValue

      CASE nKey == xbeK_ALT_A
         ::nMemory += ::nValue

      CASE nKey == xbeK_ALT_R
         ::nValue := ::nMemory

      CASE nKey == xbeK_ALT_C
         ::nMemory := 0

      CASE nKey == xbeK_SPACE
         ::cBuffer  := ""
         ::nValue   := 0
         ::LedDisplay:minus := .F.

      CASE nKey == xbeK_BS
         ::cBuffer := SubStr( ::cBuffer, 1, Len( ::cBuffer )-1 )
         ::nValue  := Val( ::cBuffer )
         IF ::LedDisplay:minus .AND. ::nValue == 0
            ::LedDisplay:minus := .F.
         ENDIF

      CASE cChr $ "rR"               //  1/x = x reverse
         ::nValue  := 1 / ::nValue
         ::lClear := .T.

      CASE cChr $ "sS"               //  x^2 = x square
         ::nValue ^= 2
         ::lClear := .T.

      CASE cChr $ "qQ"               //  x^2 = x square root
         ::nValue := SQrt( ::nValue )
         ::lClear := .T.

      CASE cChr $ "lL"               //  Log10 x
         ::nValue := Log( ::nValue ) / Log( 10 )
         ::lClear := .T.

      CASE cChr $ "nN"               //  Ln x
         ::nValue := Log( ::nValue )
         ::lClear := .T.

      CASE cChr $ "eE"               //  e ^ x
         ::nValue := Exp( ::nValue )
         ::lClear := .T.

      CASE cChr $ "tT"               //  10 ^ x
         ::nValue := 10 ^ ::nValue
         ::lClear := .T.

      CASE cChr == "#"
         ::nValue  := - ::nValue
         ::ledDisplay:minus := ( ::nValue < 0 )
         ::cBuffer := NTrim( ::nValue )

      CASE cChr $ "0123456789"
         IF ::lClear
            ::lClear  := .F.
            ::cBuffer := ""
         ENDIF

         IF Len( ::cBuffer ) < ::ledDisplay:digits .OR. ::ledDisplay:withExponent
            IF Val( ::cBuffer ) == 0 
               IF ::cBuffer == "-0"
                  ::cBuffer := "-" + cChr
               ELSEIF "." $ ::cBuffer
                  ::cBuffer += cChr
               ELSE
                  ::cBuffer := cChr
               ENDIF
            ELSE
               ::cBuffer += cChr
            ENDIF

            ::nValue  := Val( ::cBuffer )
         ELSE
            lUpdate := .F.
            Tone( 1000 )
         ENDIF

         IF ::ledDisplay:minus .AND. ! "-" $ ::cBuffer
            ::cBuffer := "-" + ::cBuffer
            ::nValue  := Val( ::cBuffer )
         ENDIF

      CASE cChr $ ".,"
         IF ! "." $ ::cBuffer
            IF ::lClear .OR. ::cBuffer == ""
               ::lClear  := .F.
               IF ::ledDisplay:minus
                  ::cBuffer := "-0"
               ELSE
                  ::cBuffer := "0"
               ENDIF
            ENDIF
            ::cBuffer += "."
         ELSE
            lUpdate := .F.
            Tone( 1000 )
         ENDIF

      CASE cChr == "=" .OR. nKey == xbeK_RETURN
         IF Len( ::cOperator ) == 0
            Tone( 1000 )
            lUpdate := .F.
         ELSE
            ::nOperand2 := ::nValue
            ::calculate()
            ::nOperand1 := 0
            ::cOperator := ""
         ENDIF

      CASE cChr $ "+-*/pP"
         IF cChr $ "pP"
            cChr := "^"
         ENDIF

         IF Len( ::cOperator ) == 0
            IF cChr == "-" .AND.  ::nValue == 0
               ::LedDisplay:minus := .T.
            ELSE
               ::cOperator := cChr
               IF ::nOperand1 == 0
                  ::nOperand1 := ::nValue
               ENDIF
               ::cBuffer  := ""
               ::nValue   := 0
               ::LedDisplay:minus := .F.
            ENDIF
         ELSE
            IF cChr == "-" .AND.  ::nValue == 0
               ::LedDisplay:minus := .T.
            ELSE
               ::nOperand2 := ::nValue
               ::calculate()
               ::cOperator := cChr
            ENDIF
         ENDIF
      ENDCASE

      IF lUpdate
         IF ::nValue == 0
            // must pass ::cBuffer for 0.000 entries
            IF ::cBuffer == "" .AND. ::LedDisplay:minus
               ::cBuffer := "-0"
            ENDIF
            ::LedDisplay:setData( ::nValue, ::cBuffer )
         ELSE
            ::LedDisplay:setData( ::nValue )
         ENDIF
      ENDIF

   RECOVER USING oError                // error object for debugging only
      ::cBuffer   := ""
      ::cOperator := ""
      ::nOperand1 := 0
      ::nOperand2 := 0
      ::nValue    := 0
      ::nResult   := 0
      ::lClear    := .F.

      ::LedDisplay:error()
   ENDSEQUENCE

   ErrorBlock( bError )
   SetAppFocus( ::drawingArea )
   IF nKey == xbeK_ESC  
      EVAL(::Close) 
   ENDIF   
RETURN



******************************************************************************
* Do the calculation
******************************************************************************
METHOD XppCalculator:calculate
   LOCAL cBlock, bBlock

   IF ::cOperator  $  "+-*/^"
      cBlock      := "{|n1,n2| n1 " + ::cOperator + " n2 }"
      bBlock      := &(cBlock)
      ::nResult   := Eval( bBlock, ::nOperand1, ::nOperand2 )
      ::ledDisplay:setData( ::nResult )
      ::nValue    := ::nResult
      ::nOperand1 := ::nResult
      ::cOperator := ""
      ::nOperand2 := 0
      ::cBuffer   := NTrim( ::nResult )
      ::lClear    := .T.
   ENDIF

RETURN self

METHOD XppCalculator:Destroy() 
   IF !EMPTY(::oSLE) //points to a control
      ::oSLE:SetData(::nValue)
      SetAppFocus(::oSLE)
   ENDIF
   ::_XppCalculator:Destroy()
RETURN self




PROCEDURE xCalcFunc(nVal,oSLE)
   LOCAL nEvent, oXbp, mp1, mp2,oDlg

   oDlg := XppCalculator():New(SetAppWindow():DrawingArea,ParentDialog(oSLE))
   oDlg:nValue := VAL(STRTRAN(oSLE:EditBuffer(),",","")) 
   oDlg:oSLE   := oSLE
   oDlg:Create() 
   oDlg:setModalState( XBP_DISP_APPMODAL)
   SetAppFocus( oDlg )
   ModalLoop(oDlg)  
   SetAppFocus(oSLE)
RETURN

* Include program code for the implementation-level class of the form
#include "_XCALC.PRG"

//EOF
/////